/*                                                                
**  Copyright (C) 1996,2007-2010,2013-2015  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*H**************************************************************************

 * FILE NAME: acis_process_events.c

 * DEVELOPEMENT: tools

 * DESCRIPTION:

 Acis_process_events program.  This tool is an ACIS Level 1 processing tool.
 It has a dual purpose. Depending on input parameters, the tool will calculate
 various coordinates, grading values, or both for events from an event list.
 The tool takes one or more event files and a parameter file as input and
 creates an output qpoe file with the requested information.

 * NOTES:

 The order of the individual pha columns of the pixels that make up an
 event island are expected to be listed in the following order:

 the pixels in the island are ordered from lower left to upper right
 with the local maxima at the center of the event island.

 ie. for a 3x3 time exposure event : A  B  C
 D  E  F
 G  H  I where E is the local max

 the pha columns should be ordered: G H I D E F A B C

 for a continuous clocking event: A  B  C where B is the local max

 the pha columns should be ordered: A B C

 for a 5x5 timed exposure event: A  B  C  D  E
 F  G  H  I  J    where M is the
 K  L  M  N  O      local max
 P  Q  R  S  T
 U  V  W  X  Y

 the pha columns should be ordered:
 U V W X Y P Q R S T K L M N O F G H I J A B C D E

 *H**************************************************************************/

#include "parameter.h"

#ifndef ACIS_PROCESS_EVENTS_H
#include "acis_process_events.h"
#endif

#ifndef L1_ASPECT_DEFS_H
#include "l1_aspect_defs.h"
#endif

#ifndef ACIS_GAIN_DEFS_H
#include "acis_gain_defs.h"
#define ACIS_GAIN_DEFS_H
#endif

#include "bad_pixel_defs.h"

extern void process_grades_acis(EVENT_SETUP_P_T,
                                short*,
                                CTI_TABLE_P_T,
                                EVENT_REC_P_T,
                                GAIN_TABLE_UPT,
                                INPUT_PARMS_P_T,
                                STATISTICS_P_T,
                                dsErrList*);

dsErrCode acis_process_events(void)
{
  /* CL INPUT PARAMETERS */
  INPUT_PARMS_T inpars; /* I - hold parameters used for processing */
  INPUT_PARMS_P_T inp_p = &inpars; /* I - pointer to input parameters      */
  register int debug; /* register reference to debug param value */
  FILE *log_ptr; /* L - pointer to logfile                  */

  /* DITHER VARIABLES */
  ALIGNMENT_INFRA_T aln_infra;
  ALIGNMENT_INFRA_P_T aln_hk_p = &aln_infra;
  ALIGNMENT_REC_T aln_rec;
  ALIGNMENT_REC_P_T aln_p = &aln_rec;

  /* ASPECT CORRECTION VARIABLES */
  ASPECT_INFRA_T a_infra; /* structure holding aspect file info      */
  ASPECT_INFRA_P_T asp_hk_p = &a_infra; /* pntr to aspect Housekeeping info*/
  ASPECT_ENTRY_T asp_rec; /* data struct to hold aspect file record  */
  ASPECT_ENTRY_P_T asp_p = &asp_rec; /* pointer to a aspect file record    */

  /* "STACK" VARIABLES */
  Stack evtstack; /* input event file(s) 'stack'             */
  int num_evtfile = 0; /* number of event files in 'stack'        */

  /* OUTPUT EVENT FILE VARIABLES */
  EVENT_SETUP_T evt_out; /* output event file information           */
  EVENT_SETUP_P_T evtout_p = &evt_out;
  EVENT_REC_T evt; /* struct containing data for an event     */
  EVENT_REC_P_T evt_p = &evt; /* pointer to the event structure          */

  STATISTICS_T stat; /* struct to keep track of event statistics*/
  STATISTICS_P_T stat_p = &stat;/* pointer to statistics structure         */
  GAIN_TABLE_UT gtable; /* struct for gain table data              */
  GAIN_TABLE_UPT gain_table = &gtable; /* pointer to gain table structure */
  BADPIX_SETUP_P_T badpix_p = NULL; /* pointer to bad pixel structure  */

  boolean setup_outfile = TRUE; /* flag to init output file only once      */
  char error_info[1] = {'\0'}; /* string to print in err msg              */
  INST_KEYWORDS_T inst; /*  instrument keyword data structure      */
  INST_KEYWORDS_P_T inst_p = &inst; /* ptr to instrument keyword structure */
  short* grade_map = NULL; /* array to store grade mappings           */
  dsErrCode err = dsNOERR; /* return error status for this routine    */
  dsErrList* err_p = NULL; /* pointer to list of errors/warnings      */

  /* cti variables */
  CTI_TABLE_T cti; /* pointer to CTI ARD table descriptors    */
  CTI_TABLE_P_T cti_p = &cti; /* pointer to CTI ARD table descriptors   */
  short applied_cti = 0; /* was the cti adjustment applied?         */

  /* tgain variables */
  TGAIN_TABLE_T tgain; /* pointer to TGAIN ARD table descriptors  */
  TGAIN_TABLE_P_T tgain_p = &tgain; /* pointer to TGAIN ARD descriptors  */

  int status = 0;

  SUBPIX_CAL subpix_t; /* Data to hold subpixel calibrations */

  inp_p->do_tgain = 0;
  inp_p->use_pharo_for_pha = 0;
  inp_p->copy_pharo_to_output = 0;
  inp_p->copy_pha_to_output = 0;
  inp_p->pharo_is_zero = 0;
  inp_p->pharo_in_eventdef = 0;

  /* initialize error list */
  dsErrCreateList(&err_p);

  memset(evtout_p, 0, sizeof(EVENT_SETUP_T));
  memset(aln_p, 0, sizeof(ALIGNMENT_REC_T));
  memset(aln_hk_p, 0, sizeof(ALIGNMENT_INFRA_T));
  memset(asp_p, 0, sizeof(ASPECT_ENTRY_T));
  memset(asp_hk_p, 0, sizeof(ASPECT_INFRA_T));

  /* read parameters from acis_process_events.par */
  status = acis_process_read_parameters(inp_p, stat_p, err_p);
  if (status != dsNOERR)
  {
    dsErrPrintList(err_p, dsErrTrue);
    return (-1);
  }

  stat_p->err_info_p = &error_info[0];
  debug = inp_p->debug;

  /* setup debug file if necessary */
  inp_p->log_p = log_ptr = acis_process_setup_debug_logfile(inp_p, err_p);

  /* read obs.par file */
  error_check_obsfile(inp_p, err_p);

  /* make stack */
  evtstack = stk_build(inp_p->stack_in);
  num_evtfile = stk_count(evtstack);

  status = error_check_infile(inp_p, err_p);
  if (status != dsNOERR)
  {
    dsErrPrintList(err_p, dsErrTrue);
    return (-1);
  }
  else
  {
    if (debug > DEBUG_LEVEL_0)
    {
      fprintf(log_ptr, " stack count is %2d items ...\n", num_evtfile);
    }
    if (num_evtfile < 1)
    {
      dsErrAdd(err_p, dsAPEEMPTYSTACKERR, Individual, Generic);
    }

    if (error_check_acaofffile(inp_p, err_p) == 0)
    {
      strcpy(inp_p->asp_file, "NONE");
      strcpy(inp_p->dth_file, "NONE");
      inp_p->asp_file_p = NULL;
      inp_p->dth_file_p = NULL;
      inp_p->use_asolfile = 0;
    }
    else
    {
      inp_p->asp_file_p = &inp_p->asp_file[0];
      inp_p->dth_file_p = &inp_p->asp_file[0];
    }

    /* aspect housekeeping */
    asp_hk_p->stack = stk_build(inp_p->asp_file_p);
    asp_hk_p->stk_cnt = stk_count(asp_hk_p->stack);
    asp_hk_p->asp_file = NULL;
    asp_p->curr = 0;
    asp_p->next = 1;
    /* alignment housekeeping */
    aln_hk_p->stack = stk_build(inp_p->dth_file_p);
    aln_hk_p->stk_cnt = stk_count(aln_hk_p->stack);
    aln_hk_p->file = NULL;

    error_check_stop(inp_p, err_p);

    /* check badpix file */
    error_check_badpixfile(inp_p);

    error_check_pix_adj(inp_p);
  }

  /* start going through stack */
  while (((inp_p->infile = stk_read_next(evtstack)) != NULL)
      && (dsErrGetFatalCt(err_p) == 0))
  {
    EVENT_SETUP_T evt_in; /* input event file information */
    EVENT_SETUP_P_T evtin_p = &evt_in;
    long row_check = dmSUCCESS;
    long num_rows = -1;
    inp_p->is_cc33_mode = 0;
    inp_p->is_tg = 0;

    /* open input file */
    acis_process_setup_input_file(evtin_p, inp_p, stat_p, err_p);

    if (set_ra_dec_targ_vals(inp_p, err_p))
    {
      return -1;
    }

    num_rows = dmTableGetNoRows(evtin_p->extension);
    if (0 == num_rows || dmBADROW == num_rows)
      dsErrAdd(err_p, dsPARAMVALUEWERR, Individual, Custom, "WARNING: Input dataset contains no rows.");

    if ((setup_outfile) && (dsErrGetFatalCt(err_p) == 0))
    {
      if (inp_p->apply_cti && ((inp_p->in_mode & MODE_GRD_TE) != 0
          || (inp_p->in_mode & MODE_GRD_CC) != 0 || (inp_p->in_mode
          & MODE_GRD_CC3x3) != 0 || (inp_p->in_mode & MODE_GRD_HISTOGRAM) != 0))
      {
        evtin_p->island_sz = TE_ISLAND_SZ;
      }

      /* substitute calibration product file names */
      access_caldb(evtin_p, inp_p, err_p);

      err = acis_process_configure_pixlib(evtin_p, inp_p, inst_p, aln_p, err_p);
      if (err)
        return (err);

      /* load grade table if necessary */
      if (inp_p->grade && num_rows > 0)
      {
        load_grade_map(inp_p, &grade_map, err_p);
      }

      /* load gain table */
      setup_acis_gain_table(inp_p, gain_table, err_p);

      /* load tgain trap density table if necessary */
      /* if (inp_p->apply_tgain && inp_p->doevtgrade) */
      if (inp_p->apply_tgain && num_rows > 0)
      {
        if ((ds_strcmp_cis(inp_p->tgainfile, "NONE") != 0)
            && (strcmp(inp_p->tgainfile, "\0") != 0)
            && (dmDatasetAccess(inp_p->tgainfile, "R") == dmTRUE))
        {
          load_tgain_map(inp_p, evtin_p, tgain_p, err_p);
        }
        else
        {
          inp_p->apply_tgain = 0;
          inp_p->do_tgain = 0;
          dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "WARNING: problem reading tgainfile, tgain adjustment will "
            "not be applied. Changing apply_tgain=yes to apply_tgain=no.");
          dmResetErrorState();
        }
      }
      set_tgain_input_conditions(inp_p, evtin_p->extension, err_p);
      set_tgain_output_conditions(inp_p, err_p);

      /* load cti trap density table if necessary */
      if (inp_p->apply_cti && num_rows > 0)
      {
        if ((ds_strcmp_cis(inp_p->ctifile, "NONE") != 0)
            && (strcmp(inp_p->ctifile, "\0") != 0)
            && (dmDatasetAccess(inp_p->ctifile, "R") == dmTRUE))
        {
          load_cti_map(inp_p, cti_p, err_p);
          applied_cti = 1;
        }
        else
        {
          inp_p->apply_cti = 0;
          dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "WARNING: problem reading ctifile, cti adjustment will not be "
            "applied. Changing apply_cti=yes to apply_cti=no.");
          dmResetErrorState();
        }

        if ((ds_strcmp_cis(inp_p->grade_image_file, "NONE") == 0)
            || (strcmp(inp_p->grade_image_file, "\0") == 0))
        {
          inp_p->apply_cti = 0;
          dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "WARNING: grade_image_file is not specified and "
            "apply_cti=yes, adjustment will not be applied. "
            "Changing apply_cti=yes to apply_cti=no.");
          dmResetErrorState();
        }
      }

      /* if PHAS_ADJ is in the eventdef and apply_cti=no*/
      if ((0 != strcmp(inp_p->eventdef, "")
          && strstr(inp_p->eventdef, "phas_adj")) && (!inp_p->apply_cti))
      {
        dsErrAdd(err_p, dsFINDKEYWORDWERR, Individual, Custom, "WARNING: The column PHAS_ADJ will not be produced because apply_cti=no.");
      }

      if (inp_p->check_badpix)
      {
        if ((badpix_p = (BADPIX_SETUP_P_T)malloc(sizeof(BADPIX_SETUP_T)))
            != NULL)
        {
          memset(badpix_p, 0, sizeof(BADPIX_SETUP_T));
          set_badpix_map_values(inp_p->badpixfile, badpix_p, err_p);
          dmResetErrorState();
        }
        else
        {
          dsErrAdd(err_p, dsALLOCERR, Individual, Custom, "ERROR: Unable to allocate memory for bad pixel check.\n");
          dsErrPrintList(err_p, dsErrTrue);
          return (-1);
        }
      }
      else
      {
        badpix_p = NULL;
      }

      if (inp_p->apply_cti && ((inp_p->in_mode & MODE_GRD_TE) != 0
          || (inp_p->in_mode & MODE_GRD_CC) != 0 || (inp_p->in_mode
          & MODE_GRD_CC3x3) != 0 || (inp_p->in_mode & MODE_GRD_HISTOGRAM) != 0))
      {
        if (acis_load_grade_images(inp_p, cti_p, err_p) != 1)
        {
          inp_p->apply_cti = 0;
          dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "WARNING: Problem with Grade Image File, will not apply cti correction.\n");
          dmResetErrorState();
        }
      }

      acis_process_setup_output_file(evtin_p, evtout_p, inp_p, inst_p, aln_p, err_p);

      /* Check to see if we are dealing with tg_resolve_events data
       * Mode is either CC33_FAINT or CC33_GRADED */
      if ((inp_p->in_mode & CC33_MODES) != 0)
      {
        inp_p->is_cc33_mode = 1;

        if (ds_strcmp_cis(inp_p->content, CONTENT_TGEVT1) == 0)
        {
          if (inp_p->debug > DEBUG_LEVEL_2)
          {
            printf("DEBUG: we are dealing with cc mode on tg resolve events data %s.\n", inp_p->content);
          }
          inp_p->is_tg = 1;
        }

        if (ds_strcmp_cis(inp_p->observation_mode, "pointing") == 0)
        {
          determine_focalpoint(inp_p);
        }/* end if obs_mode is pointing */
      }/* end if in cc3x3 mode */

      if (!initialize_acaofffile(inp_p, err_p))
      {
        dsErrPrintList(err_p, dsErrTrue);
        return (-1);
      }

      if (!initialize_subpixfile(inp_p, &subpix_t, err_p))
      {
        dsErrPrintList(err_p, dsErrTrue);
        return (-1);
      }

      /* setup info for fast window to real chip coord (acis2c) */
      if (inp_p->acis_2c_data)
      {
        get_predicted_beam_position(evtin_p->extension, inp_p);
        if ((inp_p->winsize != WINSIZE_DEFAULT_VAL) && (!inp_p->found_foa))
        {
          dsErrAdd(err_p, dsAPEFOAFPDNEERR, Individual, Generic, inp_p->infile);
        }
      }

      status = load_threshold_file(inp_p, err_p); /* load split threshold values */
      if (status != dsNOERR)
      {
        dsErrPrintList(err_p, dsErrTrue);
        return (-1);
      }

      /* intersect subspace- keeps gti extensions if rerunning */
      dmBlockIntersectSubspace(evtin_p->extension, evtout_p->extension, evtout_p->extension);

      /* warnings for columns that are not in input file or computed */
      acis_process_uninitialized_col(evtin_p, evtout_p, inp_p, err_p);

      process_warnings(err_p, log_ptr, debug);

      setup_outfile = FALSE;

    }
    else
    {
      /* check for mode consistency */
      if ((inp_p->in_mode != inp_p->supported_mode) && (num_evtfile > 1))
      {
        /* error if modes switch during a run */
        dsErrAdd(err_p, dsAFEMIXEDMODEERR, Individual, Generic);
      }

      /* merge subspaces */
      dmBlockMergeSubspace(evtin_p->extension, evtout_p->extension, evtout_p->extension);
    }

    if (dsErrGetFatalCt(err_p) == 0)
    {
      if (debug > DEBUG_LEVEL_3)
      {
        fprintf(log_ptr, "  open file : %s\n", inp_p->infile);

        if (debug > DEBUG_LEVEL_4)
        {
          fprintf(log_ptr, "    TIME        PHA1 PHA2 PHA3 PHA4 PHA5");
          fprintf(log_ptr, " PHA6 PHA7 PHA8 PHA9  SUM DET_X DET_Y\tCHIPX\tCHIPY(_TARG)\n");
        }
      }

      /************************
       * **********************
       * LOOP OVER ALL EVENTS
       * **********************
       ************************/
      /* only loop if event file contains at least 1 row */
      if (dmTableGetRowNo(evtin_p->extension) != dmBADROW && num_rows > 0)
        while ((row_check != dmNOMOREROWS) && (dsErrGetFatalCt(err_p) == 0))
        {
          /* initialize event record structure */
          memset(evt_p, 0, sizeof(EVENT_REC_T));
          evt_p->time = inp_p->start_time;
          evt_p->status = L1_STS_GOOD_EVENT;
          int index = 0;
          for (index = 0; index < ADET_NUM_COLS; index++)
          {
            evt_p->status_cent_valid[index] = TRUE;
          }
          for (index = 0; index < 9; index++)
          {
            evt_p->status_centroid[index] = L1_STS_GOOD_EVENT;
          }

          /* load data into event record */
          load_event_data(evtin_p, evt_p);

          /* The value of STATUS[k] is set to zero for bits
           * k = 1-5, 14, 16-19, and 23 (of 32)
           * bits that can be set by acis_process_events
           * if doevtgrade = yes
           * then the value of STATUS[20] is set to zero
           */
          /*    if(inp_p->doevtgrade)
           {
           evt_p->status &= ~(L1_STAT_TO_ZERO_EVTGRD);
           }
           else
           {
           evt_p->status &= ~(L1_STAT_TO_ZERO_EVTGRD);
           }*/

          /* if applying graded mode cti correction set the center phas value
           * as the pha value
           */
          if (inp_p->apply_cti && ((inp_p->in_mode & MODE_GRD_TE) != 0
              || (inp_p->in_mode & MODE_GRD_CC) != 0 || (inp_p->in_mode
              & MODE_GRD_CC3x3) != 0 || (inp_p->in_mode & MODE_GRD_HISTOGRAM)
              != 0))
          {
            evt_p->phas[4] = evt_p->pha;
            evt_p->phas_adj[4] = evt_p->pha;
            cti_p->flight_grade = evt_p->fltgrade; /* store original value */
          }

          verify_ccdid_and_node(evt_p, inp_p, err_p);

          if (set_and_verify_chip_vals(evt_p, inp_p, err_p))
          {
            return -1;
          }

          if ((inp_p->supported_mode & CC_MODES) != 0)
          {
            read_time_ro(inp_p, evt_p, err_p);
            double temp_time = evt_p->time_ro - (inp_p->chip_targ_med[ADET_Y_AXIS] + 1028) * inp_p->timedel;
            calculate_and_verify_chip_targ(temp_time, evt_p->ccdid, inp_p, err_p);

          /* keep track of earliest and latest event times */
          }

          if (evt_p->time < inp_p->evt_tstart)
          {
            inp_p->evt_tstart = evt_p->time;
          }
          else if (evt_p->time > inp_p->evt_tstop)
          {
            inp_p->evt_tstop = evt_p->time;
          }

          /* if acis-2c data then remap columns to flight format */
          if (inp_p->acis_2c_data)
          {
            convert_acis2c_to_flight(evt_p->phas, inp_p->supported_mode);
          }

          /* add time offset to event time for dither sequence */
          evt_p->time += inp_p->time_offset;

          if ((inp_p->supported_mode & CC_MODES) != 0)
          {
            read_tg_m(inp_p, evt_p, err_p);

            if (!read_chipy_tg(inp_p, evt_p))
            {
              dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "ERROR: While reading chipy_tg.  CHIPY_TG dependent"
                " computations could fail if the value of the CHIPY_TG is unphysical.");
              dsErrPrintList(err_p, dsErrTrue);
              return (-1);
            }

            read_chipy_zo(inp_p, evt_p);
          }

          if (!determine_time_chip_adj(inp_p, evt_p, err_p))
          {
            dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "ERROR: While determining time and chip_adj.  CHIPY_ADJ dependent"
              " computations could fail if the value of the CHIPY_ADJ is unphysical.");
            dsErrPrintList(err_p, dsErrTrue);
            return (-1);
          }

          if (alignment_update(aln_p, aln_hk_p, evt_p->time))
          {
            dsErrAdd(err_p, dsHPEALIGNMENTERR, Individual, Generic);
          }

          /*reset mirror after chipy_adj calc */
          pix_set_mirror(aln_p->hpy, aln_p->pos, aln_p->theta);

          /* restore original time after dither sequence */
          evt_p->time -= inp_p->time_offset;

          if (aspect_update(asp_p, asp_hk_p, evt_p->time))
          {
            dsErrAdd(err_p, dsHPEASPECTERR, Individual, Generic);
          }

          /* update statistical file counts */
          stat_p->total_events_in++;

          if (inp_p->use_pharo_for_pha)
          {
            /* CASE E:  new PHA value is computed using input PHA_RO */
            evt_p->pha = evt_p->pha_ro;
          }

          if (inp_p->pharo_is_zero)
          {
            evt_p->pha_ro = 0;
          }
          else if (!inp_p->copy_pharo_to_output)
          {
            /* CASE:  PHA_RO = Input PHA */
            evt_p->pha_ro = evt_p->pha;
          }
          /* else
           PHA_RO (input) = PHA_RO (output) */

          if (inp_p->check_badpix)
          {
            if (strcmp(inp_p->pix_adj, "CENTROID") == 0)
            {
              update_badpix_map(evt_p, badpix_p, err_p, TRUE);
            }
            else
            {
              update_badpix_map(evt_p, badpix_p, err_p, FALSE);
            }
            check_event_for_badpix(evt_p, inp_p->supported_mode, badpix_p, err_p);
          }

          /************************************************************/
          if (debug >= DEBUG_LEVEL_4)
          {
            printf("DEBUG: Initial Val (PHA %4ld, PHA_RO %4ld)\t[x %g  y %g]\n", evt_p->pha, evt_p->pha_ro, evt_p->chip[0], evt_p->chip[1]);
          }

          /* process grades */
          /* if CASE B:  PHA_RO = summed PHAS, but no CTI and TGAIN */
          if (inp_p->apply_cti && inp_p->pharo_in_eventdef)
          {
            inp_p->apply_cti = 0;
            process_grades_acis(evtout_p, grade_map, cti_p, evt_p, gain_table, inp_p, stat_p, err_p);
            if (debug >= DEBUG_LEVEL_4)
            {
              printf("DEBUG: First Iter: (PHA %4ld, PHA_RO %4ld)\t[x %g  y %g]\n", evt_p->pha, evt_p->pha_ro, evt_p->chip[0], evt_p->chip[1]);
            }

            if (inp_p->pharo_is_zero)
            {
              /* special tgain case */
              evt_p->pha_ro = 0;
            }
            else if (!inp_p->copy_pha_to_output)
            {
              /* override pha_ro with new summed PHAS pha value */
              evt_p->pha_ro = evt_p->pha;
            }
            /* else
             PHA (output) = PHA (input) */

            inp_p->apply_cti = 1;
            process_grades_acis(evtout_p, grade_map, cti_p, evt_p, gain_table, inp_p, stat_p, err_p);

            if (debug >= DEBUG_LEVEL_4)
            {
              printf("DEBUG: Final vals: (PHA %4ld, PHA_RO %4ld)\t[x %g  y %g]\n", evt_p->pha, evt_p->pha_ro, evt_p->chip[0], evt_p->chip[1]);
            }

          }
          else
          {
            process_grades_acis(evtout_p, grade_map, cti_p, evt_p, gain_table, inp_p, stat_p, err_p);
            if (!dmTableOpenColumn(evtin_p->extension, "PHA"))
              evt_p->pha_ro = evt_p->pha;
          }

          if (debug >= DEBUG_LEVEL_4)
          {
            printf("DEBUG: First Iter: (PHA %4ld, PHA_RO %4ld)\t[x %g  y %g]\n", evt_p->pha, evt_p->pha_ro, evt_p->chip[0], evt_p->chip[1]);
          }

          /* apply the tgain adjustment, TGAIN STEP (2)*/
          if (inp_p->do_tgain)
            apply_tgain_adjustment(evt_p, inp_p, tgain_p, err_p);

          if (debug >= DEBUG_LEVEL_4)
          {
            printf("DEBUG: Final vals: (PHA %4ld, PHA_RO %4ld)\t[x %g  y %g]\n", evt_p->pha, evt_p->pha_ro, evt_p->chip[0], evt_p->chip[1]);
          }

          convert_pha_to_pi(inp_p, gain_table, evt_p, err_p);
          calculate_coords_acis(&asp_p->entry[asp_p->next], inp_p, asp_hk_p->asp_file_type, evt_p, badpix_p, &subpix_t);

          if (inp_p->check_vf_pha && inp_p->supported_mode != MODE_VFNT_TE)
          {
            dsErrAdd(err_p, dsGENERICERR, Individual, Custom, "WARNING: The parameter check_vf_pha=yes, but the "
              "DATAMODE key does not indicate a very-faint mode "
              "observation.  Changing to check_vf_pha=no.");
            inp_p->check_vf_pha = 0;
          }

          /* if the VF status bit is set for this event, unset it */
          if (L1_STS_VF_BAD_PIX & evt_p->status)
          {
            evt_p->status &= ~L1_STS_VF_BAD_PIX;
          }

          /* check very faint pixels */
          if (inp_p->check_vf_pha && inp_p->in_mode == MODE_VFNT_TE)
            check_vf_pha(evt_p, inp_p, badpix_p);

          /* if acis-2c data then map columns from flight format
           back to acis-2c format for the output file */
          if (inp_p->acis_2c_data)
          {
            convert_flight_to_acis2c(evt_p->phas, inp_p->supported_mode);
          }

          if (debug > DEBUG_LEVEL_4)
          {
            fprintf(log_ptr, " %9.4f ", evt_p->time);
            fprintf(log_ptr, "%4d %4d %4d %4d %4d %4d %4d %4d %4d %4ld"
              "\t%g\t%g\n", evt_p->phas[0], evt_p->phas[1], evt_p->phas[2], evt_p->phas[3], evt_p->phas[4], evt_p->phas[5], evt_p->phas[6], evt_p->phas[7], evt_p->phas[8], evt_p->pha, evt_p->chip[0], evt_p->chip[1]);
          }

          /* write out data */
          if (write_acis_events(evtout_p, evt_p) == LEV_1_NO_ERROR)
          {
            /* update statistical file counts */
            stat_p->total_events_out++;
          }
          row_check = dmTableNextRow(evtin_p->extension);
        } /* end while (evt_next_row) */
      /* ************************
       * ************************
       * END LOOP OVER ALL EVENTS
       * ************************
       * ************************
       */
    }
    else
    {
      if (process_warnings(err_p, log_ptr, debug))
      {
        stat_p->num_bad_files++;
      }
    } /* else (dsErrGetFatalCt) */

    determine_time_adj(inp_p, evt_p);
    if (inp_p->debug > DEBUG_LEVEL_2)
    {
      printf("DEBUG: TIME_ADJ = %s\n", inp_p->time_adj);
    }
    dmKeyWrite_c(evtout_p->extension, "TIME_ADJ", inp_p->time_adj, NULL, "time adjustment algorithm");

    acis_process_evt_file_cleanup(evtin_p);

    if ((debug > DEBUG_LEVEL_0) && (inp_p->infile != NULL))
    {
      fprintf(log_ptr, "  close file: %s\n", inp_p->infile);
    }
  } /* end while (inp_p->infile) */

  /* deallocate memory fror grade mappings */
  if (grade_map != NULL)
  {
    free(grade_map);
  }

  /* free memory for alignment/aspect files */
  close_alignment_file(aln_hk_p);
  close_aspect_file(asp_hk_p);
  if (asp_hk_p->asp_type[0] != '\0')
  {
    dmKeyWrite_c(evtout_p->extension, ASP_TYPE_KEY, asp_hk_p->asp_type, NULL, NULL);
  }

  /* error if there are no good files in stack */
  if ((stat_p->num_files_in == stat_p->num_bad_files) && (stat_p->num_bad_files
      > 0))
  {
    dsErrAdd(err_p, dsHPEALLBADFILESERR, Individual, Generic);
  }

  cleanup_gain_table(gain_table);

  if (inp_p->apply_cti && applied_cti)
    cleanup_cti_table(cti_p);

  if (inp_p->apply_tgain)
    cleanup_tgain_table(tgain_p);

  /* close output file */
  acis_process_evt_file_cleanup(evtout_p);

  if (inp_p->check_badpix && badpix_p)
  {
    cleanup_bad_pix(badpix_p, err_p);
    free(badpix_p);
  }

  /* clean up memory allocated for pixlib library */
  if (inp_p->pix_init)
  {
    pix_close_pixlib();
    inp_p->pix_init = FALSE;

    if (inp_p->obsHdr)
    {
      /* free any memory used for events extension keywords */
      free(inp_p->obsHdr);
      inp_p->obsHdr = NULL;
    }
  }

  free(inp_p->aspect_data_array);
  free(inp_p->acp);

  /* free up memory allocated for stack */
  stk_close(evtstack);

  /* verify time range */
  if (dsErrGetFatalCt(err_p) == 0)
  {
    acis_process_time_check(inp_p, log_ptr, err_p);
  }

  if (debug > DEBUG_LEVEL_2)
  {
    fprintf(log_ptr, " ============ STATISTICS ===========\n");
    fprintf(log_ptr, "FILES   total in (%4ld) - bad (%2ld) = %3ld used\n", stat_p->num_files_in, stat_p->num_bad_files, stat_p->num_files_in
        - stat_p->num_bad_files);
    fprintf(log_ptr, "EVENTS  total in (%4ld)   total out = %3ld\n", stat_p->total_events_in, stat_p->total_events_out);
  }

  if (err_p->size > 0)
  {
    if (dsErrGetFatalCt(err_p) != 0)
    {
      err = dsGENERICERR;
    }
    dsErrPrintList(err_p, dsErrTrue);

    if ((debug > DEBUG_LEVEL_0) && (log_ptr != stderr) && (log_ptr != stdout))
    {
      dsErrDirectOutput(log_ptr);
      dsErrPrintList(err_p, dsErrTrue);
      dsErrDirectOutput(stderr);
    }

    while (err_p->size > 0)
      dsErrRemove(err_p);
  }

  /* close up log file */
  if ((debug > DEBUG_LEVEL_0) && (log_ptr != NULL))
  {
    fclose(log_ptr);
  }

  return (err);
}
