/*                                                                
**  Copyright (C) 1998,2007,2010,2015  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*************************************************************************

 * FILE NAME: acis_setup_columns.c
 
 * DESCRIPTION:

 The routine acis_setup_columns() is called upon by acis_process_events
 to create the columns of the output event file. It is responsible for
 obtaining the column descriptor as well as for setting the values of
 TLMIN/TLMAX/TUNITS keywords as specified in the level 1 acis to archive
 icd. (Macros defining these values are found in ds_acis_config.h).

 * NOTES

 The routine does not pass a return status back, but detected errors are
 added to the maintained error list which is passed in as a parameter.

11/13/2015- bug14203: change tnull to nan for chipy_zo, chipy_tg ; don't set tnull or nan for tdet;
 *************************************************************************/

#include <ctype.h>
#include <limits.h> 

#ifndef ACIS_PROCESS_EVENTS_H
#include "acis_process_events.h"
#define ACIS_PROCESS_EVENTS_H
#endif

#ifndef DSLIB_H
#include "dslib.h"
#define DSLIB_H
#endif

#ifndef DS_ACIS_CONFIG_H
#include "ds_acis_config.h"
#define DS_ACIS_CONFIG_H
#endif

#ifndef EVENT_DEFS_H
#include "event_defs.h"
#define EVENT_DEFS_H
#endif 

static int acis_write_column_range(dmDescriptor* desc, /* I   - column descriptor            */
dmDataType type, /* I   - column type                  */
double min, /* I   - minimum value (TLMIN)        */
double max) /* I   - maximum value (TLMAX)        */
{
  int retVal = 1;
  switch(type)
  {
    case dmSHORT:
    {
      short smin = (short)min, smax = (short)max;

      if ((min < SHRT_MIN) || (min > SHRT_MAX))
      {
        smin = SHRT_MIN; /* avoid underflow */
      }
      if ((max < SHRT_MIN) || (max > SHRT_MAX))
      {
        smax = SHRT_MAX; /* avoid overflow */
      }

      dmDescriptorSetRange_s(desc, smin, smax);
    }
      break;
    case dmLONG:
      dmDescriptorSetRange_l(desc, (long)min, (long)max);
      break;
    case dmFLOAT:
      dmDescriptorSetRange_f(desc, (float)min, (float)max);
      break;
    case dmDOUBLE:
      dmDescriptorSetRange_d(desc, (double)min, (double)max);
      break;
    default:
      retVal = 0;
      break;
  }
  return retVal;
} /* end: acis_write_column_range */

static int acis_write_column_null(dmDescriptor* desc, /* I   - column descriptor            */
dmDataType type, /* I   - column type                  */
double null /* I - nil value */)
{
  int retVal = 1;
  switch(type)
  {
    case dmSHORT:
    {
      short snull = (short)null;

      if (null < SHRT_MIN)
      {
        snull = SHRT_MIN; /* avoid underflow */
      }

      dmDescriptorSetNull_s(desc, snull);
    }
      break;
    case dmUSHORT:
      dmDescriptorSetNull_us(desc, (unsigned short)null);
      break;
    case dmLONG:
      dmDescriptorSetNull_l(desc, (long)null);
      break;
    case dmFLOAT:
      dmDescriptorSetNull_f(desc, (float)null);
      break;
    case dmULONG:
      dmDescriptorSetNull_ul(desc, (unsigned long)null);
      break;
    case dmDOUBLE:
      dmDescriptorSetNull_d(desc, (double)null);
      break;
    default:
      retVal = 0;
      break;
  }
  return retVal;
} /* end: acis_write_column_null */

static int acis_write_column_bin(dmDescriptor* desc, /* I   - column descriptor            */
dmDataType type, /* I   - column type                  */
double bin /* I - bin value */)
{
  int retVal = 1;
  switch(type)
  {
    case dmSHORT:
      dmDescriptorSetBin_s(desc, (short)bin);
      break;
    case dmUSHORT:
      dmDescriptorSetBin_us(desc, (unsigned short)bin);
      break;
    case dmLONG:
      dmDescriptorSetBin_l(desc, (long)bin);
      break;
    case dmULONG:
      dmDescriptorSetBin_ul(desc, (unsigned long)bin);
      break;
    case dmDOUBLE:
      dmDescriptorSetBin_d(desc, (double)bin);
      break;
    case dmFLOAT:
      dmDescriptorSetBin_f(desc, (float)bin);
      break;
    default:
      retVal = 0;
      break;
  }
  return retVal;
} /* end: acis_write_column_bin */

void acis_setup_columns(EVENT_SETUP_P_T evtin_p, /* I   - input file information       */
EVENT_SETUP_P_T evtout_p, /* I/O - output file inprocession     */
INPUT_PARMS_P_T inp_p, /* I/O - input parameters/data        */
Header_Type* obsHdr, /* I   - observation.par data         */
char** out_names, /* I   - output event column names    */
dsErrList* hpe_err_p) /* O   - error list pointer           */

{
  short rr;

  for (rr = 0; rr < evtout_p->num_cols; rr++)
  {
    switch(evtout_p->mapping[rr])
    {
      case EVT_SKY: /* vectored column pair- X,Y */
      {
        VEC2_LONG axlen;
        char names[2][DS_SZ_KEYWORD];
        char* cptnames[2];
        char* eqnames[] = {"RA", "DEC"};
        dmDescriptor* temp_desc;

        int ii = 0;
        long dim = 0;
        double* crpix_g;
        double* cdelt_g;
        double* dummy;

        cptnames[0] = names[0];
        cptnames[1] = names[1];
        strcpy(cptnames[0], ADET_SKY_X_COL);
        strcpy(cptnames[1], ADET_SKY_Y_COL);

        /* if output column name is lowercase, set x/y to lc */
        if (islower(out_names[rr][0]))
        {
          string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
          string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
        }

        pix_get_fp_image_cntr(inp_p->crpix);
        hdrGetKeyValue_d(obsHdr, "RA_NOM", &inp_p->crval[0]);
        hdrGetKeyValue_d(obsHdr, "DEC_NOM", &inp_p->crval[1]);
        inp_p->crval[2] = 0.0;
        inp_p->cdelt[0] = inp_p->fp_scale / -3600.0;
        inp_p->cdelt[1] = inp_p->cdelt[0] * -1;

        /* if stop_transform!=sky, then get crval,cdelt,crpix from input */
        if (inp_p->stop_transform != ADET_SKY_VAL)
        {
          dim = (long)dmGetElementDim(dmCoordGetParent(inp_p->eqposcol));
          dummy = (double*)malloc(dim * sizeof(double));
          cdelt_g = (double*)malloc(dim * sizeof(double));
          crpix_g = (double*)malloc(dim * sizeof(double));

          dmCoordGetTransform_d(inp_p->eqposcol, crpix_g, dummy, cdelt_g, dim);
          for (ii = 0; ii < dim; ii++)
          {
            inp_p->crpix[ii] = crpix_g[ii];
            inp_p->cdelt[ii] = cdelt_g[ii];
          }

          free(dummy);
          free(cdelt_g);
          free(crpix_g);
        }

        evtout_p->dim[rr] = 2;
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_SKY_TUNITS, "sky coordinates", cptnames, evtout_p->dim[rr]);
        inp_p->sky_desc
            = dmCoordCreate_d(evtout_p->desc[rr], "EQPOS", "deg", eqnames, 2, "TAN", inp_p->crpix, inp_p->crval, inp_p->cdelt, NULL);

        /* set TLMIN/TLMAX for sky coords */
        pix_get_fp_image_dimn(axlen);
        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
        {
          if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_X_TLMIN, DS_ACIS_X_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }

        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
        {
          if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_Y_TLMIN, DS_ACIS_Y_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }
      }
        break;

      case EVT_CHIP: /* vectored coord pair- CHIPX/CHIPY */
      {
        double crpix[2] = {0.5, 0.5}, crval[2] = {0.0, 0.0}, cdelt[2];
        char names[2][DS_SZ_KEYWORD];
        char* cpcnames[] = {"CPCX", "CPCY"};
        char* cptnames[2];
        dmDescriptor* temp_desc;

        int ii = 0;
        double* crpix_g;
        double* cdelt_g;
        double* dummy;
        long dim;

        cptnames[0] = names[0];
        cptnames[1] = names[1];
        strcpy(cptnames[0], ADET_CHIP_X_COL);
        strcpy(cptnames[1], ADET_CHIP_Y_COL);

        /* if output column name is lowercase, set x/y to lc */
        if (islower(out_names[rr][0]))
        {
          string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
          string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
        }

        evtout_p->dim[rr] = 2;
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "Chip coords", cptnames, evtout_p->dim[rr]);

        /* if stop_transform=none, get cdelt and crpix from input */
        if (inp_p->stop_transform == ADET_NONE_VAL)
        {
          dim = (long)dmGetElementDim(dmCoordGetParent(inp_p->cpccol));
          dummy = (double*)malloc(dim * sizeof(double));
          cdelt_g = (double*)malloc(dim * sizeof(double));
          crpix_g = (double*)malloc(dim * sizeof(double));

          dmCoordGetTransform_d(inp_p->cpccol, crpix_g, dummy, cdelt_g, dim);
          for (ii = 0; ii < dim; ii++)
          {
            crpix[ii] = crpix_g[ii];
            cdelt[ii] = cdelt_g[ii];
          }

          free(dummy);
          free(cdelt_g);
          free(crpix_g);
        }
        else
        {
          pix_get_chip_scale_in_mm(cdelt);
        }

        dmCoordCreate_d(evtout_p->desc[rr], "CPC", "mm", cpcnames, 2, "LINEAR", crpix, crval, cdelt, NULL);

        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          /* CC mode */
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
          {
            dmDescriptorSetRange_s(temp_desc, DS_ACIS_CC_CHIPX_TLMIN, DS_ACIS_CC_CHIPX_TLMAX);
          }
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
          {
            dmDescriptorSetRange_s(temp_desc, DS_ACIS_CC_CHIPY_TLMIN, DS_ACIS_CC_CHIPY_TLMAX);
          }
        }
        else
        {
          /* TE modes */
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
          {
            dmDescriptorSetRange_s(temp_desc, DS_ACIS_TE_CHIPX_TLMIN, DS_ACIS_TE_CHIPX_TLMAX);
          }
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
          {
            dmDescriptorSetRange_s(temp_desc, DS_ACIS_TE_CHIPY_TLMIN, DS_ACIS_TE_CHIPY_TLMAX);
          }
        }
      }
        break;

      case EVT_TDET: /* vectored coord pair- TDETX/TDETY */
      {
        char names[2][DS_SZ_KEYWORD];
        char* cptnames[2];
        dmDescriptor* temp_desc;

        cptnames[0] = names[0];
        cptnames[1] = names[1];
        strcpy(cptnames[0], ADET_TDET_X_COL);
        strcpy(cptnames[1], ADET_TDET_Y_COL);

        /* if output column name is lowercase, set x/y to lc */
        if (islower(out_names[rr][0]))
        {
          string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
          string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
        }

        evtout_p->dim[rr] = 2;
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_TDET_TUNITS, "ACIS tiled detector coordinates", cptnames, evtout_p->dim[rr]);

        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          /* CC mode */
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
          {
            if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_CC_TDETX_TLMIN, DS_ACIS_CC_TDETX_TLMAX))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
            if (!acis_write_column_null(temp_desc, evtout_p->types[rr], 9999))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
          }
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
          {
            if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_CC_TDETY_TLMIN, DS_ACIS_CC_TDETY_TLMAX))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
            if (!acis_write_column_null(temp_desc, evtout_p->types[rr], 9999))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
          }
        }
        else
        {
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
          {
            if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_TE_TDETY_TLMIN, DS_ACIS_TE_TDETY_TLMAX))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
            if (!acis_write_column_null(temp_desc, evtout_p->types[rr], 9999))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
          }
          if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
          {
            if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_TE_TDETY_TLMIN, DS_ACIS_TE_TDETY_TLMAX))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
            if (!acis_write_column_null(temp_desc, evtout_p->types[rr], 9999))
            {
              dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
            }
          }
        }
      }
        break;

      case EVT_DET: /* vectored coord pair- DETX/DETY */
      {
        double crpix[2], crval[2] = {0.0, 0.0}, cdelt[2];
        double tanp_params[4] = {0.0, /* Rotation (deg)         */
        270.0, /* pole longitude (deg)   */
        90.0, /* pole latitude (deg)    */
        2000.0}; /* equinox (Julian years) */
        VEC2_LONG axlen;
        char names[2][DS_SZ_KEYWORD];
        char* cptnames[2];
        char* mscnames[] = {"PHI", "THETA"};
        dmDescriptor* temp_desc;

        int ii = 0;
        double* crpix_g;
        double* cdelt_g;
        double* dummy;
        long dim;

        cptnames[0] = names[0];
        cptnames[1] = names[1];
        strcpy(cptnames[0], ADET_DET_X_COL);
        strcpy(cptnames[1], ADET_DET_Y_COL);

        /* if output column name is lowercase, set x/y to lc */
        if (islower(out_names[rr][0]))
        {
          string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
          string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
        }

        evtout_p->dim[rr] = 2;
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_DET_TUNITS, "ACIS detector coordinates", cptnames, evtout_p->dim[rr]);

        /* if stop_transform=none or stop_transform=chip, get cdelt and crpix from input */
        if (inp_p->stop_transform == ADET_NONE_VAL || inp_p->stop_transform == ADET_CHIP_VAL)
        {
          dim = (long)dmGetElementDim(dmCoordGetParent(inp_p->msccol));
          dummy = (double*)malloc(dim * sizeof(double));
          cdelt_g = (double*)malloc(dim * sizeof(double));
          crpix_g = (double*)malloc(dim * sizeof(double));

          dmCoordGetTransform_d(inp_p->msccol, crpix_g, dummy, cdelt_g, dim);
          for (ii = 0; ii < dim; ii++)
          {
            crpix[ii] = crpix_g[ii];
            cdelt[ii] = cdelt_g[ii];
          }

          free(dummy);
          free(cdelt_g);
          free(crpix_g);
        }
        else
        {
          pix_get_fp_image_cntr(crpix);
          cdelt[0] = cdelt[1] = inp_p->fp_scale / 3600.0;
        }

        dmCoordCreate_d(evtout_p->desc[rr], "MSC", "deg", mscnames, 2, "TAN-P", crpix, crval, cdelt, tanp_params);

        pix_get_fp_image_dimn(axlen);

        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
        {
          if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_DETX_TLMIN, DS_ACIS_DETX_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }

        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
        {
          if (!acis_write_column_range(temp_desc, evtout_p->types[rr], DS_ACIS_DETY_TLMIN, DS_ACIS_DETY_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }
      }
        break;

      case EVT_TIME:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_TIME_TUNITS, "S/C TT corresponding to mid-exposure");
        break;

      case EVT_TIME_RO:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_TIME_TUNITS, "S/C TT corresponding to mid-exposure");
        break;

      case EVT_PHAS:
      {
        long pha_axes[2];

        switch(inp_p->supported_mode)
        {
          case MODE_GRD_CC3x3: /* fallthrough intended */
          case MODE_GRD_CC: /* fallthrough intended */
          case MODE_GRD_TE:
            pha_axes[0] = 0;
            pha_axes[1] = 0;
            break;
          case MODE_FNT_CC: /* fallthrough intended */
          case MODE_RAW_CC:
            pha_axes[0] = 3;
            pha_axes[1] = 1;
            break;
          case MODE_VFNT_TE:
            pha_axes[0] = 5;
            pha_axes[1] = 5;
            break;
          case MODE_FNT_CC3x3: /* fallthrough intended */
          case MODE_FNT_TE: /* fallthrough intended */
          case MODE_FNT_BS: /* fallthrough intended */
          case MODE_FNT_BSMP: /* fallthrough intended */
          default:
            pha_axes[0] = 3;
            pha_axes[1] = 3;
            break;
        }

        /* only set column for arrays with non-zero axes */
        if (pha_axes[0] && pha_axes[1])
        {
          evtout_p->desc[rr]
              = dmColumnCreateNDArray(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_PHAS_TUNITS, "array of pixel pulse heights", pha_axes, 2);

          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_PHAS_TLMIN, DS_ACIS_PHAS_TLMAX);
        }
      }
        break;

      case EVT_ADJPHAS:
        if (inp_p->apply_cti)
        {
          long pha_adj_axes[2];

          switch(inp_p->supported_mode)
          {
            case MODE_GRD_CC3x3: /* fallthrough intended */
            case MODE_GRD_CC: /* fallthrough intended */
            case MODE_GRD_TE:
              pha_adj_axes[0] = 0;
              pha_adj_axes[1] = 0;
              break;
            case MODE_FNT_CC: /* fallthrough intended */
            case MODE_RAW_CC:
              pha_adj_axes[0] = 3;
              pha_adj_axes[1] = 1;
              break;
            case MODE_VFNT_TE:
              pha_adj_axes[0] = 5;
              pha_adj_axes[1] = 5;
              break;
            case MODE_FNT_CC3x3: /* fallthrough intended */
            case MODE_FNT_TE: /* fallthrough intended */
            case MODE_FNT_BS: /* fallthrough intended */
            case MODE_FNT_BSMP: /* fallthrough intended */
            default:
              pha_adj_axes[0] = 3;
              pha_adj_axes[1] = 3;
              break;
          }

          /* only set column for arrays with non-zero axes */
          if (pha_adj_axes[0] && pha_adj_axes[1])
          {
            evtout_p->desc[rr]
                = dmColumnCreateNDArray(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_PHAS_TUNITS, "array of cti adjusted pixel pulse heights", pha_adj_axes, 2);

            dmDescriptorSetRange_f(evtout_p->desc[rr], DS_ACIS_PHAS_TLMIN, DS_ACIS_PHAS_TLMAX);
          }
        }
        break;

      case EVT_PHA:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_PHA_TUNITS, "total pulse height of event");
        switch(inp_p->supported_mode)
        {
          case MODE_FNT_CC: /* fallthrough intended*/
          case MODE_GRD_CC: /* fallthrough intended*/
          case MODE_RAW_CC:
            dmDescriptorSetRange_l(evtout_p->desc[rr], DS_ACIS_CC_PHA_TLMIN, DS_ACIS_CC_PHA_TLMAX);
            break;
          case MODE_FNT_CC3x3: /* fallthrough intended*/
          case MODE_GRD_CC3x3: /* fallthrough intended*/
          case MODE_FNT_TE: /* fallthrough intended*/
          case MODE_FNT_BS: /* fallthrough intended*/
          case MODE_VFNT_TE: /* fallthrough intended*/
          case MODE_FNT_BSMP: /* fallthrough intended*/
          default:
            dmDescriptorSetRange_l(evtout_p->desc[rr], DS_ACIS_TE_PHA_TLMIN, DS_ACIS_TE_PHA_TLMAX);
            break;
        }
        dmDescriptorSetNull_l(evtout_p->desc[rr], 0);
        break;

      case EVT_PHA_RO:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_PHA_TUNITS, "total read-out pulse height of event");
        switch(inp_p->supported_mode)
        {
          case MODE_FNT_CC: /* fallthrough intended*/
          case MODE_GRD_CC: /* fallthrough intended*/
          case MODE_RAW_CC:
            dmDescriptorSetRange_l(evtout_p->desc[rr], DS_ACIS_CC_PHA_TLMIN, DS_ACIS_CC_PHA_TLMAX);
            break;
          case MODE_FNT_CC3x3: /* fallthrough intended*/
          case MODE_GRD_CC3x3: /* fallthrough intended*/
          case MODE_FNT_TE: /* fallthrough intended*/
          case MODE_FNT_BS: /* fallthrough intended*/
          case MODE_VFNT_TE: /* fallthrough intended*/
          case MODE_FNT_BSMP: /* fallthrough intended*/
          default:
            dmDescriptorSetRange_l(evtout_p->desc[rr], DS_ACIS_TE_PHA_TLMIN, DS_ACIS_TE_PHA_TLMAX);
            break;
        }
        dmDescriptorSetNull_l(evtout_p->desc[rr], 0);
        break;

      case EVT_PULSE_INV:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_PI_TUNITS, "pulse invariant energy of event");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], 1, inp_p->pi_num_bins);
        acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 0);
        break;

      case EVT_ENERGY:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_ENERGY_TUNITS, "nominal energy of event (eV)");
        switch(inp_p->supported_mode)
        {
          case MODE_FNT_CC: /* fallthrough intended*/
          case MODE_GRD_CC: /* fallthrough intended*/
          case MODE_RAW_CC:
            dmDescriptorSetRange_f(evtout_p->desc[rr], DS_ACIS_CC_ENERGY_TLMIN, DS_ACIS_CC_ENERGY_TLMAX);
            break;
          case MODE_FNT_CC3x3: /* fallthrough intended*/
          case MODE_GRD_CC3x3: /* fallthrough intended*/
          case MODE_FNT_TE: /* fallthrough intended*/
          case MODE_FNT_BS: /* fallthrough intended*/
          case MODE_VFNT_TE: /* fallthrough intended*/
          case MODE_FNT_BSMP: /* fallthrough intended*/
          default:
            dmDescriptorSetRange_f(evtout_p->desc[rr], DS_ACIS_TE_ENERGY_TLMIN, DS_ACIS_TE_ENERGY_TLMAX);
            break;
        }
        break;

      case EVT_CHIPX:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "X position of center pixel of event, chip coords");
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CC_CHIPX_TLMIN, DS_ACIS_CC_CHIPX_TLMAX);
        }
        else
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_TE_CHIPX_TLMIN, DS_ACIS_TE_CHIPX_TLMAX);
        }
        break;

      case EVT_CHIPY:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "Y position of center pixel of event, chip coords");
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CC_CHIPY_TLMIN, DS_ACIS_CC_CHIPY_TLMAX);
        }
        else
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_TE_CHIPY_TLMIN, DS_ACIS_TE_CHIPY_TLMAX);
        }
        break;
      case EVT_CHIPX_ADJ:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "modified X position of center pixel of event, chip coords");
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CC_CHIPX_TLMIN, DS_ACIS_CC_CHIPX_TLMAX);
        }
        else
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_TE_CHIPX_TLMIN, DS_ACIS_TE_CHIPX_TLMAX);
        }
        break;

      case EVT_CHIPY_ADJ:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "modified Y position of center pixel of event, chip coords");
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CC_CHIPY_TLMIN, DS_ACIS_CC_CHIPY_TLMAX);
        }
        else
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_TE_CHIPY_TLMIN, DS_ACIS_TE_CHIPY_TLMAX);
        }
        break;

      case EVT_CHIP_ADJ:
      {
        char *colnames[2] = {"chipx_adj", "chipy_adj"};
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "modified position of center pixel of event, chip coords", colnames, 2);
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CC_CHIPY_TLMIN, DS_ACIS_CC_CHIPY_TLMAX);
        }
        else
        {
          dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_TE_CHIPY_TLMIN, DS_ACIS_TE_CHIPY_TLMAX);
        }

      }
        break;

      case EVT_TDET_X:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_TDET_TUNITS, "X position of event, ACIS tiled detector coordinates");
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_CC_TDETX_TLMIN, DS_ACIS_CC_TDETX_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
          if (!acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 9999))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }
        else
        {
          /* TE MODES */
          if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_TE_TDETX_TLMIN, DS_ACIS_TE_TDETX_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
          if (!acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 9999))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }
        break;

      case EVT_TDET_Y:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_TDET_TUNITS, "Y position of event, ACIS tiled detector coordinates");
        if ((inp_p->supported_mode & CC_MODES) != 0)
        {
          if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_CC_TDETY_TLMIN, DS_ACIS_CC_TDETY_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
          if (!acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 9999))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }
        else
        {
          /* TE MODES */
          if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_TE_TDETY_TLMIN, DS_ACIS_TE_TDETY_TLMAX))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
          if (!acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 9999))
          {
            dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
          }
        }
        break;

      case EVT_DET_X:
      {
        VEC2_LONG axlen;

        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_DET_TUNITS, "X position of event, ACIS detector coordinates");
        pix_get_fp_image_dimn(axlen);
        if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_DETX_TLMIN, DS_ACIS_DETX_TLMAX))
        {
          dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
        }
      }
        break;

      case EVT_DET_Y:
      {
        VEC2_LONG axlen;

        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_DET_TUNITS, "Y position of event, ACIS detector coordinates");
        pix_get_fp_image_dimn(axlen);
        if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_DETY_TLMIN, DS_ACIS_DETY_TLMAX))
        {
          dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
        }
      }
        break;

      case EVT_SKY_X:
      {
        VEC2_LONG axlen;

        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_SKY_TUNITS, "X position of event, sky coordinates");
        pix_get_fp_image_dimn(axlen);
        if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_X_TLMIN, DS_ACIS_X_TLMAX))
        {
          dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
        }
      }
        break;

      case EVT_SKY_Y:
      {
        VEC2_LONG axlen;

        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_SKY_TUNITS, "Y position of event, sky coordinates");
        pix_get_fp_image_dimn(axlen);
        if (!acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_ACIS_Y_TLMIN, DS_ACIS_Y_TLMAX))
        {
          dsErrAdd(hpe_err_p, dsAPEDMDATATYPEERR, Individual, Generic, out_names[rr]);
        }
      }
        break;

      case EVT_SKY_1D:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "arcsec", "The 1D spatial coordinate");
        break;

      case EVT_EXPNO:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "Exposure number of CCD frame containing event");
        dmDescriptorSetRange_l(evtout_p->desc[rr], DS_ACIS_EXPNO_TLMIN, DS_ACIS_EXPNO_TLMAX);
        break;

      case EVT_CCDID:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "CCD reporting event");
        dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CCD_ID_TLMIN, DS_ACIS_CCD_ID_TLMAX);
        break;
      case EVT_CCDNODE:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "CCD serial readout amplifier node");
        dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_NODE_ID_TLMIN, DS_ACIS_NODE_ID_TLMAX);
        break;

      case EVT_PHA_MEAN:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "mean of event corner pixel PHA");
        dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CORN_PHA_TLMIN, DS_ACIS_CORN_PHA_TLMAX);
        break;

      case EVT_GRADE:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "binned event grade");
        dmDescriptorSetRange_s(evtout_p->desc[rr], inp_p->grade_min, inp_p->grade_max);
        break;

      case EVT_FLTGRADE:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "event grade, flight system");
        switch(inp_p->supported_mode)
        {
          case MODE_FNT_CC: /* fallthrough intended*/
          case MODE_GRD_CC: /* fallthrough intended*/
          case MODE_RAW_CC:
            dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_CC_FLTGRADE_TLMIN, DS_ACIS_CC_FLTGRADE_TLMAX);
            break;
          case MODE_FNT_CC3x3: /* fallthrough intended*/
          case MODE_GRD_CC3x3: /* fallthrough intended*/
          case MODE_FNT_TE: /* fallthrough intended*/
          case MODE_FNT_BS: /* fallthrough intended*/
          case MODE_VFNT_TE: /* fallthrough intended*/
          case MODE_FNT_BSMP: /* fallthrough intended*/
          default:
            dmDescriptorSetRange_s(evtout_p->desc[rr], DS_ACIS_TE_FLTGRADE_TLMIN, DS_ACIS_TE_FLTGRADE_TLMAX);
            break;
        }
        break;

      case EVT_STATUS:
        if (evtout_p->types[rr] == dmBIT)
        {
          evtout_p->desc[rr]
              = dmColumnCreateArray(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "event status bits", 4);
        }
        else
        {
          evtout_p->desc[rr]
              = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "coded", "");
        }
        break;

      case TG_GAC: /* column rd(tg_r,tg_d) */
      {
        char names[2][DS_SZ_KEYWORD];
        char* cptnames[2];
        dmDescriptor* temp_desc;

        cptnames[0] = names[0];
        cptnames[1] = names[1];
        strcpy(cptnames[0], TG_R_COL);
        strcpy(cptnames[1], TG_D_COL);

        /* if output column name is lowercase, set x/y to lc */
        if (islower(out_names[rr][0]))
        {
          string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
          string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
        }

        evtout_p->dim[rr] = 2;
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_TG_RD_TUNITS, "Grating angular coords", cptnames, evtout_p->dim[rr]);

        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
        {
          dmDescriptorSetRange_d(temp_desc, DS_GRT_TG_R_TLMIN, DS_GRT_TG_R_TLMAX);
          acis_write_column_bin(temp_desc, evtout_p->types[rr], DS_GRT_TG_R_TDBIN);
        }
        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
        {
          dmDescriptorSetRange_d(temp_desc, DS_GRT_TG_D_TLMIN, DS_GRT_TG_D_TLMAX);
          acis_write_column_bin(temp_desc, evtout_p->types[rr], DS_GRT_TG_D_TDBIN);
        }
      }
        break; /* ------------ end: TG_GAC ------------*/

      case TG_GDP: /* column gdp(gdpx,gdpy) */
      {
        double scale, crpix[2], crval[2] = {0.0, 0.0}, cdelt[2];
        char names[2][DS_SZ_KEYWORD];
        char* cptnames[2];
        char* gacnames[] = {"TG_R", "TG_D"};
        dmDescriptor* temp_desc;

        cptnames[0] = names[0];
        cptnames[1] = names[1];
        strcpy(cptnames[0], TG_GDP_X_COL);
        strcpy(cptnames[1], TG_GDP_Y_COL);

        /* if output column name is lowercase, set x/y to lc */
        if (islower(out_names[rr][0]))
        {
          string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
          string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
        }

        evtout_p->dim[rr] = 2;
        evtout_p->desc[rr]
            = dmColumnCreateVector(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_GDP_TUNITS, "Grating pixel coords", cptnames, evtout_p->dim[rr]);

        pix_get_gdp_scale_in_asec(&scale);
        cdelt[0] = scale / -3600.0;
        cdelt[1] = cdelt[0] * -1;
        pix_get_gdp_image_cntr(crpix);
        /* needs to be TAN-P but currently not supported by DM */
        dmCoordCreate_d(evtout_p->desc[rr], "GAC", "deg", gacnames, 2, "LINEAR", crpix, crval, cdelt, NULL);

        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
        {
          acis_write_column_range(temp_desc, evtout_p->types[rr], DS_GRT_GDPX_TLMIN, DS_GRT_GDPX_TLMAX);
          acis_write_column_bin(temp_desc, evtout_p->types[rr], DS_GRT_GDPX_TDBIN);
        }
        if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
        {
          acis_write_column_range(temp_desc, evtout_p->types[rr], DS_GRT_GDPY_TLMIN, DS_GRT_GDPY_TLMAX);
          acis_write_column_bin(temp_desc, evtout_p->types[rr], DS_GRT_GDPY_TDBIN);
        }
        acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 0);
      }
        break; /* ------------ end: TG_GDP ------------*/

      case TG_R: /* column tg_r */
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_TG_RD_TUNITS, "Diffraction angle, relative to zero-order centroid");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_R_TLMIN, DS_GRT_TG_R_TLMAX);
        acis_write_column_bin(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_R_TDBIN);
        break; /* ------------ end: TG_GAC_R ------------*/

      case TG_D: /* column tg_d */
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_TG_RD_TUNITS, "Cross-dispersion angle");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_D_TLMIN, DS_GRT_TG_D_TLMAX);
        acis_write_column_bin(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_D_TDBIN);
        break; /* ------------ end: TG_GAC_D ------------*/

      case TG_GDP_X: /* column gdpx */
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_GDP_TUNITS, "Grating diffraction pixel, dispersion direction");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_GDPX_TLMIN, DS_GRT_GDPX_TLMAX);
        acis_write_column_bin(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_GDPX_TDBIN);
        acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 0);
        break; /* ------------ end: TG_GDP_X ------------*/

      case TG_GDP_Y: /* column gdpy */
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_GDP_TUNITS, "Grating diffraction pixel, cross-dispersion direction");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_GDPY_TLMIN, DS_GRT_GDPY_TLMAX);
        acis_write_column_bin(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_GDPY_TDBIN);
        acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 0);
        break; /* ------------ end: TG_GDP_Y ------------*/

      case TG_PART:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "component index (HEG, MEG, LEG, HESF regions)");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_PART_TLMIN, DS_GRT_TG_PART_TLMAX);
        break; /* ------------ end: TG_PART ------------*/

      case TG_M:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "Diffraction order (m)");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_M_TLMIN, DS_GRT_TG_M_TLMAX);
        acis_write_column_null(evtout_p->desc[rr], evtout_p->types[rr], 99);
        break; /* ------------ end: TG_M ------------*/

      case TG_MLAM:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_TG_MLAM_TUNITS, "Order times wavelength (m * lambda)");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_MLAM_TLMIN, DS_GRT_TG_MLAM_TLMAX);
        acis_write_column_bin(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_MLAM_TDBIN);
        break; /* ------------ end: TG_MLAM ------------*/

      case TG_LAM:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_GRT_TG_LAM_TUNITS, "wavelength (lambda)");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_LAM_TLMIN, DS_GRT_TG_LAM_TLMAX);
        acis_write_column_bin(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_LAM_TDBIN);
        break; /* ------------ end: TG_LAM ------------*/

      case TG_SRCID:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "source ID, index from detect table");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_SRCID_TLMIN, DS_GRT_TG_SRCID_TLMAX);
        break; /* ------------ end: SRCID ------------*/

      case TG_SMAP:
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, "", "source map; flags for up to 10 sources");
        acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], DS_GRT_TG_SMAP_TLMIN, DS_GRT_TG_SMAP_TLMAX);
        break; /* ------------ end: TG_SMAP ------------*/
      case TG_CHIPY_TG:
      {
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "chip y column");
        /*acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], (double)TG_ACIS_MIN_CHIP_VAL, (double)TG_ACIS_MAX_CHIP_VAL);*/

      }/* end: case tg_chipy_tg */
        break; /* ------------ end: TG_CHIPY_TG ----------*/

      case TG_CHIPY_ZO:
      {
        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, DS_ACIS_CHIP_TUNITS, "chip y column");
        /*acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], (double)TG_ACIS_MIN_CHIP_VAL, (double)TG_ACIS_MAX_CHIP_VAL);*/

      }
        break; /* ------------ end: TG_CHIPY_ZO ----------*/
      default:
      {
        char unit[100], comm[100];
        memset(unit, 0, 100);
        memset(comm, 0, 100);

        /* check to see if the input column has the unit and comment/description
         * to use in the output column
         */
        short int ss = 0;
        for (ss = evtin_p->num_cols; ss >= 0; ss--)
        {
          if (evtout_p->mapping[rr] == evtin_p->mapping[ss])
          {
            dmGetUnit(evtin_p->desc[ss], unit, 100);
            dmGetDesc(evtin_p->desc[ss], comm, 100);
            ss = 0;
          }
        }

        evtout_p->desc[rr]
            = dmColumnCreate(evtout_p->extension, out_names[rr], evtout_p->types[rr], 0, unit, comm);

        /* set up any ranges */
        short int cc = 0;
        for (cc = evtin_p->num_cols; cc >= 0; cc--)
        {
          if (evtout_p->mapping[rr] == evtin_p->mapping[cc])
          {
            switch(evtin_p->types[cc])
            {
              case dmSHORT:
              {
                short min = 0, max = 0, nil = 0;

                if (dmDescriptorGetRange_s(evtin_p->desc[cc], &min, &max)
                    != dmFAILURE)
                {
                  acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], (double)min, (double)max);
                }
                if (dmDescriptorGetNull_s(evtin_p->desc[cc], &nil) != dmFAILURE)
                {
                  dmDescriptorSetNull_s(evtout_p->desc[rr], nil);
                }
                else
                {
                  dmDescriptorSetNull_s(evtout_p->desc[rr], 9999);
                }
              }
                break;
              case dmLONG:
              {
                long min = 0, max = 0, nil = 0;

                if (dmDescriptorGetRange_l(evtin_p->desc[cc], &min, &max)
                    != dmFAILURE)
                {
                  acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], (double)min, (double)max);
                }
                if (dmDescriptorGetNull_l(evtin_p->desc[cc], &nil) != dmFAILURE)
                {
                  dmDescriptorSetNull_l(evtout_p->desc[rr], nil);
                }
                else
                {
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 9999);
                }
              }
                break;
              case dmFLOAT:
              {
                float min = 0.0, max = 0.0, nil = 0.0;

                if (dmDescriptorGetRange_f(evtin_p->desc[cc], &min, &max)
                    != dmFAILURE)
                {
                  acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], (double)min, (double)max);
                }
                if (dmDescriptorGetNull_f(evtin_p->desc[cc], &nil) != dmFAILURE)
                {
                  dmDescriptorSetNull_f(evtout_p->desc[rr], nil);
                }
                else
                {
                  dmDescriptorSetNull_f(evtout_p->desc[rr], 9999.0);
                }
              }
                break;
              case dmDOUBLE:
              {
                double min = 0.0, max = 0.0, nil = 0.0;

                if (dmDescriptorGetRange_d(evtin_p->desc[cc], &min, &max)
                    != dmFAILURE)
                {
                  acis_write_column_range(evtout_p->desc[rr], evtout_p->types[rr], min, max);
                }
                if (dmDescriptorGetNull_d(evtin_p->desc[cc], &nil) != dmFAILURE)
                {
                  dmDescriptorSetNull_d(evtout_p->desc[rr], nil);
                }
                else
                {
                  dmDescriptorSetNull_d(evtout_p->desc[rr], 9999.0);
                }
              }
                break;
                /* unknown type, don't set the ranges */
              default:
                break;
            }/*end switch on types*/
            cc = 0; /* don't need to keep looking */
          }
        }/*end loop over num cols */
      }
        break;
    }/*end switch on the mapping column */
  }/*end loop over num cols */
}
