/*                                                                
**  Copyright (C) 2014-2015  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*H**************************************************************************

 * FILE NAME: ape_ccmode.c

 * DEVELOPEMENT: tools

 * DESCRIPTION: Routines used for ccd mode gratings

 * NOTES:

 * REVISION HISTORY:

 Ref. No.        Date
 --------        ----

11/13/2015- bug14203: change tnull to nan for chipy_zo, chipy_tg ; don't set tnull or nan for tdet;
 *H**************************************************************************/

#ifndef ACIS_PROCESS_EVENTS_H
#include "acis_process_events.h"
#endif

#include <math.h>

int compare(const void *val1, const void *val2);
int compute_chip_targ_median(INPUT_PARMS_P_T inp_p, dsErrList* err_p);
void retrieve_coord(INPUT_PARMS_P_T inp_p, double time);

void determine_focalpoint(INPUT_PARMS_P_T inp_p /* I - input param data structure     */)
{
  if (inp_p->debug > DEBUG_LEVEL_2)
  {
    printf("DEBUG: Determining the focal point ccd_id\n");
  }

  static double crpix[2] = {4096.5, 4096.5};
  static double chp_targ[2] = {512.0, 512.0};
  static double fpc_targ[2];
  static double crdelt[2] = {(-0.492 / 3600.0), (0.492 / 3600.0)};
  short chipid;
  static double targ_crds[2];
  double pnts[3];

  dmBlock *in_block = dmTableOpen(inp_p->infile);
  dmKeyRead_d(in_block, "RA_PNT", pnts + 0);
  dmKeyRead_d(in_block, "DEC_PNT", pnts + 1);
  dmKeyRead_d(in_block, "ROLL_PNT", pnts + 2);

  dmKeyRead_d(in_block, "RA_TARG", targ_crds + 0);
  dmKeyRead_d(in_block, "DEC_TARG", targ_crds + 1);

  dmTanWorldToPix(targ_crds, pnts, crpix, crdelt, fpc_targ);
  pix_fpc_to_extended_chip_and_id(fpc_targ, &chipid, chp_targ);

  dmTableClose(in_block);

  inp_p->ccd_id_focalpoint = chipid;
  if (inp_p->debug > DEBUG_LEVEL_2)
  {
    printf("DEBUG: CCD_ID focalpoint = %d\n", inp_p->ccd_id_focalpoint);
  }
}/* end determine_aimpoint */

/* check to determine if datamode == cc33_faint or datamode == cc33_graded
 * happens in acis_process_events.c before function call
 * if not cc mode, then only want the time min and time max values */
int initialize_acaofffile(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
dsErrList* err_p) /* O   - error list pointer  */
{
  /* Only initialize if obs_mode is pointing and asp_file is not None */
  if ((ds_strcmp_cis(inp_p->observation_mode, "pointing") == 0)
      && inp_p->use_asolfile)
  {
    if (inp_p->debug > DEBUG_LEVEL_2)
    {
      printf("DEBUG: Initializing the acaofffile %s\n", inp_p->asp_file);
    }

    /* Loop over all aspect solution files and create an array
     * of aspect_data structs */
    Stack asol_stack;
    char *current_asol = NULL;
    char *name_with_cols = NULL;
    long asol_count = 0;
    int ii = 0;
    long num_rows = 0;
    long retval = 0;
    long tot_num_rows = 0;
    long curr_row = 0;
    dmBlock *inBlock = NULL;
    dmDescriptor *timeCol = NULL;
    dmDescriptor *dyCol = NULL;
    dmDescriptor *dzCol = NULL;
    dmDescriptor *dthetaCol = NULL;
    dmDescriptor *raCol = NULL;
    dmDescriptor *decCol = NULL;
    dmDescriptor *rollCol = NULL;
    char columnlist[256];
    int retVal = 0;

    if (inp_p->is_cc33_mode)
    {
      strcpy(columnlist, "[columns time,ra,dec,roll,dy,dz,dtheta]");
    }
    else
    {
      strcpy(columnlist, "[columns time]");
    }

    /* Initialize stack */
    asol_stack = stk_build(inp_p->asp_file);
    /* get count of the number of files in the stack */
    asol_count = stk_count(asol_stack);

    /* loop over aspect files to get the total number of rows to create array */
    /* also perform data checks/validations */
    for (ii = 1; ii <= asol_count; ii++)
    {
      if ((current_asol = stk_read_num(asol_stack, ii)) != NULL)
      {
        inBlock = dmTableOpen(current_asol);

        num_rows = dmTableGetNoRows(inBlock);
        if (0 == num_rows || dmBADROW == num_rows)
        {
          dsErrAdd(err_p, dsPARAMVALUEWERR, Individual, Custom, "WARNING: Input aspect dataset contains no rows.");
        }
        tot_num_rows += num_rows;
        dmTableClose(inBlock);
        free(current_asol);
      }
    }

    inp_p->row_count = 0;
    inp_p->tot_num_rows = tot_num_rows;
    inp_p->aspect_data_array
        = (ASPECT_DATA_P)malloc(sizeof(ASPECT_DATA_T[tot_num_rows]));
    stk_rewind(asol_stack);
    for (ii = 1; ii <= asol_count; ii++)
    {
      if ((current_asol = stk_read_num(asol_stack, ii)) != NULL)
      {
        name_with_cols = malloc(sizeof(char) * 1024);
        if (name_with_cols == NULL)
        {
          dsErrAdd(err_p, dsALLOC2ERR, Individual, Generic, "input file name %s\n", current_asol);
          return 0;
        }

        strcpy(name_with_cols, current_asol);
        /* add in cols so we don't read in entire file */
        strcat(name_with_cols, columnlist);
        inBlock = dmTableOpen(name_with_cols);
        timeCol = dmTableOpenColumn(inBlock, "TIME");
        raCol = dmTableOpenColumn(inBlock, "RA");
        decCol = dmTableOpenColumn(inBlock, "DEC");
        rollCol = dmTableOpenColumn(inBlock, "ROLL");
        dyCol = dmTableOpenColumn(inBlock, "DY");
        dzCol = dmTableOpenColumn(inBlock, "DZ");
        dthetaCol = dmTableOpenColumn(inBlock, "DTHETA");

        do
        {
          inp_p->aspect_data_array[curr_row].time = dmGetScalar_d(timeCol);
          if (inp_p->is_cc33_mode)
          {
            inp_p->aspect_data_array[curr_row].ra = dmGetScalar_d(raCol);
            inp_p->aspect_data_array[curr_row].dec = dmGetScalar_d(decCol);
            inp_p->aspect_data_array[curr_row].roll = dmGetScalar_d(rollCol);
            inp_p->aspect_data_array[curr_row].dy = dmGetScalar_d(dyCol);
            inp_p->aspect_data_array[curr_row].dz = dmGetScalar_d(dzCol);
            inp_p->aspect_data_array[curr_row].dtheta
                = dmGetScalar_d(dthetaCol);
          }
          retval = dmTableNextRow(inBlock);

          curr_row++;
        }while (retval != dmNOMOREROWS);
      }/* end if current asol file */
      free(current_asol);
      free(name_with_cols);
      dmTableClose(inBlock);
    }/* end loop over aspect files */

    stk_close(asol_stack);

    double* time_vals = (double *)calloc(tot_num_rows, sizeof(double));
    /*add all the events to the array */
    for (ii = 0; ii < tot_num_rows; ii++)
    {
      time_vals[ii] = inp_p->aspect_data_array[ii].time;
    }

    /* sort time values */
    qsort(time_vals, tot_num_rows, sizeof(double), compare);

    /*(a) TIME_min and TIME_max */
    inp_p->time_min_max[0] = time_vals[tot_num_rows - 1];
    inp_p->time_min_max[1] = time_vals[0];

    /*(b) CHIPX_TARG_MED and CHIPY_TARG_med  */
    /* only compute median value if cc mode */
    if (inp_p->is_cc33_mode)
    {
      retVal = compute_chip_targ_median(inp_p, err_p);
      if(!retVal)
        return 0;
    }

    free(time_vals);
  }
  return 1;
}

void read_tg_m(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
EVENT_REC_P_T evt_p, /* record containing input event data */
dsErrList* err_p /* O   - error list pointer           */)
{
  /* Check we are in cc mode with gratings*/
  if ((inp_p->is_cc33_mode) && (inp_p->is_tg)
      && (ds_strcmp_cis(inp_p->observation_mode, "pointing") == 0))
  {
    /* Validation */
    /* Part A. */
    if (evt_p->tg_m < -99)
    {
      evt_p->tg_m = -99;
      dsErrAdd(err_p, dsGENERICERR, Accumulation, Custom, "WARNING: TG_M = -99\n");
    }
    /* Part B. */
    if (evt_p->tg_m > 99)
    {
      evt_p->tg_m = 99;
      dsErrAdd(err_p, dsGENERICERR, Accumulation, Custom, "WARNING: TG_M = 99\n");
    }
  }
}

int read_chipy_tg(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
EVENT_REC_P_T evt_p /* record containing input event data */)
{
  int retVal = 1;
  /* Check we are in cc mode with gratings*/
  if ((inp_p->is_cc33_mode) && (inp_p->is_tg)
      && (ds_strcmp_cis(inp_p->observation_mode, "pointing") == 0))
  {
    /* Validation */
    /* Part A. *//*Checks for null*/  /*11/13/2015: replace chipy_tg<9999 w/ nan */
    if (((evt_p->tg_m > -99) && (evt_p->tg_m < 99)) && (!ds_dNAN(evt_p->chipy_tg))
        && ((evt_p->chipy_tg <= 0) && (evt_p->chipy_tg >= 1025)))
    {
      retVal = 0;
    }
    else
    {
      /* Part B. *//*Checks for null*/ /*11/13/2015: replace chipy_tg<9999 w/ nan */
      if ((evt_p->tg_m > -99) && (evt_p->tg_m < 99) && (!ds_dNAN(evt_p->chipy_tg))
          && (evt_p->chipy_tg < 1))
      {
        evt_p->chipy_tg = 1;
      }
      /* Part C. */ /*11/13/2015: replace chipy_tg<9999 w/ nan */
      else if ((evt_p->tg_m > -99) && (evt_p->tg_m < 99) && (!ds_dNAN(evt_p->chipy_tg))
               && (evt_p->chipy_tg > 1024))
      {
        evt_p->chipy_tg = 1024;
      }
      else
      {
        /* no op */
        /* evt_p->chipy_tg = evt_p->chipy_tg */
      }
    }
  }

  return retVal;
}

void read_chipy_zo(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
EVENT_REC_P_T evt_p /* record containing input event data */)
{
  /* Check we are in cc mode with gratings*/
  if ((inp_p->is_cc33_mode) && (inp_p->is_tg)
      && (ds_strcmp_cis(inp_p->observation_mode, "pointing") == 0))
  {
    /* The value already read in based on columns in the infile */
    /* place holder function in case algorithm changes for determining chipy_zo
     * apart from just reading it in from infile */
    /* no op */
    /* evt_p->chipy_zo = evt_p->chipy_zo */
  }
}

void read_time_ro(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
EVENT_REC_P_T evt_p, /* record containing input event data */
dsErrList* err_p /* O   - error list pointer           */)
{
  /* (a) CC mode:
   If
   DATAMODEin = CC33 FAINT or (181)
   DATAMODEin = CC33 GRADED, (182)
   */
  if ((inp_p->is_cc33_mode))
  {
    /*
     then
     i. Read:
     A. Level 0:
     If
     CONTENTin = EVT0, (183)
     */
    if (ds_strcmp_cis(inp_p->content, CONTENT_EVT0) == 0)
    {
      /* then
       the value of TIME RO for an event is given by TIMEin. */
      evt_p->time_ro = evt_p->time;
    }
    /*
     B. Level 1, 1.5, or 2:
     If
     CONTENTin = EVT1, or (184)
     CONTENTin = TGEVT1, or (185)
     CONTENTin = EVT2, (186)
     */
    if ((ds_strcmp_cis(inp_p->content, CONTENT_EVT1) == 0)
        || (ds_strcmp_cis(inp_p->content, CONTENT_TGEVT1) == 0)
        || (ds_strcmp_cis(inp_p->content, CONTENT_EVT2) == 0))
    {
      /* then
       then the value of TIME RO for an event is given by TIME ROin. */
      /* no-op as this is what the value is read in as */
      if (evt_p->time_ro > 0)
      {
        /*evt_p->time_ro = evt_p->time_ro;*/
      }
      else
      {
        evt_p->time_ro = evt_p->time;
      }
    }

    /*
     ii. Validation:
     */
    if ((evt_p->time_ro < 0) || (evt_p->time_ro >= 3000000000))
    {
      dsErrAdd(err_p, dsGENERICERR, Accumulation, Custom, "WARNING: TIME_RO < 0 or TIME_RO >= 3*10^9\n");
    }
  }
}

int determine_time_chip_adj(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
EVENT_REC_P_T evt_p, /* record containing input event data */
dsErrList* err_p /* O   - error list pointer           */)
{
  int retVal = 1;
  double temp_time = 0.0;

  /*
   TIME and CHIPY ADJ:
   (a) Read or calculate:

   i. TE mode:
   */
  if (inp_p->supported_mode & TE_MODES)
  {
    /* no op on time since it is the time read in */
    /*evt_p->time = evt_p->time;*/
    evt_p->chip_adj[ADET_Y_AXIS] = evt_p->chip[ADET_Y_AXIS];
  }

  /*
   ii. CC mode:
   */
  else if (inp_p->is_cc33_mode)
  {
    /* A. */
    evt_p->chip_adj[ADET_Y_AXIS] = 512;
    evt_p->time_work = evt_p->time_ro - (evt_p->chip_adj[ADET_Y_AXIS] + 1028)
        * inp_p->timedel;

    /* B. */
    if (inp_p->use_asolfile)
    {
      evt_p->chipy_adj_work = inp_p->chip_targ_eff[ADET_Y_AXIS];
      evt_p->time_work = evt_p->time_ro - (evt_p->chipy_adj_work + 1028)
          * inp_p->timedel;

      /* calculate chipy_adj */
      compute_chip(inp_p, evt_p->time_work, FALSE);
      evt_p->chip_adj[ADET_Y_AXIS] = inp_p->chip_adj_cc_mode[ADET_Y_AXIS];
    }

    /* C. */ /*11/13/2015: replace chipy_tg<tnull w/ nan */
    if ((inp_p->is_tg == 1) && (evt_p->tg_m > -99) && (evt_p->tg_m < 99)
        && (!ds_dNAN(evt_p->chipy_tg)) )
    {
      evt_p->chip_adj[ADET_Y_AXIS] = evt_p->chipy_tg;
    }

    /* D. */
    if ((inp_p->is_tg == 1) && ((evt_p->tg_m == -99) || (evt_p->tg_m == 99)))
    {
      evt_p->chip_adj[ADET_Y_AXIS] = 512;
    }

    temp_time = evt_p->time_work;
    if (inp_p->use_asolfile && ((temp_time < inp_p->time_min_max[0])
        || (temp_time >= inp_p->time_min_max[1])))
    {
      /* E. */
      evt_p->chip_adj[ADET_Y_AXIS] = 512;

      /* F. */
      if(((inp_p->ccd_id_focalpoint == 0) || (inp_p->ccd_id_focalpoint == 2)) &&
          ((evt_p->ccdid == 0) || (evt_p->ccdid == 2)))
      {
        evt_p->chip_adj[ADET_Y_AXIS] = inp_p->chip_targ_med[ADET_Y_AXIS];
      }

      /* G. */
      if(((inp_p->ccd_id_focalpoint == 1) || (inp_p->ccd_id_focalpoint == 3)) &&
          ((evt_p->ccdid == 1) || (evt_p->ccdid == 3)))
      {
        evt_p->chip_adj[ADET_Y_AXIS] = inp_p->chip_targ_med[ADET_Y_AXIS];
      }

      /* H. */
      if((inp_p->ccd_id_focalpoint >= 4) && (evt_p->ccdid >= 4))
      {
        evt_p->chip_adj[ADET_Y_AXIS] = inp_p->chip_targ_med[ADET_Y_AXIS];
      }
    }


    /* I. */
    if (!(inp_p->use_asolfile) && !(inp_p->is_tg))
    {
      evt_p->chip_adj[ADET_Y_AXIS] = 512;
    }

    /* J. */
    if ((inp_p->ccd_id_focalpoint <= 3) && (evt_p->ccdid >= 4))
    {
      evt_p->chip_adj[ADET_Y_AXIS] = 512;
    }

    /* K. */
    if ((inp_p->ccd_id_focalpoint >= 4) && (evt_p->ccdid <= 3))
    {
      evt_p->chip_adj[ADET_Y_AXIS] = 512;
    }

    /* L. */
    if ((ds_strcmp_cis(inp_p->observation_mode, "secondary") == 0))
    {
      evt_p->chip_adj[ADET_Y_AXIS] = 512;
    }

    /* M. */
    if ((evt_p->chip_adj[ADET_Y_AXIS] < 0.5) || (evt_p->chip_adj[ADET_Y_AXIS]
        >= 1024.5))
    {
      evt_p->chip_adj[ADET_Y_AXIS] = 512;
    }

    /* N. */
    evt_p->time = evt_p->time_ro - (evt_p->chip_adj[ADET_Y_AXIS] + 1028)
        * inp_p->timedel;
  }

  /*
   (b) Validation:
   */
  if ((evt_p->time < 0) || (evt_p->time >= 3000000000))
  {
    dsErrAdd(err_p, dsGENERICERR, Accumulation, Custom, "WARNING: TIME < 0 or TIME >= 3*10^9\n");
  }

  if ((evt_p->chip_adj[1] < 0.5) || (evt_p->chip_adj[1] >= 1024.5))
  {
    /*printf("chipy_adj %f, %f, %d, %ld, %f\n", evt_p->chip_adj[1], evt_p->time, evt_p->ccdid, evt_p->expno, evt_p->chip[0]);*/
    retVal = 0;
  }

  return retVal;
}

void determine_time_adj(INPUT_PARMS_P_T inp_p, /* I - input param data structure     */
EVENT_REC_P_T evt_p /* record containing input event data */)
{
  /*
   (a) TE mode
   */
  if (inp_p->supported_mode & TE_MODES)
  {
    strcpy(inp_p->time_adj, "NONE");
  }
  else
  {
    strcpy(inp_p->time_adj, "MIDCHIP");

    if (ds_strcmp_cis(inp_p->observation_mode, "pointing") == 0)
    {
      if (inp_p->use_asolfile)
      {
        strcpy(inp_p->time_adj, "TARGET");
      }
      if (inp_p->is_tg)
      {
        strcpy(inp_p->time_adj, "GRATING");
      }
    }
  }
}

/*
 * Helper Functions
 */

int compare(const void *val1, const void *val2)
{
  const double *tmpVal1 = val1;
  const double *tmpVal2 = val2;

  double diff = *tmpVal1 - *tmpVal2;

  return ((diff >= 0) ? ((diff > 0) ? -1 : 0) : +1);
}

void retrieve_coord(INPUT_PARMS_P_T inp_p, double time)
{
  double t_asol = 0;
  double t_asol_next = 0;
  long ii = 0;
  long final_row = 0;
  bool set_to_zero = FALSE;
  /*check if for whatever reason row_count < 0, set to 0 */
  if (inp_p->row_count < 0)
  {
    inp_p->row_count = 0;
  }

  long start = inp_p->row_count;

  /*need to determine which row of the asol file to use */
  for (ii = start; ii < inp_p->tot_num_rows; ii++)
  {
    inp_p->row_count++;
    t_asol = inp_p->aspect_data_array[ii].time;

    if ((ii == 0) && (time < inp_p->time_min_max[0]))
    {
      break;
    }
    else
    {
      if((ii + 1) != inp_p->tot_num_rows)
      {
        t_asol_next = inp_p->aspect_data_array[ii + 1].time;
      }
      else
      {
        if(time > inp_p->time_min_max[1])
        {
          break;
        }

        t_asol_next = inp_p->time_min_max[1];
      }

      if (time < t_asol_next)
      {
        if ((time >= t_asol))
        {
          break;
        }
        else
        {
          /*to speed up things, I store where we are in the asol file as
           * row_count.  Since we are progressing through the event file
           * in chronological order, we don't need to go all the way back
           * to the beginning of the asol file for each iteration.  However,
           * with time adjustments, we may need a few rows before current row.
           * If this is the case, back up a row at a time and retry.
           */
          while (time < t_asol)
          {
            ii--;
            t_asol = inp_p->aspect_data_array[ii].time;

            /* if we get back to the beginning, use first row */
            if (ii == 0)
            {
              set_to_zero = TRUE;
              break;
            }
          }
          break;
        }
      }
    }
  }

  final_row = ii;

  if (set_to_zero)
    inp_p->row_count = 0;
  else
    inp_p->row_count--;

  inp_p->coord_data.dy = inp_p->aspect_data_array[final_row].dy;
  inp_p->coord_data.dz = inp_p->aspect_data_array[final_row].dz;
  inp_p->coord_data.dtheta = inp_p->aspect_data_array[final_row].dtheta;
  inp_p->coord_data.ra = inp_p->aspect_data_array[final_row].ra;
  inp_p->coord_data.dec = inp_p->aspect_data_array[final_row].dec;
  inp_p->coord_data.roll = inp_p->aspect_data_array[final_row].roll;
}

int compute_chip_targ_median(INPUT_PARMS_P_T inp_p, dsErrList* err_p)
{
  VEC2_DBLE mpy; /* mirror pitch and yaw angles in degrees */
  VEC3_DBLE stage; /* moving part of  3-D HXDS position in mm */
  VEC3_DBLE stage_ang;
  static double crpix[2] = {4096.5, 4096.5};
  static double chp_targ[2] = {512.0, 512.0};
  static double fpc_targ[2];
  static double crdelt[2] = {(-0.492 / 3600.0), (0.492 / 3600.0)};
  short chipid;
  static double targ_crds[2];
  double pnts[3];
  double median_x = 0.0;
  double median_y = 0.0;
  double *chipy_targ_median = (double*)calloc(inp_p->tot_num_rows,sizeof(double));
  double *chipx_targ_median = (double*)calloc(inp_p->tot_num_rows,sizeof(double));

  if ((NULL == chipy_targ_median) || (NULL == chipx_targ_median))
  {
    err_msg("ERROR: Out of memory while computing the chip targ median values\n");
    return 0;
  }

  targ_crds[0] = inp_p->ra_targ_evt;
  targ_crds[1] = inp_p->dec_targ_evt;

  long ii = 0;
  for (ii = 0; ii < inp_p->tot_num_rows; ii++)
  {
    /* initialize values for mirror/alignment */
    mpy[0] = mpy[1] = 0.0;
    stage[0] = stage[1] = 0.0;
    stage_ang[0] = stage_ang[1] = stage_ang[2] = 0.0;

    stage[1] = inp_p->aspect_data_array[ii].dy * (-1.0);
    stage[2] = inp_p->aspect_data_array[ii].dz * (-1.0);
    stage_ang[0] = inp_p->aspect_data_array[ii].dtheta;

    pnts[0] = inp_p->aspect_data_array[ii].ra;
    pnts[1] = inp_p->aspect_data_array[ii].dec;
    pnts[2] = inp_p->aspect_data_array[ii].roll;

    pix_set_mirror(mpy, stage, stage_ang);
    dmTanWorldToPix(targ_crds, pnts, crpix, crdelt, fpc_targ);
    pix_fpc_to_extended_chip_and_id(fpc_targ, &chipid, chp_targ);

    chipx_targ_median[ii] = chp_targ[ADET_X_AXIS];
    chipy_targ_median[ii] = chp_targ[ADET_Y_AXIS];
  }

  /* sort column */
  qsort(chipx_targ_median, inp_p->tot_num_rows, sizeof(double), compare);
  qsort(chipy_targ_median, inp_p->tot_num_rows, sizeof(double), compare);

  /* get median */
  if ((inp_p->tot_num_rows % 2) == 1)
  {
    median_x = (chipx_targ_median[(inp_p->tot_num_rows / 2) - 1]
        + chipx_targ_median[inp_p->tot_num_rows / 2]) / 2;
    median_y = (chipy_targ_median[(inp_p->tot_num_rows / 2) - 1]
            + chipy_targ_median[inp_p->tot_num_rows / 2]) / 2;
  }
  else if (inp_p->tot_num_rows > 0)
  {
    median_x = chipx_targ_median[(inp_p->tot_num_rows / 2) - 1];
    median_y = chipy_targ_median[(inp_p->tot_num_rows / 2) - 1];
  }
  else
  {
    ds_MAKE_DNAN(median_x);
    ds_MAKE_DNAN(median_y);
  }

  if (inp_p->debug > DEBUG_LEVEL_2)
  {
    printf("DEBUG: Median value of chipx_targ = %f\n", median_x);
    printf("DEBUG: Median value of chipy_targ = %f\n", median_y);
  }

  /*set the value in the input struct*/
  inp_p->chip_targ_med[ADET_X_AXIS] = median_x;
  inp_p->chip_targ_med[ADET_Y_AXIS] = median_y;
  
  free(chipy_targ_median);
  free(chipx_targ_median);

  return 1;
}

void compute_chip(INPUT_PARMS_P_T inp_p,
                   double time,
                   bool isTarg)
{
  VEC2_DBLE mpy; /* mirror pitch and yaw angles in degrees */
  VEC3_DBLE stage; /* moving part of  3-D HXDS position in mm */
  VEC3_DBLE stage_ang;
  double fpc_targ[2];
  static double crpix[2] = {4096.5, 4096.5};
  static double crdelt[2] = {(-0.492 / 3600.0), (0.492 / 3600.0)};
  double targ_crds[2] = {inp_p->ra_targ_evt, inp_p->dec_targ_evt};
  double pnts[3];

  /* initialize values for mirror/alignment */
  mpy[0] = mpy[1] = 0.0;
  stage[0] = stage[1] = 0.0;
  stage_ang[0] = stage_ang[1] = stage_ang[2] = 0.0;

  /*call to retrieve stf */
  retrieve_coord(inp_p, time);

  stage[1] = inp_p->coord_data.dy * (-1.0);
  stage[2] = inp_p->coord_data.dz * (-1.0);
  stage_ang[0] = inp_p->coord_data.dtheta;

  pnts[0] = inp_p->coord_data.ra;
  pnts[1] = inp_p->coord_data.dec;
  pnts[2] = inp_p->coord_data.roll;

  pix_set_mirror(mpy, stage, stage_ang);
  dmTanWorldToPix(targ_crds, pnts, crpix, crdelt, fpc_targ);
  if (isTarg)
  {
    pix_fpc_to_extended_chip(fpc_targ, inp_p->ccd_id_focalpoint, inp_p->chip_targ_evt);
  }
  else
  {
    pix_fpc_to_extended_chip(fpc_targ, inp_p->ccd_id_focalpoint, inp_p->chip_adj_cc_mode);
  }
}
