/*                                                                
**  Copyright (C) 1996,2007,2010-2011,2013-2015  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*H***********************************************************************
 
 * FILE NAME: coordinate_transforms.c
 
 * DEVELOPEMENT: tools
 
 * DESCRIPTION: Coordinate_transforms.c contains the routines used by
 acis_process_events to compute various coordinate transformations.
 
 * NOTES:

 * REVISION HISTORY:
 
 Ref. No.        Date
 --------        ----
 1.1             18 Jul 1996

11/4/2015-bug 14174 (detY shouldn't be very different w/ pix_adj=random or edser)
11/13/2015- bug14203: change tnull to nan for chipy_zo, chipy_tg ; don't set tnull or nan for tdet;
 *H***********************************************************************/


#include <stdio.h>
#include <ctype.h> 

#include <stdlib.h> 

#ifndef ACIS_PROCESS_EVENTS_H
#include "acis_process_events.h"
#define ACIS_PROCESS_EVENTS_H
#endif 

#ifndef L1_ASPECT_DEFS_H
#include "l1_aspect_defs.h"
#define L1_ASPECT_DEFS_H
#endif

static void calculate_sky_1d(INPUT_PARMS_P_T, EVENT_REC_P_T);

/*H**************************************************************************
 
 * DESCRIPTION:
 
 The module calculate_coords_acis() is used by acis_process_events to convert
 coordinates from the start coordinate system specified in the acis_process_events
 parameter file to the stop coordinates specified in the parameter file.
 The conversions are done primarily through pixel library routines.
 
 * NOTES:
 
 *H**************************************************************************/

void calculate_coords_acis(ASPECT_REC_P_T aspect, /* I   aspect information             */
INPUT_PARMS_P_T inp_p, /* I   input parameter data structure */
short asp_type_flag, /* I  asp corr offsets or solution?   */
EVENT_REC_P_T evt_p, /* I/O event record data structure    */
BADPIX_SETUP_P_T bp_p, /* I/O - bad pix housekeeping structure*/
SUBPIX_CAL *subpix_p)
{
  VEC2_DBLE real_val_coord;
  VEC2_DBLE nint;

  if (strcmp(inp_p->pix_adj, "CENTROID") == 0)
  {
    calculate_centroid(evt_p, inp_p, bp_p);
  }
  else if (strcmp(inp_p->pix_adj, "EDSER") == 0)
  {
    apply_edser(evt_p, subpix_p, inp_p);
  }
  else if (strcmp(inp_p->pix_adj, "NONE") == 0)
  {
    /* no op
     * chipx_adj = chipx_adj
     * chipy_adj = chipy_adj
     * time = time */
  }
  else if (strcmp(inp_p->pix_adj, "RANDOMIZE") == 0)
  {
    double rand_x, rand_y;

    /* apply a random value form -randpixsize to +randpixsize
     * to avoid quantization artifacts. (pix_short_to_double
     * doesn't accept 0 is input so use 1 and subtract it from
     * the returned value.
     */
    pix_short_to_double(1, 1, &rand_x);
    pix_short_to_double(1, 1, &rand_y);
    rand_x--;
    rand_y--;
    rand_x *= (2 * inp_p->randpixsize);
    rand_y *= (2 * inp_p->randpixsize);

    evt_p->chip_adj[ADET_X_AXIS] = evt_p->chip_adj[ADET_X_AXIS] + rand_x;  /*11/4/2015*/
    evt_p->chip_adj[ADET_Y_AXIS] = evt_p->chip_adj[ADET_Y_AXIS] + rand_y; 

    if (inp_p->is_cc33_mode)
    {
      evt_p->time = evt_p->time - (rand_y * inp_p->timedel);
    }
  }

  /* convert fast window chip coords to real chip coords (only for acis-2c) */
  if ((inp_p->winsize != WINSIZE_DEFAULT_VAL) && (inp_p->found_foa))
  {
    VEC2_DBLE real_chip;

    pix_wchip_to_chip(inp_p->winsize, inp_p->cy0, evt_p->chip_adj, real_chip);

    evt_p->chip_adj[ADET_X_AXIS] = real_chip[ADET_X_AXIS];
    evt_p->chip_adj[ADET_Y_AXIS] = real_chip[ADET_Y_AXIS];
  }

  if (evt_p->chip_adj[ADET_X_AXIS] < 0.5)
    evt_p->chip_adj[ADET_X_AXIS] = 1;
  if (evt_p->chip_adj[ADET_X_AXIS] >= 1024.5)
    evt_p->chip_adj[ADET_X_AXIS] = 1024;
  if (evt_p->chip_adj[ADET_Y_AXIS] < 0.5)
    evt_p->chip_adj[ADET_Y_AXIS] = 1;
  if (evt_p->chip_adj[ADET_Y_AXIS] >= 1024.5)
    evt_p->chip_adj[ADET_Y_AXIS] = 1024;

  switch(inp_p->stop_transform)
  {
    case ADET_SKY_VAL: /* fallthru intended */
    {
      /* i. */
      if (inp_p->supported_mode & TE_MODES)
      {
        real_val_coord[0] = evt_p->chip_adj[ADET_X_AXIS];
        real_val_coord[1] = evt_p->chip_adj[ADET_Y_AXIS];
      }
      /* ii. */
      else if (inp_p->is_cc33_mode)
      {
        /* A. */
        real_val_coord[0] = evt_p->chip_adj[ADET_X_AXIS];
        real_val_coord[1] = inp_p->chip_targ_eff[ADET_Y_AXIS];

        /* B.
         * This is commented out purposefully
         * Initially was added as a hack before there was a
         * pix_extended_chip_to_fpc below
         * Leaving it here for now*/
        /*if((inp_p->chip_targ_med[ADET_Y_AXIS] < 16.5)
         && (inp_p->ccd_id_focalpoint >=4)
         && (evt_p->ccdid >= 4))
         {
         real_val_coord[1] = 512 + (inp_p->chip_targ_eff[ADET_Y_AXIS] -
         inp_p->chip_targ_med[ADET_Y_AXIS]);
         }*/
      }

      /* Used with the above hack B. but not needed now the extended chip
       * version is available.
       * pix_chip_to_fpc(evt_p->ccdid, real_val_coord, evt_p->fppos);*/
      pix_extended_chip_to_fpc(evt_p->ccdid, real_val_coord, evt_p->fppos);

      if (inp_p->datatype == EVT_FLT_FORMAT)
      {
        VEC2_DBLE fpc;

        fpc[0] = evt_p->fppos[ADET_X_AXIS];
        fpc[1] = evt_p->fppos[ADET_Y_AXIS];

        if (asp_type_flag == ASP_FTYPE_OFFSETS)
        {
          pix_apply_aspect(fpc, aspect->asp_sol, evt_p->sky);
        }
        else
        {
          double cel[2];
          if (((inp_p->in_mode & CC_MODES) != 0) && ((evt_p->time
              < inp_p->time_min_max[0]) || (evt_p->time
              > inp_p->time_min_max[1])))
          {
            /*aspect values may be zero in the aspect struct (because
             * time < time_min or time > time_max)
             * so if ccmode use the values obtained directly from the asol file
             */
            double asp_sol[3];
            asp_sol[0] = inp_p->coord_data.ra;
            asp_sol[1] = inp_p->coord_data.dec;
            asp_sol[2] = inp_p->coord_data.roll;
            dmTanPixToWorld(fpc, asp_sol, inp_p->crpix, inp_p->cdelt, cel);
          }
          else
          {
            dmTanPixToWorld(fpc, aspect->asp_sol, inp_p->crpix, inp_p->cdelt, cel);
          }
          dmTanWorldToPix(cel, inp_p->crval, inp_p->crpix, inp_p->cdelt, evt_p->sky);
        }

        if ((inp_p->in_mode & CC_MODES) != 0)
        {
          calculate_sky_1d(inp_p, evt_p);
        }
      }
      else
      {
        evt_p->sky[ADET_X_AXIS] = evt_p->fppos[ADET_X_AXIS];
        evt_p->sky[ADET_Y_AXIS] = evt_p->fppos[ADET_Y_AXIS];
      }
    }
    case ADET_DET_VAL: /* fallthru intended */
    {
      real_val_coord[0] = evt_p->chip_adj[ADET_X_AXIS];
      real_val_coord[1] = evt_p->chip_adj[ADET_Y_AXIS];

      pix_chip_to_fpc(evt_p->ccdid, real_val_coord, evt_p->det);
    }
    case ADET_TDET_VAL: /* fallthru intended */
    {
      nint[0] = floor(evt_p->chip_adj[0] + 0.5);
      nint[1] = floor(evt_p->chip_adj[1] + 0.5);
      pix_chip_to_tdet(evt_p->ccdid, nint, evt_p->tdet);
      break;
    }
    case ADET_NONE_VAL:
    {
      /* do no coordinate transformations */
      break;
    }
    default:
      break;
  }
  /* check if the values need to be set to nan/null */
  double temp_evt_time = evt_p->time;
  bool is_out_asol = FALSE;

  if (inp_p->use_asolfile)
  {
    is_out_asol = ((temp_evt_time < inp_p->time_min_max[0]) || (temp_evt_time
        >= inp_p->time_min_max[1]));

    if ((!(inp_p->is_tg) && is_out_asol) || (inp_p->is_tg && ds_dNAN(evt_p->chipy_tg)))
    {
      /* only set the tdet vals to null if running on gratings data
       * tg_resolve_events data only run when cc mode and hetg */
      if(inp_p->is_tg)
      {
        evt_p->tdet[ADET_X_AXIS] = 9999;
        evt_p->tdet[ADET_Y_AXIS] = 9999;
      }

      ds_MAKE_DNAN(evt_p->det[ADET_X_AXIS]);
      ds_MAKE_DNAN(evt_p->det[ADET_Y_AXIS]);

      ds_MAKE_DNAN(evt_p->sky[ADET_X_AXIS]);
      ds_MAKE_DNAN(evt_p->sky[ADET_Y_AXIS]);

      ds_MAKE_DNAN(evt_p->chipy_zo );    /* 11/13/2015 */
      ds_MAKE_DNAN(evt_p->chipy_tg );

      if ((inp_p->supported_mode & CC_MODES) != 0)
      {
        ds_MAKE_DNAN(evt_p->sky_1d);
      }
    }
  }
}

/****************************************************************************

 * DESCRIPTION

 The routine parse_coord_range is used to translate start and stop coordinate
 strings into integer values. The integer values may then be used elsewhere
 in the code to avoid having to perform string compares.

 * NOTES

 If the specified string is unknown, a value of ADET_INV_VAL is returned.

 ****************************************************************************/

short parse_coord_range(char* string) /* I - start/stop string */
{
  short val = ADET_INV_VAL;

  if (strcmp(string, ADET_NO_COORD) == 0)
  {
    val = ADET_NONE_VAL;
  }
  else if (strcmp(string, ADET_CHIP_COORD) == 0)
  {
    val = ADET_CHIP_VAL;
  }
  else if (strcmp(string, ADET_TDET_COORD) == 0)
  {
    val = ADET_TDET_VAL;
  }
  else if (strcmp(string, ADET_DET_COORD) == 0)
  {
    val = ADET_DET_VAL;
  }
  else if (strcmp(string, ADET_TAN_COORD) == 0)
  {
    val = ADET_TAN_VAL;
  }
  else if (strcmp(string, ADET_SKY_COORD) == 0)
  {
    val = ADET_SKY_VAL;
  }

  return (val);
}

/* Spec comes from http://space.mit.edu/ASC/docs/acis_l1_2.3.pdf
 * Last paragraph on page 14
 */
static void calculate_sky_1d(INPUT_PARMS_P_T inp_p, /* I   - input parameter data structure */
EVENT_REC_P_T evt_p) /* I/O - event record data structure    */
{
  double theta, temp;
  double phi, c1;

  dmCoordCalc_d(inp_p->sky_desc, evt_p->sky, evt_p->radec);

  temp = cos(evt_p->radec[1] * DEG2RAD) * cos(inp_p->dec_targ * DEG2RAD)
      * cos((evt_p->radec[0] - inp_p->ra_targ) * DEG2RAD) + sin(evt_p->radec[1]
      * DEG2RAD) * sin(inp_p->dec_targ * DEG2RAD);
  theta = acos(temp);

  if ((evt_p->ccdid >= 0) && (evt_p->ccdid < 4))
  {
    phi = 90;
  }
  else if (evt_p->ccdid <= 9)
  {
    phi = 0;
  }
  else
  {
    /* error- ccdid out of range */
  }

  c1 = cos((inp_p->roll_nom - phi) * DEG2RAD) * (evt_p->sky[0]
      - inp_p->crpix[0]) - sin((inp_p->roll_nom - phi) * DEG2RAD)
      * (evt_p->sky[1] - inp_p->crpix[1]);

  if (c1 < 0)
  {
    evt_p->sky_1d = inp_p->crpix[0] + 3600 * theta * RAD2DEG;
  }
  else
  {
    evt_p->sky_1d = inp_p->crpix[0] - 3600 * theta * RAD2DEG;
  }
}
