/*                                                                
**  Copyright (C) 1997-2007,2010,2014-2015  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/************************************************************************
 
* FILE NAME: tg_process_event_data.c
 
* DESCRIPTION:
 
  The routine tg_process_event_data() is called by tg_resolve_events to 
  calculate level 1.5 products such as coords (r,d, gdpx, gdpy), mlambda,
  and lamdba. The coordinate transformations are acutally handled via a 
  call to the routine compute_rd() which is also contained in this file.   
  The routine tg_process_event_data_CC() does mostly the same thing
  as tg_process_event_data(), but for ACIS Continuous Clocking (CC) mode
  dataset only (HETG gratings only).

  None of the routines return an error status. 

  LAST MODIFIED:
  6/2000:(1)Modified the tool to handle the new rmf table.  The
	    energy_hi and energy_lo in the new rmf table are not only
	    ccd dependent but also chipx and chipy dependent.
	    The tool can still handle the old rmf table.
 (2/2005)-no check out on the min/max ranges of chipxy before 
          processing data(compute_rd).
 (6/2005)- rename  part->e_part, t_part or parrt; PART->PARRT; rr->rrr;
         - add comments;
7/2007 - add chipy_zo and set it to zo_chip[ Y ]. 
       - remove unused variables.
4/2010 - code cleanup

10/2014 - no code changes except:  doc and renames variables ( see  tg_r_e.h ) 
( the new file : tg_process_evtData_hetgCC_2ndRun.c  came from tg_process_event_data_CC() )

6/10/2015 :
 - add new fltfile param; 
   If evt_adj_time falls off flt or pcad time range, then
   set chipy_zo=chipy_tg=tnull=99999.0 and set tg_part=1(letg) or 2(hetg), tg_m=null=99 

6/17/2015 - add is_sky_nan;
  (note:  if input sky=NaN -> src_tags[0]=TG_NOT_IN_SRC=0 -> is_sky_nan=1 )

6/29/2015 - new way to set time_gti for flt GTIn
          - use set_gti_flag() to set final flt gti flag.
          - use chipy_tg to set evt_adj_time instead of using chipy_zo
6/30/2015 - merge code for CC 1st and 2nd pass :
            tweaked tg_process_event_data_CC() to make it work for CC 2nd pass as well.
8/24/2015 - pull the entire flt gtin check out
            ( rename  set_gti_flag() to  set_pcad_sky_flag() )
            ( rename is_in_gti      to  pcad_sky_flag )
11/13/2015- bug14203: change tnull to nan for chipy_zo, chipy_tg; don't set tnull or nan for tdet;
* ************************************************************************/

#ifndef TG_RESOLVE_EVENTS_H
#include "tg_resolve_events.h"
#define TG_RESOLVE_EVENTS_H
#endif

#ifndef TG_REGION_TABLE_H
#include "tg_region_table.h"
#define TG_REGION_TABLE_H
#endif 



/* 4/29/2015 - make this public instead of  "static void compute_rd()"  */
void compute_rd( 
   EVENT_REC_P_T    evt_p,  /* I/O - event record data structure pointer    */
   GRATING_EVT_P_T  grt_p,  /* I/O - gratings data structure pointer	    */ 
   INPUT_PARMS_P_T  inp_p,  /* I - input params struct (chip ranges)	    */
   dsErrList*	    err_p)  /* O - error list pointer			    */
{
  /* Only compute for valid chip values inp_p->chip_minrange. */ 
  /* That means the following condition should hold:
   * evt_p->chip_work[0] >= inp_p->chip_minrange
   * evt_p->chip_work[0] <= inp_p->chip_maxrange[0]
   * evt_p->chip_work[1] >= inp_p->chip_minrange
   * evt_p->chip_work[1] >= inp_p->chip_maxrange[1]
   */

      if (inp_p->randpixsize > FLT_EPSILON)
      {
	 double rand_x, rand_y;

	 /* apply a random value from -randpixsize to +randpixsize
	  * to avoid quantization artifacts. (pix_short_to_double
	  * doesn't accept 0 is input so use 1 and subtract it from
	  * the returned value.
	  */
	 pix_short_to_double(1, 1, &rand_x);
	 pix_short_to_double(1, 1, &rand_y);
	 rand_x--;
	 rand_y--;
	 rand_x *= (2*inp_p->randpixsize);
	 rand_y *= (2*inp_p->randpixsize);

	 if ( inp_p->cc_mode ) {
	   evt_p->chip_work[TG_X_AXIS] = evt_p->chip[TG_X_AXIS] + rand_x;
	   evt_p->chip_work[TG_Y_AXIS] = evt_p->chip_work[TG_Y_AXIS] + rand_y;
	 } else {
	   evt_p->chip_work[TG_X_AXIS] = evt_p->chip[TG_X_AXIS] + rand_x;
	   evt_p->chip_work[TG_Y_AXIS] = evt_p->chip[TG_Y_AXIS] + rand_y;
	 }
      }

      pix_chip_to_fpc(evt_p->chipid, evt_p->chip_work, evt_p->det_work);
      pix_fpc_to_gdp(evt_p->det_work, grt_p->t_gdp_xy);

      /* CRC - Nov 18 2002 */
      /* J. Davis' patch from message
       * 200211160032.gAG0W1Ee003454@aluche.mit.edu */
      if (1)
      { 
          double fc[3] = { 0, 0, 0};
          double gzo[3] = { 0, 0, 0};
	   
          pix_chip_to_fc (evt_p->chipid, evt_p->chip_work, fc);
          pix_fc_to_gzo(fc, gzo);
          pix_gzo_to_gac(gzo, grt_p->t_gac_rd);
      }  
      else
      {  
          pix_gdp_to_gac(grt_p->t_gdp_xy, grt_p->t_gac_rd);
      }  

      evt_p->chip_work_new = evt_p->chip_work[TG_Y_AXIS] ; /*6/30/3015-add; outCol chipy_tg*/

} /* end: compute_rd */


/* ------------------------------------------------------------------------- */


/* there is a copy of this function, called tg_process_event_data_CC(), 
   which has CC mode additions.	 NOTE:	when updating this function, please 
   update tg_process_event_data_CC() as well, since there are similarities 
   in the processing.  this copy was made in order to save time 
   in non-cc mode processing */

void tg_process_event_data(
        SOURCE_ENTRY_A_T          srcs,   /*I - src list */  
        SRC_TAGS_A_T              tags,   /*I - track which srcs an evt is in */
        RSP_MTRX_REC_P_T          rm_p,   /*I - energy limits struct. ptr */ 
        NEW_RSP_MTRX_REC_P_T  new_rm_p,   /*I - new format energy limits struct.ptr */
        float              **rm_energy,   /*I - energy data array for NEW rmf table */
        float           **rm_energy_lo,   /*I - energy low data array for NEW rmf table */
        float           **rm_energy_hi,   /*I - energy hi data array for NEW rmf table */
        long       energy_vec_col_size,   /*I - dim of energy vector for NEW rmf table */
        INPUT_PARMS_P_T          inp_p,   /*I - input parameter structure ptr */
        ASPECT_REC_P_T           asp_p,   /*I - aspect record structure ptr */ 
        ALIGNMENT_INFRA_P_T   aln_hk_p,   /*I - alignment housekeeping ptr */ 
        ALIGNMENT_REC_P_T        aln_p,   /*I - alignment record structure ptr */
        short            asp_type_flag,   /*I - type of aspect correction */
        EVENT_REC_P_T            evt_p,   /*I/O - evt record structure ptr */
        enum RMFileType   rm_type_flag,
        dsErrList*                err_p   /*O - error list ptr */
    )   

{   
   VEC2_DBLE zo_fpc = {0.0, 0.0}; 
   VEC2_DBLE zo_chip;
   VEC3_DBLE zo_fc	 = {0.0, 0.0}; 
   short rrr, cnt = 0, chipid;
   GRATING_EVT_T grt_a[MAX_NUM_SOURCES*2]; /* assume heg & meg per src */ 

   /* the following are needed for applying aspect correction from the 
      aspect solution */
   static double nominal_asp[3];
   static double crpix[2];
   static double cdelt[2];
   static short first_run = 1;
   
   if ( first_run )
   {  
      nominal_asp[0] = inp_p->kywd_nom_ra;
      nominal_asp[1] = inp_p->kywd_nom_dec;
      nominal_asp[2] = 0.0;
      
      pix_get_fp_image_cntr(crpix);
      
      cdelt[1] = inp_p->fp_scale/3600.0;  
      cdelt[0] = -cdelt[1];   
      
      first_run = 0;
   }
   
   memset(grt_a, 0, (MAX_NUM_SOURCES*2*sizeof(GRATING_EVT_T))); 
   
   /* ----------- BEGIN for src_cnt from the region file --------*/
   for (rrr = 0; rrr < inp_p->src_cnt; rrr++)
     {
       short piD = 0; 
       SRC_TAGS_T mask = tags[rrr];
       
       /* -------------- */ 

       if (tags[rrr] != TG_NOT_IN_SRC) 
	 {
	   /* aspect correct to convert from sky to fpc */ 
	   if (asp_type_flag == ASP_FTYPE_OFFSETS)
	     {	
	       pix_deapply_aspect(srcs[rrr].s_center, asp_p->asp_sol, zo_fpc); 
	     }
	   else
	     {
	       double cel[2];
	       
	       dmTanPixToWorld( srcs[rrr].s_center, nominal_asp, crpix, cdelt, cel );
	       dmTanWorldToPix( cel, asp_p->asp_sol, crpix, cdelt, zo_fpc );
	     }
 
	   
	   /* update sim alignment if necessary */
	   
	   /* add time offset to event time for dither sequence */
	    evt_p->time += inp_p->time_offset;
 
	    if (alignment_update(aln_p, aln_hk_p, evt_p->time))
	      {
		/* error- eventually should pass back to calling routine */
	      }
	    /* restore original time after dither sequence */
	    evt_p->time -= inp_p->time_offset;
	    
	    pix_fpc_to_fc(zo_fpc, zo_fc); 

            chipid = evt_p->ccdid; 
            pix_fpc_to_extended_chip( zo_fpc, chipid, zo_chip );
            evt_p->chipy_zo = zo_chip[ TG_Y_AXIS ] ;
	    
	    /* compute gratings coords in pixels and angles */ 
/*	    pix_chipid_to_chip(inp_p->tdetsys, evt_p->chipid, evt_p->chipname);
 */
	    
	    while (mask > TG_NOT_IN_SRC)
	      {
		if ((mask & 0x01) == 0x01)
		  {
		    switch((grt_a[cnt].t_part = piD)) 
		      {

		      case TG_ZERO_ORD_PARRT_ID:
			pix_set_grating(inp_p->zero_order_grating); 
			pix_set_gzo(zo_fc); 
			compute_rd(evt_p, &grt_a[cnt], inp_p, err_p); 
			grt_a[cnt].t_mlamda = 99; 
			mask = 0; 
			break; 

		      case TG_HEG_PARRT_ID:
			pix_set_grating("HEG"); 
			pix_set_gzo(zo_fc); 
			compute_rd(evt_p, &grt_a[cnt], inp_p, err_p); 
			/* mlamda = Period * r	where r is in radians 
			 * since t_gac_rd is in degrees need to convert to 
			 * radians using  pi/180   
			 */  
			grt_a[cnt].t_mlamda = inp_p->TG_Period_HEG * 
                            sin ( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
			break; 
		      case TG_MEG_PARRT_ID:
			pix_set_grating("MEG"); 
			pix_set_gzo(zo_fc); 
			compute_rd(evt_p, &grt_a[cnt], inp_p, err_p); 
			/* mlamda = Period * r	where r is in radians 
			 * since t_gac_rd is in degrees need to convert to 
			 * radians using  pi/180   
			 */  
			grt_a[cnt].t_mlamda = inp_p->TG_Period_MEG * 
                            sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
			break; 
		      case TG_LETG_PARRT_ID:
			pix_set_grating("LEG"); 
			pix_set_gzo(zo_fc); 
			compute_rd(evt_p, &grt_a[cnt], inp_p, err_p); 
			/* mlamda = Period * r	where r is in radians 
			 * since t_gac_rd is in degrees need to convert to 
			 * radians using  pi/180   
			 */  
                        /* J. Davis' patch from message
                         * 200211160032.gAG0W1Ee003454@aluche.mit.edu */
			grt_a[cnt].t_mlamda = inp_p->TG_Period_LETG *
                            sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
			break; 
		      } /*end: switch (t_part) */

		    
		    /* reapply aspect to get updated sky coords */
		    if (asp_type_flag == ASP_FTYPE_OFFSETS)
		      {	 
			pix_apply_aspect(evt_p->det_work, asp_p->asp_sol, 
					 evt_p->sky_work); 
		      }
		    else
		      {
			double cel[2];
			
			dmTanPixToWorld( evt_p->det_work, asp_p->asp_sol,
                                         crpix, 
					 cdelt, cel );
			dmTanWorldToPix( cel, nominal_asp, crpix, cdelt, 
					 evt_p->sky_work );
		      }
		    
		    
		    grt_a[cnt].t_srcid = srcs[rrr].s_srcid;  /* from regionFile*/
		    grt_a[cnt].ccdid = evt_p->ccdid;     /* from infile */
		    /* keep track of number of sources */
		    cnt++;  

		  }  /* end: if ((mask & 0x01) == 0x01) */

		mask >>= 1;
		piD++;	

	    } /* end : while (mask > TG_NOT_IN_SRC) */


	    /* update source mask */
	    evt_p->e_smap |= (1 << rrr); 

	 }  /*----------- end : if (tags[rrr] != TG_NOT_IN_SRC) */

     }  /*--------END :  for (rrr=0; rrr < src_cnt; rrr++) ------- */

   /**************************************************************/
   if (cnt == 0) 
     {
       /* background */ 
       evt_p->e_gdp_xy[0] = evt_p->e_gdp_xy[1] = evt_p->e_srcid = evt_p->e_smap = 0; 
      evt_p->e_part = evt_p->e_order = 99; 
      ds_MAKE_DNAN(evt_p->e_gac_rd[0]); 
      ds_MAKE_DNAN(evt_p->e_gac_rd[1]); 
      ds_MAKE_DNAN(evt_p->e_mlamda); 
      ds_MAKE_DNAN(evt_p->e_lamda); 
   }
   else if (cnt == 1) 
   {
      /* single source */ 
      evt_p->e_gdp_xy[0] = grt_a[0].t_gdp_xy[0]; 
      evt_p->e_gdp_xy[1] = grt_a[0].t_gdp_xy[1]; 
      evt_p->e_gac_rd[0] = grt_a[0].t_gac_rd[0]; 
      evt_p->e_gac_rd[1] = grt_a[0].t_gac_rd[1]; 
      evt_p->e_srcid = grt_a[0].t_srcid;

      /* --------- determine order for ACIS --------- */ 
      if (grt_a[0].t_part != TG_ZERO_ORD_PARRT_ID)    /* !=0 */
      {
	 if (inp_p->instrume_type == TG_INSTR_ACIS)
	 {
	    short ndx; 
   
	    if (!ds_fNAN(evt_p->energy) && 
	       ((ndx = tg_order_resolve(rm_p, new_rm_p, rm_energy,
					rm_energy_lo, rm_energy_hi,
					energy_vec_col_size,
					inp_p, grt_a, cnt, evt_p->energy,
					evt_p->chip[TG_X_AXIS],
					evt_p->chip[TG_Y_AXIS],
					rm_type_flag, err_p)) != -1))
	    {
	       evt_p->e_order = grt_a[ndx].t_order;
	       evt_p->e_lamda = grt_a[ndx].t_lamda;
	       evt_p->e_mlamda = grt_a[ndx].t_mlamda;
	       evt_p->e_part = grt_a[ndx].t_part;
	    } 
	    else
	    {
	       /* unable to resolve orders- set to unresolved status */
	       evt_p->e_order = 99;
	       ds_MAKE_DNAN(evt_p->e_lamda); 
	       evt_p->e_mlamda = grt_a[0].t_mlamda; 
	       evt_p->e_part = grt_a[0].t_part;
	    }
	 }  /* end :  TG_INSTR_ACIS */
	 else
	 {
	    /* HRC instrument - can't resolve orders */
	       if ((evt_p->e_lamda = evt_p->e_mlamda = grt_a[0].t_mlamda) > 0.0) 
	       {
		  evt_p->e_order = 1;
	       } 
	       else 
	       {
		  evt_p->e_lamda *= (evt_p->e_order = -1);
	       } 
	       evt_p->e_part = grt_a[0].t_part;
	 }  /* end :  TG_INSTR_HRC  */

      }  /* Grating part not 0th order (t_part !=TG_ZERO_ORD_PARRT_ID)*/

      /* --------- end : determine order for ACIS --- */ 

   } /* end :  else if (cnt == 1)  */

   else       /***********  cnt > 1 ********** */
   {
      /* multiple sources- resolve for acis instrument only */ 
      if (inp_p->instrume_type == TG_INSTR_ACIS) 
      {
	 short ndx; 

	 if (!ds_fNAN(evt_p->energy) && 
	    ((ndx = tg_order_resolve(rm_p, new_rm_p, rm_energy,
				     rm_energy_lo, rm_energy_hi,
				     energy_vec_col_size,
				     inp_p, grt_a, cnt, 
				     evt_p->energy, evt_p->chip[TG_X_AXIS],
				     evt_p->chip[TG_Y_AXIS],
				     rm_type_flag, err_p))!= -1))
	 {
	    evt_p->e_part = grt_a[ndx].t_part;
	    evt_p->e_order = grt_a[ndx].t_order;
	    evt_p->e_lamda = grt_a[ndx].t_lamda;
	    evt_p->e_mlamda = grt_a[ndx].t_mlamda;
	    evt_p->e_gdp_xy[0] = grt_a[ndx].t_gdp_xy[0];
	    evt_p->e_gdp_xy[1] = grt_a[ndx].t_gdp_xy[1];
	    evt_p->e_gac_rd[0] = grt_a[ndx].t_gac_rd[0];
	    evt_p->e_gac_rd[1] = grt_a[ndx].t_gac_rd[1];
	    evt_p->e_srcid = grt_a[ndx].t_srcid;
	    evt_p->e_smap &= 1 << ndx; 
	 }
	 else
	 {
	    /* unable to resolve orders- set to unresolved status */
	    evt_p->e_part = evt_p->e_order = 99;
	    evt_p->e_srcid = 0;  
	    ds_MAKE_DNAN(evt_p->e_lamda); 
	 }
      }	 /* end : if instrume_type is  TG_INSTR_ACIS */  
   }  /* end : if ( cnt > 1 )  */

} /* end: tg_process_event_data */


/* ------------------------------------------------------------------------- */


/* this is a copy of tg_process_event_data() function, with CC mode
   additions.  NOTE:  when updating this function, please update
   tg_process_event_data() as well, since there are similarities 
   in the processing.  this copy was made in order to save time 
   in non-cc mode processing */

void tg_process_event_data_CC(
        SOURCE_ENTRY_A_T          srcs,   /*I - src list */  
        SRC_TAGS_A_T              tags,   /*I - track which srcs an evt is in */
        RSP_MTRX_REC_P_T          rm_p,   /*I - energy limits struct. ptr */ 
        NEW_RSP_MTRX_REC_P_T  new_rm_p,   /*I - new format energy limits struct.ptr */
        float              **rm_energy,   /*I - energy data array for NEW rmf table */
        float           **rm_energy_lo,   /*I - energy low data array for NEW rmf table */
        float           **rm_energy_hi,   /*I - energy hi data array for NEW rmf table */
        long       energy_vec_col_size,   /*I - dim of energy vector for NEW rmf table */
        INPUT_PARMS_P_T          inp_p,   /*I - input parameter structure ptr */
        ASPECT_REC_P_T           asp_p,   /*I - aspect record structure ptr */ 
        ALIGNMENT_INFRA_P_T   aln_hk_p,   /*I - alignment housekeeping ptr */ 
        ALIGNMENT_REC_P_T        aln_p,   /*I - alignment record structure ptr */
        short            asp_type_flag,   /*I - type of aspect correction */
        EVENT_REC_P_T            evt_p,   /*I/O - evt record structure ptr */
        enum RMFileType   rm_type_flag,
        dsErrList*                err_p   /*O - error list ptr */
    )   

{   
   VEC2_DBLE zo_fpc = {0.0, 0.0}; 
   VEC2_DBLE fpc_2d, zo_chip, cc_chip;
   VEC3_DBLE zo_fc	 = {0.0, 0.0}; 
   short rrr = 0, cnt = 0, chipid;
   GRATING_EVT_T grt_a[MAX_NUM_SOURCES*2]; /* assume heg & meg per src */ 
   boolean   zero_order=NO;
   short piD = 0; 
   SRC_TAGS_T mask = tags[rrr];
   double meg_chipy_work;
   short ndx=0;


   /* the following are needed for applying aspect correction from the 
      aspect solution */
   static double nominal_asp[3];
   static double crpix[2];
   static double cdelt[2];
   static short first_run = 1;
   static double chp_targ[2] = { 512.0, 512.0 };
   static double targ_crds[2], sky_xy[2];
   static short ccd_targ;
   static double noms[3];   

   static int ppp5 = 0 ;    /* 6/10/2015 - debug, infile's row# */
   ppp5 = ppp5 + 1 ;        /* row # in input and output evtfile */

   if ( first_run )
     {	
       
       nominal_asp[0] = inp_p->kywd_nom_ra;
       nominal_asp[1] = inp_p->kywd_nom_dec;
       nominal_asp[2] = 0.0;
       
       pix_get_fp_image_cntr(crpix);
       
       cdelt[1] = inp_p->fp_scale/3600.0;  
       cdelt[0] = -cdelt[1];   



       inp_p->crval[0] = inp_p->kywd_nom_ra;
       inp_p->crval[1] = inp_p->kywd_nom_dec;
       
       inp_p->crpix[0] = crpix[0];
       inp_p->crpix[1] = crpix[1];
       inp_p->cdelt[0] = cdelt[0];
       inp_p->cdelt[1] = cdelt[1];

       /* Determine nom target position */
       targ_crds[0] = inp_p->kywd_targ_ra;
       targ_crds[1] = inp_p->kywd_targ_dec;
       dmCoordInvert_d( inp_p->sky_desc, targ_crds, sky_xy );

       noms[0] = inp_p->crval[0]; 
       noms[1] = inp_p->crval[1];
       noms[2] = inp_p->kywd_nom_roll;
       
       if (asp_type_flag == ASP_FTYPE_OFFSETS)
	 {
	   pix_deapply_aspect(sky_xy, noms, fpc_2d);
	 }
       else
	 {
	   double cel[2];
	   dmTanPixToWorld(sky_xy, inp_p->crval, inp_p->crpix,
			   inp_p->cdelt, cel );
	   dmTanWorldToPix(cel, noms, inp_p->crpix,
			   inp_p->cdelt, fpc_2d);
	   
	 }
       /* pix_fpc_to_chip( fpc_2d, &ccd_targ, chp_targ ); */
       pix_fpc_to_extended_chip( fpc_2d, ccd_targ, chp_targ );
       
       first_run = 0;
     }
   
   
   /* update sim alignment if necessary */
   /* add time offset to event time for dither sequence */
   evt_p->time += inp_p->time_offset;
   
   if (alignment_update(aln_p, aln_hk_p, evt_p->time))
     {
       /* error- eventually should pass back to calling routine */
     }
   /* restore original time after dither sequence */
   evt_p->time -= inp_p->time_offset;
   



   /* Find a working chip coordinate */
   if (asp_type_flag == ASP_FTYPE_OFFSETS) {
       pix_deapply_aspect(sky_xy, asp_p->asp_sol, fpc_2d);
   }
   else {
       double cel[2];
       
       /* Map this chip center to fpc */
       pix_chip_to_fpc(evt_p->ccdid, chp_targ, fpc_2d);
       
       /* Map this chip center to sky at nominal aspect */
       noms[0] = inp_p->crval[0]; 
       noms[1] = inp_p->crval[1];
       noms[2] = inp_p->kywd_nom_roll;;
       
       dmTanPixToWorld(fpc_2d, noms, inp_p->crpix, inp_p->cdelt, cel );
       /* Map sky to fpc at THIS aspect */
       dmTanWorldToPix(cel, asp_p->asp_sol, inp_p->crpix,
		       inp_p->cdelt, fpc_2d);
       
     }
   
   pix_fpc_to_extended_chip( fpc_2d, evt_p->ccdid, cc_chip );
   chipid = evt_p->ccdid;
   evt_p->chip_work[TG_Y_AXIS] = cc_chip[TG_Y_AXIS];

   
   memset(grt_a, 0, (MAX_NUM_SOURCES*2*sizeof(GRATING_EVT_T))); 
   
   if (tags[rrr] != TG_NOT_IN_SRC) 
     {

       /* aspect correct to convert from sky to fpc */ 
       
       if ( asp_type_flag == ASP_FTYPE_OFFSETS )
	 {
	   pix_deapply_aspect(srcs[rrr].s_center, asp_p->asp_sol, fpc_2d); 
	 }
       else
	 {
	   double cel[2];
	   
	   dmTanPixToWorld( srcs[rrr].s_center, nominal_asp, crpix, cdelt, cel );
	   dmTanWorldToPix( cel, asp_p->asp_sol, crpix, cdelt, zo_fpc );
	 }
       
       /* Get 0th order chip positon */
       pix_fpc_to_fc(zo_fpc, zo_fc); 
       /* pix_fpc_to_chip( zo_fpc, &chipid, zo_chip ); */
       pix_fpc_to_extended_chip( zo_fpc, chipid, zo_chip );

       evt_p->chipy_zo = zo_chip[ TG_Y_AXIS ] ;

       evt_p->chipy_zo_new = evt_p->chipy_zo ;     /* 6/30/2015 - add ; outCol chipy_zo;*/
	   
       /* compute gratings coords in pixels and angles */ 
       /***
	 pix_chipid_to_chip(inp_p->tdetsys, evt_p->chipid, evt_p->chipname);
	 ***/
	   
       while (mask > TG_NOT_IN_SRC)
	 {

	   if ((mask & 0x01) == 0x01)
	     {
	       grt_a[cnt].t_srcid = srcs[rrr].s_srcid; 
	       grt_a[cnt].ccdid = evt_p->ccdid;
	       
	       switch((grt_a[cnt].t_part = piD)) 
		 {
		 case TG_ZERO_ORD_PARRT_ID:

             if (inp_p->hetgCC_2ndRun!=1)    /* 6/30/2015 - add : IF */
             {
                   /* CC 1st pass */

		   evt_p->e_part=TG_ZERO_ORD_PARRT_ID;
		   pix_set_grating(inp_p->zero_order_grating);
		   pix_set_gzo(zo_fc);
		   tg_cc_calculate_chip_y(srcs, tags, rm_p,
					  inp_p, asp_p, aln_hk_p, aln_p,
                                          evt_p, zo_chip, chipid, err_p);
		   compute_rd(evt_p, &grt_a[cnt], inp_p, err_p);
		   grt_a[cnt].t_mlamda = 99;
		   mask = 0;
		   zero_order=YES;
             }
             else 
             {
                 /* 6/30/2015 - add */
                 /* for CC 2nd pass : copy the input info. */
                 grt_a[ cnt ].t_gdp_xy[0] = evt_p->e_gdp_xy[0] ; /* sky(x,y)*/
                 grt_a[ cnt ].t_gdp_xy[1] = evt_p->e_gdp_xy[1] ; /* sky(x,y)*/
                 grt_a[ cnt ].t_gac_rd[0] = evt_p->e_gac_rd[0] ; /* rd(tg_r,tg_d)*/
                 grt_a[ cnt ].t_gac_rd[1] = evt_p->e_gac_rd[1] ; /* rd(tg_r,tg_d)*/
                 grt_a[ cnt ].t_mlamda = evt_p->e_mlamda ;
                 mask = 0;
                 /*  evt_p->chip_work_new (ie. outCol: chipy_tg)  set earlier from input file */
             }

		   break; 

		 case TG_MEG_PARRT_ID:
		   {
		     evt_p->e_part=TG_MEG_PARRT_ID;
                     pix_set_grating("MEG"); 
                     pix_set_gzo(zo_fc); 
		     tg_cc_calculate_chip_y(srcs, tags, rm_p, 
					    inp_p, asp_p, aln_hk_p, aln_p,
                                            evt_p, zo_chip, chipid, err_p);
		     /*Check to see if the predicted chip y is on chip.
		      * If (On Chip)
		      *     -compute the tg coordinates assuming the event
                      *     belongs to MEG. Then resolve the order.
		      *	   If (order is odd)
		      *	       -done
		      *	   Else if (order is even)
		      *	       -assume the event belongs to HEG, then calculate
		      *		the predicted chip y.  Check if the predicted
		      *		chip y is on chip.
		      *		If (On Chip):
		      *		   -compute the tg coordinates assuming
                      *             the event belongs to HEG.
		      *		Else If (not on Chip):
		      *		   -leave the event as MEG even order.	Use the
		      *		    tg coordinates, and order determined in an
		      *		    earlier step.
		      * Else If (not on chip)
		      *	     -compute the tg coordinates assuming the event
                      *       belongs to HEG.
                      */
		     
		     if (( evt_p->chip_work[TG_Y_AXIS] <=
                           inp_p->chip_maxrange[TG_Y_AXIS] ) &&
			 ( evt_p->chip_work[TG_Y_AXIS] >=
                           inp_p->chip_minrange ))
                     {
                         compute_rd(evt_p, &grt_a[cnt], inp_p, err_p); 
                         /* mlamda = Period * r	 where r is in radians 
                          * since t_gac_rd is in degrees need to convert to 
                          * radians using  pi/180   
                          */
                         grt_a[cnt].t_mlamda = inp_p->TG_Period_MEG * 
                             sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
                         
                         /* using (cnt+1), since cnt is not increased
                            for CC mode, until order is resolve, and
                            grating arm is chosen */
                         if (!ds_fNAN(evt_p->energy) && 
                             (( ndx =
                                tg_order_resolve( rm_p, new_rm_p, rm_energy,
                                                  rm_energy_lo, rm_energy_hi,
                                                  energy_vec_col_size,
                                                  inp_p, grt_a, cnt+1, 
                                                  evt_p->energy,
                                                  evt_p->chip[TG_X_AXIS],
                                                  evt_p->chip_work[TG_Y_AXIS],
                                                  rm_type_flag, err_p)) != -1))
                         {
                             evt_p->e_order = grt_a[0].t_order;
                             evt_p->e_lamda = grt_a[0].t_lamda;
                             evt_p->e_mlamda = grt_a[0].t_mlamda;
                             evt_p->e_part = grt_a[0].t_part;
                         }
                         else
                         {
			     /* Now let's try HEG */
			     meg_chipy_work=evt_p->chip_work[TG_Y_AXIS];
                             /* change to HEG */
			     grt_a[cnt].t_part=TG_HEG_PARRT_ID;  
                             /* change to HEG */
			     evt_p->e_part=TG_HEG_PARRT_ID;  
                             pix_set_grating("HEG");
                             pix_set_gzo(zo_fc);
			     tg_cc_calculate_chip_y( srcs, tags, rm_p,
						     inp_p, asp_p,
                                                     aln_hk_p, aln_p,
						     evt_p, zo_chip,
                                                     chipid, err_p);
			     
			     if ( (evt_p->chip_work[TG_Y_AXIS] <=
                                   inp_p->chip_maxrange[TG_Y_AXIS] ) &&
				  (evt_p->chip_work[TG_Y_AXIS] >=
                                   inp_p->chip_minrange )   )
			       {
				 compute_rd(evt_p, &grt_a[cnt], inp_p, err_p);
				 grt_a[cnt].t_mlamda = inp_p->TG_Period_HEG *
                                     sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
				 if (!ds_fNAN(evt_p->energy) && 
				     (( ndx =
                                        tg_order_resolve(rm_p, new_rm_p,
                                                         rm_energy,
                                                         rm_energy_lo,
                                                         rm_energy_hi,
                                                         energy_vec_col_size,
                                                         inp_p, grt_a, cnt+1, 
                                                         evt_p->energy,
                                                         evt_p->chip[TG_X_AXIS],
                                                         evt_p->chip_work[TG_Y_AXIS],
                                                         rm_type_flag, err_p)) != -1))
				   {
				     evt_p->e_order = grt_a[0].t_order;
				     evt_p->e_lamda = grt_a[0].t_lamda;
				     evt_p->e_mlamda = grt_a[0].t_mlamda;
				     evt_p->e_part = grt_a[0].t_part;
				   }				 
				 else
				   {
				     pix_set_grating("MEG"); 
				     pix_set_gzo(zo_fc); 
				     evt_p->chip_work[TG_Y_AXIS] =
                                         meg_chipy_work;
				     compute_rd( evt_p, &grt_a[cnt],
                                                 inp_p, err_p); 
				     grt_a[cnt].t_mlamda = inp_p->TG_Period_MEG *
                                         sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
				     evt_p->e_order = 99;
				     ds_MAKE_DNAN(evt_p->e_lamda); 
				     evt_p->e_mlamda = grt_a[0].t_mlamda;  
				     grt_a[cnt].t_part=TG_MEG_PARRT_ID;
				     evt_p->e_part=TG_MEG_PARRT_ID;
				   }
			       } /* else HEG not on chip */
			     else
			       {
				 evt_p->e_part=TG_MEG_PARRT_ID;
				 /* unable to resolve orders- set
                                  * to unresolved status */
				 pix_set_grating("MEG"); 
				 pix_set_gzo(zo_fc); 
				 tg_cc_calculate_chip_y( srcs, tags, rm_p, 
                                                         inp_p, asp_p,
                                                         aln_hk_p, aln_p,
                                                         evt_p, zo_chip,
                                                         chipid, err_p );
				 compute_rd(evt_p, &grt_a[cnt], inp_p, err_p); 
				 grt_a[cnt].t_mlamda = inp_p->TG_Period_MEG * 
                                     sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
				 evt_p->e_order = 99;
				 ds_MAKE_DNAN(evt_p->e_lamda); 
				 evt_p->e_mlamda = grt_a[0].t_mlamda;  
				 grt_a[cnt].t_part=TG_MEG_PARRT_ID;
				 evt_p->e_part=TG_MEG_PARRT_ID;
			       } 
			   } /* end else try HEG part */

			 if (( evt_p->e_order % 2 == 0) &&
                             /* when MEG even order */
			     ( evt_p->e_part == TG_MEG_PARRT_ID) )
			   {
			     /* store the previously calculated 
			      * evt_p->chip_work[TG_Y_AXIS] to meg_chipy_work,
			      * since we may need it if the event is finally
			      * determined to belong to MEG */
			     meg_chipy_work=evt_p->chip_work[TG_Y_AXIS];
                             /* change to HEG */
			     grt_a[cnt].t_part=TG_HEG_PARRT_ID;
                             /* change to HEG */
			     evt_p->e_part=TG_HEG_PARRT_ID;
                             pix_set_grating("HEG");
                             pix_set_gzo(zo_fc);
			     tg_cc_calculate_chip_y( srcs, tags, rm_p,
                                                     inp_p, asp_p,
                                                     aln_hk_p, aln_p,
                                                     evt_p, zo_chip,
                                                     chipid, err_p);
			     
			     if ( (evt_p->chip_work[TG_Y_AXIS] <=
                                   inp_p->chip_maxrange[TG_Y_AXIS] ) &&
				  (evt_p->chip_work[TG_Y_AXIS] >=
                                   inp_p->chip_minrange ) )
			       {
				 compute_rd(evt_p, &grt_a[cnt], inp_p, err_p);
				 grt_a[cnt].t_mlamda = inp_p->TG_Period_HEG *
                                     sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
				 
			       }
			     else
			       {
				 /* Use the tg_r, tg_d and order that we have 
				  * calculated for MEG. Set grating part back
				  * to TG_MEG_PARRT_ID. */
				 evt_p->chip_work[TG_Y_AXIS]=meg_chipy_work;
				 grt_a[cnt].t_part=TG_MEG_PARRT_ID;
				 evt_p->e_part=TG_MEG_PARRT_ID;
				 pix_set_grating("MEG");
				 pix_set_gzo(zo_fc);
				 compute_rd(evt_p, &grt_a[cnt], inp_p, err_p);
				 grt_a[cnt].t_mlamda = inp_p->TG_Period_MEG *
				   grt_a[cnt].t_gac_rd[0] * DEG2RAD;
			       }
			   } /* End if (evt_p->e_order % 2 == 0) */
		       }
		     /* End if ((evt_p->chip_work[TG_Y_AXIS] -
                      *          inp_p->chip_minrange) > -DBL_EPSILON) &&
		      *	       ((inp_p->chip_maxrange[1] -
                      *		 evt_p->chip_work[TG_Y_AXIS]) > -DBL_EPSILON)
		      */
		     else
		       {
			 grt_a[cnt].t_part=TG_HEG_PARRT_ID;
			 evt_p->e_part=TG_HEG_PARRT_ID;
			 pix_set_grating("HEG");
			 pix_set_gzo(zo_fc);
			 tg_cc_calculate_chip_y( srcs, tags, rm_p,
                                                 inp_p, asp_p,
                                                 aln_hk_p, aln_p, evt_p,
                                                 zo_chip, chipid, err_p );
			 compute_rd(evt_p, &grt_a[cnt], inp_p, err_p);
			 grt_a[cnt].t_mlamda = inp_p->TG_Period_HEG *
                             sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
		       }
		     break;
		   }
		 case TG_LETG_PARRT_ID:

                    /* 6/30/2015 Note - not for CC 2nd pass, ie. inp_p->hetgCC_2ndRun!=1*/

		    evt_p->e_part = TG_LETG_PARRT_ID;
		    pix_set_grating( "LEG" );
		    pix_set_gzo( zo_fc );
                    tg_cc_calculate_chip_y( srcs, tags, rm_p,
                                            inp_p, asp_p, aln_hk_p, aln_p,
                                            evt_p, zo_chip, chipid, err_p );
		    compute_rd( evt_p, &grt_a[cnt], inp_p, err_p );
		    grt_a[cnt].t_mlamda = inp_p->TG_Period_LETG *
                        sin( grt_a[cnt].t_gac_rd[0] * DEG2RAD );
		    break;
		 } /* end switch */
	       
	       /* keep track of number of sources */
	       cnt++;	
	       
	       /* reapply aspect to get updated sky coords */
	       if (asp_type_flag == ASP_FTYPE_OFFSETS)
		 {
		   pix_apply_aspect(evt_p->det_work, asp_p->asp_sol, 
				    evt_p->sky_work); 
		 }
	       else
		 {
		   double cel[2];
		   
		   dmTanPixToWorld( evt_p->det_work, asp_p->asp_sol, crpix, 
				    cdelt, cel );
		   dmTanWorldToPix( cel, nominal_asp, crpix, cdelt, 
				    evt_p->sky_work );
		 }
	       
	       
	       
	     }	/* end of if ((mask & 0x01) == 0x01) */

	   mask >>= 1;
	   piD++;
	 }  /* end while loop (mask > TG_NOT_IN_SRC) */
       
       /* update source mask (smap column) */
       evt_p->e_smap |= (1 << rrr); 
     }	/* end of if (tags[rrr]) loop */
   

   /* 8/24/2015 - set the final gti flag, including pcad time range, is sky nan ?
    *             ( rename set_gti_flag() to  set_pcad_sky_flag()  )  
    *             ( rename is_in_gti      to  pcad_sky_flag )
    */             
   /* 6/29/2015 - set the final gti flag, including flt gti per ccdid, pcad time range, is sky nan ?
    */
   bool pcad_sky_flag = set_pcad_sky_flag ( inp_p, evt_p, ppp5 ) ;


   if (cnt == 0) 
   {
       /* background */ 

     if ((inp_p->hetgCC_2ndRun!=1)||                /* 6/30/2015 - add IF */
        ((inp_p->hetgCC_2ndRun==1 )&&(pcad_sky_flag)))
     {
        /* 6/30/2015:
         *  For 1st pass, regardless it's in gti or not, set the values.
         *  For 2nd pass, set the values only if it's in gti.
         *                if not in gti, copy values from infile, ie. do nothing here. */

       evt_p->e_gdp_xy[0] = evt_p->e_gdp_xy[1] = evt_p->e_srcid = evt_p->e_smap = 0; 
       evt_p->e_part = evt_p->e_order = 99; 
       ds_MAKE_DNAN(evt_p->e_gac_rd[0]); 
       ds_MAKE_DNAN(evt_p->e_gac_rd[1]); 
       ds_MAKE_DNAN(evt_p->e_mlamda); 
       ds_MAKE_DNAN(evt_p->e_lamda); 
     }
   /*else {do nothing;} */


       /* 6/10/2015 - for 1st pass & 2nd pass, overwrite some values */
       /* 11/13/2015-change tnull to nan for chipy_zo, chipy_tg; don't set tnull or nan for tdet;*/
       if (!pcad_sky_flag)       /* FALSE */
       {
          ds_MAKE_DNAN( evt_p->chipy_zo );             /* chipy_zo */
          ds_MAKE_DNAN( evt_p->chip_work[TG_Y_AXIS] ); /* chipy_zo */

          ds_MAKE_DNAN( evt_p->chipy_zo_new );         /* cc 2nd pass: outCol chipy_zo*/
          ds_MAKE_DNAN( evt_p->chip_work_new );        /* cc 2nd pass: outCol chipy_tg*/

          evt_p->e_order = 99 ;    /* 6/30/2015 - add */


          /* 6/17/2015 - add */
          if ( inp_p->grating_flag == 1 )      /* GRATING kVal is HETG */
             evt_p->e_part = TG_MEG_PARRT_ID ; /* 2  */
          else if ( inp_p->grating_flag == 2 ) /* GRATING kVal is LETG */
             evt_p->e_part = TG_LETG_PARRT_ID ; /* 3 */
       }
   }
   else if (cnt == 1)
   {
       /* single source */ 
       evt_p->e_gdp_xy[0] = grt_a[0].t_gdp_xy[0]; 
       evt_p->e_gdp_xy[1] = grt_a[0].t_gdp_xy[1]; 
       evt_p->e_gac_rd[0] = grt_a[0].t_gac_rd[0]; 
       evt_p->e_gac_rd[1] = grt_a[0].t_gac_rd[1]; 
       evt_p->e_srcid = grt_a[0].t_srcid;

       /* && (grt_a[cnt-1].t_part != TG_MEG_PARRT_ID))  *//* already done MEG */
       /* determine order if acis */ 
       if (grt_a[0].t_part != TG_ZERO_ORD_PARRT_ID)
       {
	   if (!ds_fNAN(evt_p->energy) &&  (evt_p->e_order != 99 ) &&
	       ((ndx = tg_order_resolve(rm_p, new_rm_p, rm_energy, 
					rm_energy_lo, rm_energy_hi,
					energy_vec_col_size,
					inp_p, grt_a, cnt, 
					evt_p->energy, 
					evt_p->chip[TG_X_AXIS],
					evt_p->chip_work[TG_Y_AXIS],
					rm_type_flag, err_p)) != -1))
           {
	       evt_p->e_order = grt_a[ndx].t_order;
	       evt_p->e_lamda = grt_a[ndx].t_lamda;
	       evt_p->e_mlamda = grt_a[ndx].t_mlamda;
	       evt_p->e_part = grt_a[ndx].t_part;
           } 
	   else
	   {
	       /* unable to resolve orders- set to unresolved status */
	       evt_p->e_order = 99;
	       ds_MAKE_DNAN(evt_p->e_lamda); 
	       evt_p->e_mlamda = grt_a[0].t_mlamda;
	       evt_p->e_part = grt_a[0].t_part;
	   }
      } /* end if (grt_a[0].t_part != TG_ZERO_ORD_PARRT_ID) */


      /* 6/10/2015 - add this at the end of the condition 'if (cnt == 1)' :
       *   
       * if ( evtTime is not in flt or evt gti, or is not in pcad time range,  )
       * then set both chipy_zo, chipy_tg to tnull
       */
      if (!pcad_sky_flag)        /* FALSE */
      {
         /*11/13/2015-change tnull to nan for chipy_zo, chipy_tg; don't set tnull or nan for tdet;*/
         ds_MAKE_DNAN( evt_p->chipy_zo );             /* chipy_zo */
         ds_MAKE_DNAN( evt_p->chip_work[TG_Y_AXIS] ); /* chipy_zo */

         ds_MAKE_DNAN( evt_p->chipy_zo_new );         /* cc 2nd pass: outCol chipy_zo*/
         ds_MAKE_DNAN( evt_p->chip_work_new );        /* cc 2nd pass: outCol chipy_tg*/

         if ( inp_p->grating_flag == 1 )      /* GRATING kVal is HETG */
            evt_p->e_part = TG_MEG_PARRT_ID ; /* 2  */
         else if ( inp_p->grating_flag == 2 ) /* GRATING kVal is LETG */
            evt_p->e_part = TG_LETG_PARRT_ID ; /* 3 */

         evt_p->e_order = 99;
      }  

   } /* end : (cnt == 1) */

} /* end: tg_process_event_data_CC */

/* ------------------------------------------------------------------------- */

void tg_cc_calculate_chip_y (
    SOURCE_ENTRY_A_T	srcs,	  /* I   - source list			     */
    SRC_TAGS_A_T	tags,	  /* I   - which sources an event is in	     */
    RSP_MTRX_REC_P_T	rm_p,	  /* I   - energy limits structure pointer   */
    INPUT_PARMS_P_T	inp_p,	  /* I   - input parameter structure pointer */
    ASPECT_REC_P_T	asp_p,	  /* I   - aspect record structure pointer   */
    ALIGNMENT_INFRA_P_T	aln_hk_p, /* I   - alignment housekeeping pointer    */
    ALIGNMENT_REC_P_T	aln_p,	  /* I   - alignment record structure ptr    */
    EVENT_REC_P_T	evt_p,	  /* I/O - event record structure pointer    */
    VEC2_DBLE		zo_chip,  /* I   - zero order chip position	     */
    short		zo_ccdid, /* I   - zero order ccdid		     */
    dsErrList*		err_p )	  /* O   - error list pointer		     */
{   

   /**********************************************************************
    * NOTE: pix_set_grating() and pix_set_gzo() MUST be called before
    * this routine is called.  The small angle approximation correction
    * at the end depends upon this.
    *********************************************************************/

   short  delta_ccdid = 0;
   double gap = 0.0;
   double ccd_size = (double) DEFAULT_ACIS_S_CCD_SIZE;
   double alpha = 0;

   double min_tgd = 0.0;
   short  min_chipy = 0;
   double fc[3];
   double gzo[3];
   double rd[3];
    
   short ll = 0;

   /*  How to calculate the new chip_y; an email from Dave H.:
    *  The predicted CHIPY is:
    *
    * pCHIPY = CY0 + [CHIPX + Delta_CCDID * (gap+size) - CX0] * tan(alpha)
    *	
    * Where:
    *
    * pCHIPY = predicted CHIPY
    *
    * CX0 = chip x coordinate of zero-order centroid.  This is a
    *	    real-valued coordinate and is obtained from the
    *	    transformation of the sky X centroid to chip at a specific
    *	    time. (BUT: see notes below. (*) )
    *
    * CY0 = chip y coordinate of zero-order centroid (see CX0) (*)
    *
    * CHIPX = observed event CHIPX
    *
    * Delta_CCDID = difference in event CCDID and zero-order CCDID:
    * Delta_CCDID = CCDID - CCDID_ZO.
    *		    (note: this is <0 for negative orders).
    *
    * gap = size of gap between CCDs, in real pixel units; 
    * gap = gapmm/pixsize = 0.431mm /0.024 mm/pix).
    *
    * size = size of CCD in pixels (1024).
    *
    * alpha = grating clocking angle (about 5 (MEG) or -5 (HEG) degrees.
    */


   /* angles are revered in chip coords from sky coords */
   switch ( evt_p->e_part ) {
       case TG_LETG_PARRT_ID:
           alpha = inp_p->TG_Angle_LEG;	break;
           
       case TG_MEG_PARRT_ID:
           alpha = inp_p->TG_Angle_MEG;	break;
           
       case TG_HEG_PARRT_ID:
           alpha = inp_p->TG_Angle_HEG;	break;
           
       default:	/* eg, TG_ZERO_ORD_PARRT_ID */
           alpha = 0.0;				break;
   }
   /* convert from degrees to radians (for tangent) */
   alpha *= DEG2RAD;
   
   /* for now, use gap = 17.9553 pixels; use PIXLIB later */
   gap = 17.9553;
   
   /* calculate the delta ccdid between event and zero order */
   delta_ccdid = fabs( (double) ( evt_p->ccdid - zo_ccdid ));

   /* calculate the predicted chip_y for the event
    * This is only an approximation -- it assumes chip_y is parallel
    * with sim_y which is not true
    */
   if (( evt_p->ccdid - zo_ccdid) >= 0 )
   {
     evt_p->chip_work[TG_Y_AXIS] =
         zo_chip[TG_Y_AXIS] + ( evt_p->chip_work[TG_X_AXIS] -
                                zo_chip[TG_X_AXIS] +
                                delta_ccdid * (gap+ccd_size) ) *
         			tan( alpha * -1.0 );
   }
   else
   {
     evt_p->chip_work[TG_Y_AXIS] =
         zo_chip[TG_Y_AXIS] - ( zo_chip[TG_X_AXIS] -
                                evt_p->chip_work[TG_X_AXIS] +
                                delta_ccdid * (gap+ccd_size) ) *
         			tan( alpha * -1.0 );
   }

   evt_p->chip_work_new = evt_p->chip_work[TG_Y_AXIS] ;  /*6/30/3015-add; outCol chipy_tg*/


   /* is chip_y off chip */
   if (( evt_p->chip_work[TG_Y_AXIS] >= inp_p->chip_maxrange[TG_Y_AXIS] ) ||
       ( evt_p->chip_work[TG_Y_AXIS] <= inp_p->chip_minrange )) {
       /* MEG/HEG switching part, let it re-run w/ different arm */
       return;
   }
   

   /*
    * Okay, now we have a guess at chip_y.  The real question is, what
    * chip_y gives us the minimum tg_d.  This assumes that all the
    * events are diffracted, which while isn't true; is the best guess
    * we can make to get the truely diffracted events.
    *
    * So, starting at the guessed value above, we step 1 pixel in the
    * + and - chip_y directions searching for the minimum.
    *
    * Despite chip_y being a floating point value, we're only using
    * integer chip_y's so even the guess above gets floor'ed.  There
    * could be a discussion of randomization of the chip coordinates
    * at this point, but given all the guessing going on I'm not sure
    * aliasing is going to be the dominant error.
    *
    * Need to be sure that pix_set_grating() and pix_set_gzo() are
    * called before doing this.  Could call compute_rd(), but may need
    * to pass in more info, besides compute_rd() also computes gacs'
    * which we're not interested in at this point.
    */

   /* do this to integer clip value */
   min_chipy =
       evt_p->chip_work[TG_Y_AXIS] = floor( evt_p->chip_work[TG_Y_AXIS] );
   
   /* Find the tg_d for this initial guess */
   pix_chip_to_fc( evt_p->chipid, evt_p->chip_work, fc );
   pix_fc_to_gzo( fc, gzo );
   pix_gzo_to_gac( gzo, rd );
   min_tgd = fabs( rd[1] );

   /* go in +chip_y direction */
   for ( ll = min_chipy; ll <= 1024; ll++ ) {

       /* This code is identical to the loop immediately below */
       
       evt_p->chip_work[TG_Y_AXIS] = ll;

       /* calculate tg_d for chip_y = ll */
       pix_chip_to_fc( evt_p->chipid, evt_p->chip_work, fc );
       pix_fc_to_gzo( fc, gzo );
       pix_gzo_to_gac( gzo, rd );

       /* keep going until tg_d( chip_y = ll ) > tg_d( chip_y = ll+1 ) */
       if ( fabs(rd[1]) <= min_tgd ) {
           min_tgd = fabs( rd[1] );
           min_chipy = ll;
       } else
           break;
       
   }

   /* Once finished in the +chip_y, try back in the -chip_y direction */
   for ( ll = min_chipy; ll >= 1; ll-- ) {

       /* This code is identical to the loop immediately above */
       
       evt_p->chip_work[TG_Y_AXIS] = ll;
     
       pix_chip_to_fc( evt_p->chipid, evt_p->chip_work, fc );
       pix_fc_to_gzo( fc, gzo );
       pix_gzo_to_gac( gzo, rd );
     
       /* keep going until tg_d( chip_y = ll-1 ) > tg_d( chip_y = ll ) */
       if ( fabs(rd[1]) <= min_tgd ) {
           min_tgd = fabs(rd[1]);
           min_chipy = ll;
       } else
           break;
   }

   /* Finially, store the value */
   evt_p->chip_work[TG_Y_AXIS] = min_chipy;
   evt_p->chip_work_new = evt_p->chip_work[TG_Y_AXIS] ;  /*6/30/3015-add; outCol chipy_tg*/

} /* end: tg_cc_calculate_chip_y */
