/*                                                                
**  Copyright (C) 1998-2008,2012,2014-2015  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*************************************************************************

* FILE NAME: tg_setup_columns.c
 
* DESCRIPTION:

  The routine tg_setup_columns() is called upon by tg_resolve_events 
  to create the columns of the output event file. It is responsible for 
  obtaining the column descriptor as well as for setting the values of
  TLMIN/TLMAX/TUNITS keywords as specified in the level 1 tg to archive
  icd. (Macros defining these values are found in ds_tg_config.h).

  Tg_setup_columns calls upon the local function tg_write_column_range()
  to write out the TLMIN and TLMAX values to a column descriptor based
  on the dmTYPE of the column. It takes in the tlmin and tlmax values 
  as doubles, casts them to the appropriate data type, and calls the
  appropriate datamodel routine to update the descriptors. If a tlmin
  or tlmax value is outside the range of a short then SHRT_MIN or SHRT_MAX
  (as defined in limits.h) is used when the column is of type dmSHORT. 

* NOTES

  The routine does not pass a return status back, but detected errors are
  added to the maintained error list which is passed in as a parameter.  

  (2/2005) - correct tlmax for chipy
  (7/2007) - setup output columns : pha_ro, chipy_tg, chipy_zo.
     obsolete 8/2008 -> pha_ro won't be created if it doesn't exist in infile.
(8/2008)- to be consistent with other columns in eventdef, pha_ro and raw will
          be created even they don't exist in infile.
9/24/2012 - if the units of columns exist, copy it to output. bug 13374 ;

10/2014 - TG_SRC_MAP* -> TG_SMAP* ;  TG_RD -> TG_GAC ;
6/10/2015 - set tnull value (99999.0) for chipy_tg & chipy_zo cols.
6/17/2015 - set tnull values for tdetx, tdety
          - if ( tdet is in infile )
               if ( tnull is defined ) 
                  copy tnull to output ;
               else 
                  set to 9999 for acis ; set 99999 for hrc  ;
            else
               do nothing
           - if ( chipy_zo / chipy_tg col. is in infile )
                if ( tnull is defined ) 
                   copy tnull to output ;
                else 
                   set to 99999.0  ;
             else   
                set to 99999.0  ;
9/22/2015 - bug14186: remove the col. range for chipy_tg, chipy_zo
11/13/2015- bug14203: change tnull to nan for chipy_zo, chipy_tg ; don't set tnull or nan for tdet;
11/18/2015- bug14203: put tnull back to tdet;
*************************************************************************/

#include <ctype.h>
#include <limits.h>

#ifndef TG_RESOLVE_EVENTS_H
#include "tg_resolve_events.h"
#define TG_RESOLVE_EVENTS_H
#endif

#ifndef DSLIB_H
#include "dslib.h"
#define DSLIB_H
#endif
 
#ifndef DS_ACIS_CONFIG_H
#include "ds_acis_config.h"
#define DS_ACIS_CONFIG_H
#endif

#ifndef DS_HRC_CONFIG_H
#include "ds_hrc_config.h"
#define DS_HRC_CONFIG_H
#endif

#ifndef DS_GRATINGS_CONFIG_H
#include "ds_gratings_config.h"
#define DS_GRATINGS_CONFIG_H
#endif

/* ---------------------------------------------------------- */

static void tg_write_column_range(
   dmDescriptor*   desc,          /* I   - column descriptor            */
   dmDataType      type,          /* I   - column type                  */
   double          min,           /* I   - minimum value (TLMIN)        */
   double          max)           /* I   - maximum value (TLMAX)        */ 
{
   switch (type)
   {
      case dmSHORT:
      {
         short smin = (short) min,
               smax = (short) max;

         if ((min < SHRT_MIN) || (min > SHRT_MAX))
         {
            smin = SHRT_MIN;  /* avoid underflow */ 
         }
         if ((max < SHRT_MIN) || (max > SHRT_MAX))
         {
            smax = SHRT_MAX;  /* avoid overflow */ 
         }

         dmDescriptorSetRange_s(desc, smin, smax);
      } 
      break;
      case dmLONG:
         dmDescriptorSetRange_l(desc, (long) min, (long) max);
      break;
      case dmFLOAT:
         dmDescriptorSetRange_f(desc, (float) min, (float) max);
      break;
      case dmDOUBLE:
         dmDescriptorSetRange_d(desc, (double) min, (double) max);
      break;
      default:
         /* do not write out range */ 
      break;
   }

} /* end: tg_write_column_range */



/* ---------------------------------------------------------- */

void tg_setup_columns(
   EVENT_SETUP_P_T evtin_p,       /* I   - input file information       */
   EVENT_SETUP_P_T evtout_p,      /* I/O - output file information      */
   INPUT_PARMS_P_T inp_p,         /* I/O - input parameters/data        */
   dsErrList*      err_p)         /* O   - error list pointer           */

{ 
   register int rr;

   for (rr=0; rr < evtout_p->num_cols; rr++)
   {
      switch (evtout_p->mapping[rr])
      {
         case TG_SKY:    /* vectored column pair- X,Y */  
         {
            VEC2_LONG axlen;
            double crpix[2], crval[2], cdelt[2];
            char  names[2][DS_SZ_KEYWORD];
            char* cptnames[2];
            char* eqnames[] = {"RA", "DEC"};
            dmDescriptor* temp_desc;
	    double sky_x_min, sky_x_max, sky_y_min, sky_y_max;

            cptnames[0] = names[0];
            cptnames[1] = names[1];
            strcpy(cptnames[0], TG_SKY_X_COL);
            strcpy(cptnames[1], TG_SKY_Y_COL);
   
            /* if output column name is lowercase, set x/y to lc */
            if (islower(evtout_p->names[rr][0]))
            {
               string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
               string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
            }
   
            pix_get_fp_image_cntr(crpix);
            crval[0] = inp_p->kywd_nom_ra;
            crval[1] = inp_p->kywd_nom_dec;
            cdelt[0] = inp_p->fp_scale/-3600.0;
            cdelt[1] = cdelt[0]*-1;
            evtout_p->dim[rr] = 2;
            evtout_p->desc[rr] =
               dmColumnCreateVector(evtout_p->extension, evtout_p->names[rr],
                                    evtout_p->types[rr], 0, 
                                    DS_ACIS_SKY_TUNITS, "Sky coords",
                                    cptnames, evtout_p->dim[rr]);
            inp_p->sky_desc= dmCoordCreate_d(evtout_p->desc[rr], "EQPOS", 
					     "deg", eqnames, 2, "TAN", 
					     crpix, crval, cdelt,
					     NULL);
    
            /* set TLMIN/TLMAX for sky coords */
            pix_get_fp_image_dimn(axlen);

	    
	    sky_x_min = 0.5 ; sky_x_max = DS_HRCS_X_TLMAX;
	    sky_y_min = 0.5 ; sky_y_max = DS_HRCS_Y_TLMAX;

	    if ( dmGetCpt(evtout_p->desc[rr], 1) && 
		 dmGetCpt(evtout_p->desc[rr], 2 ) )
	      {
		register int cc;
		for (cc = evtin_p->num_cols; cc--;)
		  {
		    if (evtin_p->mapping[cc] == TG_SKY_X )
		      {
			dmDescriptorGetRange_d( evtin_p->desc[cc], &sky_x_min, 
						&sky_x_max );
		      }
		    if (evtin_p->mapping[cc] == TG_SKY_Y )
		      {
			dmDescriptorGetRange_d( evtin_p->desc[cc], &sky_y_min, 
						&sky_y_max );
			
		      }
		  }
	      }

            /* ------------ TG_SKY data type ----------- */
            switch (evtout_p->types[rr])
            {
               case dmLONG:
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_l(temp_desc, 
                                            sky_x_min, sky_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
		    dmDescriptorSetRange_l(temp_desc, 
					   sky_y_min, sky_y_max);
                  }
               break;

               case dmSHORT:
               {
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_s(temp_desc, 
					    sky_x_min, sky_x_max );

                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_s(temp_desc, 
					    sky_y_min, sky_y_max );
                  }
               }
               break;

               case dmFLOAT:
               {
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_f(temp_desc,
					    sky_x_min, sky_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_f(temp_desc, 
					    sky_y_min, sky_y_max );
                  }
               }
               break;

               case dmDOUBLE:
               {
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_d(temp_desc, 
					    sky_x_min, sky_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_d(temp_desc, 
					    sky_y_min, sky_y_max );
                  }
               }
               break;
               default:
                  dsErrAdd(err_p, dsAPEDMDATATYPEERR, Individual, 
                           Generic, evtout_p->names[rr]);
               break;
            }
         }  
         break;    /* ------------ end: case TG_SKY */


         case TG_DET:   /* vectored coord pair- DETX/DETY */ 
         {
            double    crpix[2],
                      crval[2] = {0.0, 0.0},
                      cdelt[2];
            double    tanp_params[4] = {0.0,     /* Rotation (deg)         */
                                        270.0,   /* pole longitude (deg)   */
                                        90.0,    /* pole latitude (deg)    */
                                        2000.0}; /* equinox (Julian years) */
            VEC2_LONG axlen;
            char  names[2][DS_SZ_KEYWORD];
            char* cptnames[2];
            char* mscnames[] = {"PHI", "THETA"};
            dmDescriptor* temp_desc;
	    double det_x_min, det_x_max, det_y_min, det_y_max;
            cptnames[0] = names[0];
            cptnames[1] = names[1];
            strcpy(cptnames[0], TG_DET_X_COL);
            strcpy(cptnames[1], TG_DET_Y_COL);
 
            /* if output column name is lowercase, set x/y to lc */
            if (islower(evtout_p->names[rr][0]))
            {
               string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
               string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
            }

            evtout_p->dim[rr] = 2;
            evtout_p->desc[rr] =
               dmColumnCreateVector(evtout_p->extension, evtout_p->names[rr],
                                    evtout_p->types[rr], 0,
                                    DS_ACIS_DET_TUNITS, "Det coords",
                                    cptnames, evtout_p->dim[rr]);

            pix_get_fp_image_cntr(crpix);
            cdelt[0] = cdelt[1] = inp_p->fp_scale/3600.0;
            dmCoordCreate_d(evtout_p->desc[rr], "MSC", "deg",
                            mscnames, 2, "TAN-P", crpix, crval, cdelt,
                            tanp_params);

            pix_get_fp_image_dimn(axlen);
	    
	    det_x_min = 0.5 ; det_x_max = DS_HRCS_X_TLMAX;
	    det_y_min = 0.5 ; det_y_max = DS_HRCS_Y_TLMAX;

	    if ( dmGetCpt(evtout_p->desc[rr], 1) && 
		 dmGetCpt(evtout_p->desc[rr], 2 ) )
	      {
		register int cc;
		for (cc = evtin_p->num_cols; cc--;)
		  {
		    if (evtin_p->mapping[cc] == TG_DET_X )
		      {
			dmDescriptorGetRange_d( evtin_p->desc[cc], &det_x_min, 
						&det_x_max );
		      }
		    if (evtin_p->mapping[cc] == TG_DET_Y )
		      {
			dmDescriptorGetRange_d( evtin_p->desc[cc], &det_y_min, 
						&det_y_max );
			
		      }
		  }
	      }


            switch (evtout_p->types[rr])    /* --- TG_DET data type --- */
            {
               case dmLONG:
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_l(temp_desc, 
					    det_x_min, det_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_l(temp_desc, 
					    det_y_min, det_y_max );
                  }
               break;

               case dmSHORT:
               {
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_s(temp_desc,
					    det_x_min, det_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_s(temp_desc, 
					    det_y_min, det_y_max );
                  }
               }
               break;

               case dmFLOAT:
               {
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_f(temp_desc,
					    det_x_min, det_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_f(temp_desc,
					    det_y_min, det_y_max );
                  }
               }
               break;

               case dmDOUBLE:
               {
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
                  {
                     dmDescriptorSetRange_d(temp_desc, 
					    det_x_min, det_x_max );
                  }
                  if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
                  {
                     dmDescriptorSetRange_d(temp_desc, 
					    det_y_min, det_y_max );
                  }
               }
               break;

               default:
                  dsErrAdd(err_p, dsAPEDMDATATYPEERR, Individual, 
                           Generic, evtout_p->names[rr]);
               break;
            }
         } 
         break;      /* ------------ end: case TG_DET -------- */

         case TG_CHIP:   /* vectored coord pair- CHIPX/CHIPY */ 
         {
            double crpix[2] = {0.5, 0.5},
                   crval[2] = {0.0, 0.0},
                   cdelt[2];
            char  names[2][DS_SZ_KEYWORD];
            char* cptnames[2];
            char* cpcnames[] = {"CPCX", "CPCY"};
            dmDescriptor* temp_desc;

            cptnames[0] = names[0];
            cptnames[1] = names[1];
            strcpy(cptnames[0], TG_CHIP_X_COL);
            strcpy(cptnames[1], TG_CHIP_Y_COL);

 
            /* if output column name is lowercase, set x/y to lc */
            if (islower(evtout_p->names[rr][0]))
            {
               string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
               string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
            }
 
            evtout_p->dim[rr] = 2;
            evtout_p->desc[rr] =
               dmColumnCreateVector(evtout_p->extension, evtout_p->names[rr],
                                    evtout_p->types[rr], 0,
                                    DS_HRC_CHIP_TUNITS, "Chip coords",
                                    cptnames, evtout_p->dim[rr]);

            pix_get_chip_scale_in_mm(cdelt);
            dmCoordCreate_d(evtout_p->desc[rr], "CPC", "mm",
                            cpcnames, 2, "LINEAR", crpix, crval, cdelt,
                            NULL);

            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
            {
               register int cc;

               for (cc = evtin_p->num_cols; cc--;)
               {
                  if (evtin_p->mapping[cc] == TG_CHIP_X)
                  {
                     short min = 0, max = 0;
 
                     dmDescriptorGetRange_s(evtin_p->desc[cc], &min, &max);
                     dmDescriptorSetRange_s(temp_desc, min, max);
                     cc = 0; 
                  } 
               } 
            } 
            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
            {
               register int cc;

               for (cc = evtin_p->num_cols; cc--;)
               {
                  if (evtin_p->mapping[cc] == TG_CHIP_Y)
                  {
                     short min = 0, max = 0;
 
                     dmDescriptorGetRange_s(evtin_p->desc[cc], &min, &max);
                     dmDescriptorSetRange_s(temp_desc, min, max);
                     cc = 0; 
                  }
               } 
            } 

         }
         break;  /* ----------- end: TG_CHIP --------------- */



        /* ------------------------------------------------------------------
         * (8/2008) tweaked the original code to work for TG_TDET and TG_RAW.
         *------------------------------------------------------------------*/
         case TG_TDET:   case TG_RAW:       /* uuu */
         {
            dmDescriptor* temp_desc;

            char  names[2][DS_SZ_KEYWORD];
            char* cptnames[2];
            cptnames[0] = names[0];
            cptnames[1] = names[1];

       char unit[100], comm[100] ;
       short tmp_xy_map[2] ;
       if (evtout_p->mapping[rr]==TG_TDET)
       {
          strcpy(cptnames[0], TG_TDET_X_COL);
          strcpy(cptnames[1], TG_TDET_Y_COL);
          strcpy(unit, DS_HRC_TDET_TUNITS );
          strcpy(comm, "Tdet coords" );
          tmp_xy_map[0] = TG_TDET_X ;
          tmp_xy_map[1] = TG_TDET_Y ;
       }
       else if (evtout_p->mapping[rr]==TG_RAW)
       {
          strcpy(cptnames[0], TG_RAW_X_COL);
          strcpy(cptnames[1], TG_RAW_Y_COL);
          strcpy( unit, DS_HRC_RAW_TUNITS );
          strcpy( comm, "raw column" );
          tmp_xy_map[0] = TG_RAW_X ;
          tmp_xy_map[1] = TG_RAW_Y ;
       }

            /* if output column name is lowercase, set x/y to lc */
            if (islower(evtout_p->names[rr][0]))     /* tdet, raw */
            {
               string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
               string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
            }
 
            evtout_p->dim[rr] = 2;
            evtout_p->desc[rr] =
               dmColumnCreateVector(evtout_p->extension, evtout_p->names[rr],
                                    evtout_p->types[rr], 0,
                                    unit, comm,
                                    cptnames, evtout_p->dim[rr]);
            /*----------------------------------------------------------
             *  BEGIN: loop for vector components TDETX/RAWX (idx=1) 
             *                                and TDETY/RAWY (idx=2).
             *----------------------------------------------------------*/
int idx ;
for (idx=1; idx<=2; idx++)
{
   temp_desc = NULL ;

            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], idx)))  /* 1=Xcomp; 2=Ycomp;*/
            {
               register int cc;
 
               /* 8/2008:   for (cc = evtin_p->num_cols; cc--;) */
               for (cc = (evtin_p->num_cols-1); cc>=0; cc-- )
               {
                  if (evtin_p->mapping[cc] == tmp_xy_map[idx-1] )
                  {
                     switch(evtin_p->types[cc])
                     {
                        case dmSHORT:
                        {
                           short min = 0, max = 0;
                           dmDescriptorGetRange_s(evtin_p->desc[cc], &min, &max);
                           dmDescriptorSetRange_s(temp_desc, min, max);

                           /*6/17/2015 + 11/18/2015 */
                           if (evtout_p->mapping[rr]==TG_TDET)
                           {  short nil = 0 ;
                              if (dmDescriptorGetNull_s(evtin_p->desc[cc], &nil) != dmFAILURE)
                                 dmDescriptorSetNull_s(temp_desc, nil);
                              else 
                                 dmDescriptorSetNull_s(temp_desc, 9999 ) ;
                           }

                        }
                        break;
                        case dmUSHORT:
                        {
                           unsigned short min = 0, max = 0;
                           dmDescriptorGetRange_us(evtin_p->desc[cc], &min, &max);
                           dmDescriptorSetRange_us(temp_desc, min, max);

                           /*6/17/2015 + 11/18/2015 */
                           if (evtout_p->mapping[rr]==TG_TDET)
                           {  short nil = 0 ;
                              if (dmDescriptorGetNull_s(evtin_p->desc[cc], &nil) != dmFAILURE)
                                 dmDescriptorSetNull_s(temp_desc, nil);
                              else 
                                 dmDescriptorSetNull_s(temp_desc, 9999 ) ;
                           }

                        }
                        break;
                        case dmLONG:
                        {
                           long min = 0, max = 0;
                           dmDescriptorGetRange_l(evtin_p->desc[cc], &min, &max);
                           dmDescriptorSetRange_l(temp_desc, min, max);

                           /*6/17/2015  + 11/18/2015 - add */
                           if (evtout_p->mapping[rr]==TG_TDET)
                           {
                              long int nil = 0 ;
                              if (dmDescriptorGetNull_l(evtin_p->desc[cc], &nil) != dmFAILURE)
                                 dmDescriptorSetNull_l(temp_desc, nil);
                              else 
                              {
                                 /* HRC is in long dtype, min:max=1:49368 */ 
                                 if ( inp_p->instrume_type == TG_INSTR_HRC )
                                 {
                                    dmDescriptorSetNull_l(temp_desc, 99999 ) ;
                                 }
                                 else
                                 {
                                    dmDescriptorSetNull_l(temp_desc, 9999 ) ;
                                 }
                              }
                           }

                        }
                        break;
                        case dmULONG:
                        {
                           unsigned long min = 0, max = 0;
                           dmDescriptorGetRange_ul(evtin_p->desc[cc], &min, &max);
                           dmDescriptorSetRange_ul(temp_desc, min, max);

                           /*6/17/2015 + 11/18/2015 */
                           if (evtout_p->mapping[rr]==TG_TDET)
                           {
                              long int nil = 0 ;
                              if (dmDescriptorGetNull_l(evtin_p->desc[cc], &nil) != dmFAILURE)
                                 dmDescriptorSetNull_l(temp_desc, nil);
                              else 
                              {
                                 if ( inp_p->instrume_type == TG_INSTR_HRC )
                                    dmDescriptorSetNull_l(temp_desc, 99999 ) ;
                                 else
                                    dmDescriptorSetNull_l(temp_desc, 9999 ) ;
                              }
                           }
                        }
                        break;
                        case dmFLOAT:
                        {
                           float min = 0.0, max = 0.0;
                           dmDescriptorGetRange_f(evtin_p->desc[cc], &min, &max);
                           dmDescriptorSetRange_f(temp_desc, min, max);

                           /*6/17/2015 + 11/18/2015 */
                           if (evtout_p->mapping[rr]==TG_TDET)
                           {
                              float nil = 0 ;
                              if (dmDescriptorGetNull_f(evtin_p->desc[cc], &nil) != dmFAILURE)
                                 dmDescriptorSetNull_f(temp_desc, nil);
                              else 
                              {
                                 if ( inp_p->instrume_type == TG_INSTR_HRC )
                                    dmDescriptorSetNull_f(temp_desc, 99999.0 ) ;
                                 else
                                    dmDescriptorSetNull_f(temp_desc, 9999.0 ) ;
                              }
                           }

                        }
                        break;
                        case dmDOUBLE:
                        {
                           double min = 0.0, max = 0.0;
                           dmDescriptorGetRange_d(evtin_p->desc[cc], &min, &max);
                           dmDescriptorSetRange_d(temp_desc, min, max);

                           /*6/17/2015 + 11/18/2015 */
                           if (evtout_p->mapping[rr]==TG_TDET)
                           {
                              double nil = 0 ;
                              if (dmDescriptorGetNull_d(evtin_p->desc[cc], &nil) != dmFAILURE)
                                 dmDescriptorSetNull_d(temp_desc, nil);
                              else 
                              {
                                 if ( inp_p->instrume_type == TG_INSTR_HRC )
                                    dmDescriptorSetNull_d(temp_desc, 99999.0 ) ;
                                 else
                                    dmDescriptorSetNull_d(temp_desc, 9999.0 ) ;
                              }
                           }

                        }
                        break;
                        case dmTEXT:  /*fallthrough intended */
                        default:
                           dsErrAdd(err_p, dsAPEDMDATATYPEERR, Individual, 
                              Generic, evtout_p->names[rr]);
                        break; 
                     } /* end:  switch ( evtin_p->types[cc] ) */ 
                     cc = 0;
                  } /* end: if (evtin_p->mapping[cc]==tmp_xy_map[idx-1] ) */
               } /* end:  loop for evtin_p->num_cols */ 
            }  /* end: if (dmGetCpt(evtout_p->desc[rr],2))!=NULL)*/
} /* END: loop of vector components idx=1,2 ; */
         }
         break;  /* ------------ end:  case TG_TDET ; TG_RAW ; ------------*/



         case TG_GAC:     /* column rd(tg_r,tg_d) */ 
         {
            char  names[2][DS_SZ_KEYWORD];
            char* cptnames[2];
            dmDescriptor* temp_desc;

            cptnames[0] = names[0];
            cptnames[1] = names[1];
            strcpy(cptnames[0], TG_GAC_R_COL);
            strcpy(cptnames[1], TG_GAC_D_COL);

 
            /* if output column name is lowercase, set x/y to lc */
            if (islower(evtout_p->names[rr][0]))
            {
               string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
               string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
            }
 
            evtout_p->dim[rr] = 2;
            evtout_p->desc[rr] =
               dmColumnCreateVector(evtout_p->extension, evtout_p->names[rr],
                                    evtout_p->types[rr], 0,
                                    DS_GRT_TG_RD_TUNITS, 
                                    "Grating angular coords",
                                    cptnames, evtout_p->dim[rr]);

            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
            {
               dmDescriptorSetRange_d(temp_desc, DS_GRT_TG_R_TLMIN, 
                                      DS_GRT_TG_R_TLMAX);
               switch (evtout_p->types[rr])
               {
                  case dmDOUBLE:
                     dmDescriptorSetBin_d(temp_desc, DS_GRT_TG_R_TDBIN);
                  break;
                  case dmFLOAT:
                     dmDescriptorSetBin_f(temp_desc, DS_GRT_TG_R_TDBIN);
                  break;
                  default:
                     /* do nothing */
                  break;
               }
            } 
            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
            {
               dmDescriptorSetRange_d(temp_desc, DS_GRT_TG_D_TLMIN,
                                      DS_GRT_TG_D_TLMAX);
               switch (evtout_p->types[rr])
               {
                  case dmDOUBLE:
                     dmDescriptorSetBin_d(temp_desc, DS_GRT_TG_D_TDBIN);
                  break;
                  case dmFLOAT:
                     dmDescriptorSetBin_f(temp_desc, DS_GRT_TG_D_TDBIN);
                  break;
                  default:
                     /* do nothing */
                  break;
               }
            } 
         }
         break;  /* ------------ end: TG_GAC ------------*/

         case TG_GDP:     /* column gdp(gdpx,gdpy) */ 
         {
            double    scale,
                      crpix[2],
                      crval[2] = {0.0, 0.0},
                      cdelt[2];
            char      names[2][DS_SZ_KEYWORD];
            char*     cptnames[2];
            char*     gacnames[] = {"TG_R", "TG_D"};
            dmDescriptor* temp_desc;

            cptnames[0] = names[0];
            cptnames[1] = names[1];
            strcpy(cptnames[0], TG_GDP_X_COL);
            strcpy(cptnames[1], TG_GDP_Y_COL);
 
            /* if output column name is lowercase, set x/y to lc */
            if (islower(evtout_p->names[rr][0]))
            {
               string_to_lowercase(cptnames[0], DS_SZ_KEYWORD);
               string_to_lowercase(cptnames[1], DS_SZ_KEYWORD);
            }
 
            evtout_p->dim[rr] = 2;
            evtout_p->desc[rr] =
               dmColumnCreateVector(evtout_p->extension, evtout_p->names[rr],
                                    evtout_p->types[rr], 0,
                                    DS_GRT_GDP_TUNITS, 
                                    "Grating pixel coords",
                                    cptnames, evtout_p->dim[rr]);

            pix_get_gdp_scale_in_asec(&scale); 
            cdelt[0] = scale/-3600.0;
            cdelt[1] = cdelt[0]*-1;
            pix_get_gdp_image_cntr(crpix);
            /* needs to be TAN-P but currently not supported by DM */
            dmCoordCreate_d(evtout_p->desc[rr], "GAC", "deg",
                            gacnames, 2, "LINEAR", crpix, crval, cdelt,
                            NULL);

            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 1)))
            {
               tg_write_column_range(temp_desc, evtout_p->types[rr],
                                  DS_GRT_GDPX_TLMIN, DS_GRT_GDPX_TLMAX); 
               switch (evtout_p->types[rr])
               {
                  case dmSHORT:
                     dmDescriptorSetBin_s(temp_desc, DS_GRT_GDPX_TDBIN);
                  break;
                  case dmUSHORT:
                     dmDescriptorSetBin_us(temp_desc, DS_GRT_GDPX_TDBIN);
                  break;
                  case dmLONG:
                     dmDescriptorSetBin_l(temp_desc, DS_GRT_GDPX_TDBIN);
                  break;
                  case dmULONG:
                     dmDescriptorSetBin_ul(temp_desc, DS_GRT_GDPX_TDBIN);
                  break;
                  default:
                     /* do nothing */
                  break;
               }
            }
            if ((temp_desc = dmGetCpt(evtout_p->desc[rr], 2)))
            {
               tg_write_column_range(temp_desc, evtout_p->types[rr],
                                  DS_GRT_GDPY_TLMIN, DS_GRT_GDPY_TLMAX); 
               switch (evtout_p->types[rr])
               {
                  case dmSHORT:
                     dmDescriptorSetBin_s(temp_desc, DS_GRT_GDPY_TDBIN);
                  break;
                  case dmUSHORT:
                     dmDescriptorSetBin_us(temp_desc, DS_GRT_GDPY_TDBIN);
                  break;
                  case dmLONG:
                     dmDescriptorSetBin_l(temp_desc, DS_GRT_GDPY_TDBIN);
                  break;
                  case dmULONG:
                     dmDescriptorSetBin_ul(temp_desc, DS_GRT_GDPY_TDBIN);
                  break;
                  default:
                     /* do nothing */
                  break;
               }
            }

            switch (evtout_p->types[rr])
            {
               case dmSHORT:
                  dmDescriptorSetNull_s(evtout_p->desc[rr], 0);
               break;
               case dmUSHORT:
                  dmDescriptorSetNull_us(evtout_p->desc[rr], 0);
               break;
               case dmLONG:
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 0);
               break;
               case dmULONG:
                  dmDescriptorSetNull_ul(evtout_p->desc[rr], 0);
               break;
               default:
                  /* do nothing */
               break;
            }
         }
         break;   /* ------------ end: TG_GDP ------------*/

         case TG_GAC_R:         /* column tg_r */
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, 
                              DS_GRT_TG_RD_TUNITS, 
                              "Diffraction angle, relative to zero-order centroid");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_R_TLMIN, DS_GRT_TG_R_TLMAX); 
            switch (evtout_p->types[rr])
            {
               case dmDOUBLE:
                  dmDescriptorSetBin_d(evtout_p->desc[rr], DS_GRT_TG_R_TDBIN);
               break;
               case dmFLOAT:
                  dmDescriptorSetBin_f(evtout_p->desc[rr], DS_GRT_TG_R_TDBIN);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;  /* ------------ end: TG_GAC_R ------------*/

         case TG_GAC_D:        /* column tg_d */
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, 
                              DS_GRT_TG_RD_TUNITS, 
                              "Cross-dispersion angle");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_D_TLMIN, DS_GRT_TG_D_TLMAX); 
            switch (evtout_p->types[rr])
            {
               case dmDOUBLE:
                  dmDescriptorSetBin_d(evtout_p->desc[rr], DS_GRT_TG_D_TDBIN);
               break;
               case dmFLOAT:
                  dmDescriptorSetBin_f(evtout_p->desc[rr], DS_GRT_TG_D_TDBIN);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;  /* ------------ end: TG_GAC_D ------------*/

         case TG_GDP_X:         /* column gdpx */
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, 
                              DS_GRT_GDP_TUNITS, 
                            "Grating diffraction pixel, dispersion direction");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_GDPX_TLMIN, DS_GRT_GDPX_TLMAX); 

            switch (evtout_p->types[rr])
            {
               case dmSHORT:
                  dmDescriptorSetNull_s(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_s(evtout_p->desc[rr], DS_GRT_GDPX_TDBIN);
               break;
               case dmUSHORT:
                  dmDescriptorSetNull_us(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_us(evtout_p->desc[rr], DS_GRT_GDPX_TDBIN);
               break;
               case dmLONG:
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_l(evtout_p->desc[rr], DS_GRT_GDPX_TDBIN);
               break;
               case dmULONG:
                  dmDescriptorSetNull_ul(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_ul(evtout_p->desc[rr], DS_GRT_GDPX_TDBIN);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;  /* ------------ end: TG_GDP_X ------------*/

         case TG_GDP_Y:         /* column gdpy */
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0,
                              DS_GRT_GDP_TUNITS,
                              "Grating diffraction pixel, cross-dispersion direction");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_GDPY_TLMIN, DS_GRT_GDPY_TLMAX); 

            switch (evtout_p->types[rr])
            {
               case dmSHORT:
                  dmDescriptorSetNull_s(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_s(evtout_p->desc[rr], DS_GRT_GDPY_TDBIN);
               break;
               case dmUSHORT:
                  dmDescriptorSetNull_us(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_us(evtout_p->desc[rr], DS_GRT_GDPY_TDBIN);
               break;
               case dmLONG:
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_l(evtout_p->desc[rr], DS_GRT_GDPY_TDBIN);
               break;
               case dmULONG:
                  dmDescriptorSetNull_ul(evtout_p->desc[rr], 0);
                  dmDescriptorSetBin_ul(evtout_p->desc[rr], DS_GRT_GDPY_TDBIN);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;   /* ------------ end: TG_GDP_Y ------------*/

         case TG_PART:
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, "", "component index (HEG, MEG, LEG, HESF regions)");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_PART_TLMIN, DS_GRT_TG_PART_TLMAX); 
         break;   /* ------------ end: TG_PART ------------*/

         case TG_M:
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, "", "Diffraction order (m)");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_M_TLMIN, DS_GRT_TG_M_TLMAX); 

            switch (evtout_p->types[rr])
            {
               case dmSHORT:
                  dmDescriptorSetNull_s(evtout_p->desc[rr], 99);
               break;
               case dmUSHORT:
                  dmDescriptorSetNull_us(evtout_p->desc[rr], 99);
               break;
               case dmLONG:
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 99);
               break;
               case dmULONG:
                  dmDescriptorSetNull_ul(evtout_p->desc[rr], 99);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;  /* ------------ end: TG_M ------------*/

         case TG_MLAM:
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, 
                              DS_GRT_TG_MLAM_TUNITS, "Order times wavelength (m * lambda)");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_MLAM_TLMIN, DS_GRT_TG_MLAM_TLMAX); 
            switch (evtout_p->types[rr])
            {
               case dmDOUBLE:
                  dmDescriptorSetBin_d(evtout_p->desc[rr], DS_GRT_TG_MLAM_TDBIN);
               break;
               case dmFLOAT:
                  dmDescriptorSetBin_f(evtout_p->desc[rr], DS_GRT_TG_MLAM_TDBIN);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;   /* ------------ end: TG_MLAM ------------*/

         case TG_LAM:
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, 
                              DS_GRT_TG_LAM_TUNITS, "wavelength (lambda)");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_LAM_TLMIN, DS_GRT_TG_LAM_TLMAX); 
            switch (evtout_p->types[rr])
            {
               case dmDOUBLE:
                  dmDescriptorSetBin_d(evtout_p->desc[rr], DS_GRT_TG_LAM_TDBIN);
               break;
               case dmFLOAT:
                  dmDescriptorSetBin_f(evtout_p->desc[rr], DS_GRT_TG_LAM_TDBIN);
               break;
               default:
                  /* do nothing */
               break;
            }
         break;   /* ------------ end: TG_LAM ------------*/

         case TG_SRCID:
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, "", "source ID, index from detect table");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_SRCID_TLMIN, 
                                  DS_GRT_TG_SRCID_TLMAX); 
         break;   /* ------------ end: SRCID ------------*/

         case TG_SMAP:
            evtout_p->desc[rr] =
               dmColumnCreate(evtout_p->extension,
                              evtout_p->names[rr],
                              evtout_p->types[rr], 0, "", "source map; flags for up to 10 sources");
            tg_write_column_range(evtout_p->desc[rr], evtout_p->types[rr],
                                  DS_GRT_TG_SMAP_TLMIN, DS_GRT_TG_SMAP_TLMAX); 
         break;   /* ------------ end: TG_SMAP ------------*/

         case TG_PHAS:      
         {
            long pha_axes[2]; 

            if (evtout_p->island_sz == 25)
            {
               pha_axes[0] = pha_axes[1] = 5; 
            }
            else 
            {
               pha_axes[0] = pha_axes[1] = 3; 
            }

            evtout_p->desc[rr] =
               dmColumnCreateNDArray(evtout_p->extension, evtout_p->names[rr],
                                     evtout_p->types[rr], 0,
                                     DS_ACIS_PHAS_TUNITS,
                                     "array of pixel pulse heights",
                                     pha_axes, 2);

               dmDescriptorSetRange_s(evtout_p->desc[rr],
                                      DS_ACIS_PHAS_TLMIN, DS_ACIS_PHAS_TLMAX);
         }
         break;   /* ------------ end: TG_PHAS ------------*/

         case TG_PHAS_ADJ:
         {
            long pha_axes[2];

            if (evtout_p->island_sz == 25)
            {
               pha_axes[0] = pha_axes[1] = 5;
            }
            else 
            {
               pha_axes[0] = pha_axes[1] = 3;
            }

            evtout_p->desc[rr] =
               dmColumnCreateNDArray(evtout_p->extension, evtout_p->names[rr],
                                     evtout_p->types[rr], 0,
                                     DS_ACIS_PHAS_TUNITS,
                                     "array of cti adjusted pulse heights",
                                     pha_axes, 2);

               dmDescriptorSetRange_f(evtout_p->desc[rr],
                                      DS_ACIS_PHAS_TLMIN, DS_ACIS_PHAS_TLMAX);
         }
         break;    /* ------------ end: TG_PHAS_ADJ ------------*/

         case TG_BIAS:      
            evtout_p->desc[rr] =
               dmColumnCreateArray(evtout_p->extension, evtout_p->names[rr],
                                   evtout_p->types[rr], 0, "", 
                                   "3x3 array of pixel bias map values", 9); 
         break;    /* ------------ end: TG_BIAS ------------*/

         case TG_TIME:
           evtout_p->desc[rr] =
              dmColumnCreate(evtout_p->extension, evtout_p->names[rr],
                             evtout_p->types[rr], 0, DS_HRC_TIME_TUNITS, 
                             "time tag of data record");
         break;    /* ------------ end: TG_TIME ------------*/

         case TG_TIME_RO:
           evtout_p->desc[rr] =
              dmColumnCreate(evtout_p->extension, evtout_p->names[rr],
                             evtout_p->types[rr], 0, DS_ACIS_TIME_TUNITS, 
                             "S/C TT corresponding to mid-exposure");
         break;    /* ------------ end: TG_TIME_RO ------------*/

         case TG_CHIPY_TG:
         {
           evtout_p->desc[rr] = 
              dmColumnCreate(evtout_p->extension, evtout_p->names[rr], evtout_p->types[rr],
                             0, DS_ACIS_CHIP_TUNITS, "chip y column");
           /* tg_write_column_range (evtout_p->desc[rr], evtout_p->types[rr],            */
           /*               (double)TG_ACIS_MIN_CHIP_VAL, (double)TG_ACIS_MAX_CHIP_VAL); */

         } /* end: case tg_chipy_tg */
         break;   /* ------------ end: TG_CHIPY_TG ----------*/

         case TG_CHIPY_ZO:
         {
           evtout_p->desc[rr] = 
              dmColumnCreate(evtout_p->extension, evtout_p->names[rr], evtout_p->types[rr],
                             0, DS_ACIS_CHIP_TUNITS, "chip y column");
           /* tg_write_column_range (evtout_p->desc[rr], evtout_p->types[rr],              */
           /*               (double)TG_ACIS_MIN_CHIP_VAL, (double)TG_ACIS_MAX_CHIP_VAL);   */

         } /* end: case tg_chipy_zo */
         break;   /* ------------ end: TG_CHIPY_ZO ----------*/

         case TG_STATUS:
           if (evtout_p->types[rr] == dmBIT)
           {
              register int cc;
              int size = 4;

              for (cc = evtin_p->num_cols; cc--;)
              {
                 if (evtin_p->mapping[cc] == TG_STATUS)
                 {
                    if (dmGetArraySize(evtin_p->desc[cc]) == 2)
                    {
                       size = 2; 
                    } 
                 }
              } 

              evtout_p->desc[rr] =
                 dmColumnCreateArray(evtout_p->extension, evtout_p->names[rr],
                    evtout_p->types[rr],  0, "", "event status bits", size);
           }
           else
           {
              evtout_p->desc[rr] =
                dmColumnCreate(evtout_p->extension, evtout_p->names[rr],
                               evtout_p->types[rr],  0, "coded", 
                               "event status bits");
           }
         break;    /* ------------ end: TG_STATUS ------------*/


       /* ------------------------------------------------------------------------
        * (8/2008) any scalar column (eg. pha_ro/rawx/rawy) can be created here. 
        * set the output column range if infile contains it. 
        * -----------------------------------------------------------------------*/
         default:
         {
            /*## printf("\n evtout_p->names[rr]=%s,\n", evtout_p->names[rr]) ; */
            char  unit[100], comm[100];
            memset(unit,0,100);
            memset(comm,0,100);
            if (evtout_p->mapping[rr]==TG_PHA_RO)
            {
               strcpy( unit, DS_ACIS_PHA_TUNITS );
               strcpy( comm, "readout pulse height column" );
            }
            else if ((evtout_p->mapping[rr]==TG_RAW_X)||
                     (evtout_p->mapping[rr]==TG_RAW_Y) )
            {
               strcpy( unit, DS_HRC_RAW_TUNITS );
               strcpy( comm, "raw column" );
            }
            /* 9/24/2012 - bug 13374 */
            else  
            {
               short int stop=0, ss;
               for (ss = evtin_p->num_cols; ss>=0 && stop==0 ; ss--)
               {
                  if (evtout_p->mapping[rr] == evtin_p->mapping[ss])
                  {
                     dmGetUnit( evtin_p->desc[ss], unit, 100 );
                     stop = 1 ; 
                  }
               }
            }


            register int cc; 

            evtout_p->desc[rr] =
              dmColumnCreate(evtout_p->extension, evtout_p->names[rr],
                             evtout_p->types[rr], 0, unit, comm );

            for (cc = evtin_p->num_cols; cc--;)
            {
               if (evtout_p->mapping[rr] == evtin_p->mapping[cc])
               {
                  switch (evtin_p->types[cc])
                  {

        /* -----------------------------------------------------------------
         * 6/17/2015+ 11/18/2015 : add tnulls for the scalar columns of tdetx,tdety : 
         *  ( this is only for when eventdef has s:tdetx,s:tdety )
         *  ( For eventdef has s:tdet, then we set tnull for tdetx,tdety  at uuu )
         * ----------------------------------------------------------------- */              
                     case dmSHORT:
                     {
                        short min = 0, max = 0, nil = 0; 

                        if (dmDescriptorGetRange_s(evtin_p->desc[cc],  
                            &min, &max) != dmFAILURE)  
                        {
                           tg_write_column_range(evtout_p->desc[rr],
                                                 evtout_p->types[rr],
                                                 (double) min, (double) max);
                        }
                        if (dmDescriptorGetNull_s(evtin_p->desc[cc], 
                            &nil) != dmFAILURE)
                        {
                           dmDescriptorSetNull_s(evtout_p->desc[rr], nil); 
                        }
         else  
         {
            /* 6/17/2015 + 11/18/2015 */
            if ((evtin_p->mapping[cc]==TG_TDET_X)||(evtin_p->mapping[cc]==TG_TDET_Y))
               dmDescriptorSetNull_s(evtout_p->desc[rr], 9999 ); 
         }
                     }
                     break;
                     case dmLONG:
                     {
                        long min = 0, max = 0, nil = 0; 
  
                        if (dmDescriptorGetRange_l(evtin_p->desc[cc], 
                            &min, &max) != dmFAILURE) 
                        {
                           tg_write_column_range(evtout_p->desc[rr],
                                                 evtout_p->types[rr],
                                                 (double) min, (double) max);
                        }
                        if (dmDescriptorGetNull_l(evtin_p->desc[cc],
                            &nil) != dmFAILURE)
                        {
                           dmDescriptorSetNull_l(evtout_p->desc[rr], nil); 
                        }
         else  
         {
            /* 6/17/2015 + 11/18/2015 */
            if ((evtin_p->mapping[cc]==TG_TDET_X)||(evtin_p->mapping[cc]==TG_TDET_Y))
            {
               /* HRC is in long dtype, min:max=1:49368 */
               if ( inp_p->instrume_type == TG_INSTR_HRC )
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 99999 ) ;
               else
                  dmDescriptorSetNull_l(evtout_p->desc[rr], 9999 ) ;
            }
         }
                     }
                     break;
                     case dmFLOAT:
                     {
                        float min = 0.0, max = 0.0, nil = 0.0;
 
                        if (dmDescriptorGetRange_f(evtin_p->desc[cc], 
                            &min, &max) != dmFAILURE)
                        {
                           tg_write_column_range(evtout_p->desc[rr], 
                                                 evtout_p->types[rr],
                                                 (double) min, (double) max); 
                        }
                        if (dmDescriptorGetNull_f(evtin_p->desc[cc],
                            &nil) != dmFAILURE)
                        {
                           dmDescriptorSetNull_f(evtout_p->desc[rr], nil); 
                        }
         else  
         {
            /* 6/17/2015 + 11/18/2015 */
            if ((evtin_p->mapping[cc]==TG_TDET_X)||(evtin_p->mapping[cc]==TG_TDET_Y))
            {
               /* HRC is in long dtype, min:max=1:49368 */
               if ( inp_p->instrume_type == TG_INSTR_HRC )
                  dmDescriptorSetNull_f(evtout_p->desc[rr], 99999.0 ) ;
               else
                  dmDescriptorSetNull_f(evtout_p->desc[rr], 9999.0 ) ;
            }
         }

                     }
                     break;
                     case dmDOUBLE:
                     {
                        double min = 0.0, max = 0.0, nil = 0.0;
 
                        if (dmDescriptorGetRange_d(evtin_p->desc[cc], 
                            &min, &max) != dmFAILURE)
                        { 
                           tg_write_column_range(evtout_p->desc[rr], 
                                                 evtout_p->types[rr],
                                                 min, max); 
                        }
                        if (dmDescriptorGetNull_d(evtin_p->desc[cc],
                            &nil) != dmFAILURE)
                        {
                           dmDescriptorSetNull_d(evtout_p->desc[rr], nil); 
                        }
         else  
         {
            /* 6/17/2015 + 11/18/2015 */
            if ((evtin_p->mapping[cc]==TG_TDET_X)||(evtin_p->mapping[cc]==TG_TDET_Y))
            {
               /* HRC is in long dtype, min:max=1:49368 */
               if ( inp_p->instrume_type == TG_INSTR_HRC )
                  dmDescriptorSetNull_f( evtout_p->desc[rr], 99999.0 ) ;
               else
                  dmDescriptorSetNull_f( evtout_p->desc[rr], 9999.0 ) ;
            }
         }
                     }
                     break;
                        /* unknown type- don't set range */
                     default:
                     break; 
                  }
                  cc = 0;
               }
            }
         } /* end default mapping */  

         break; 
      }  /* end switch (mapping) */ 
   } /* end for (rr) */ 

} /* end: tg_setup_columns() */

